/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/** \mainpage

    \htmlonly
    <a href="../latex/refman.pdf">PDF documentation</a>
    \endhtmlonly

    \hline
    \section part_ug_section Quick Reference to User's Guide
    - \ref part_section
    - \ref nuclei_section
    - \ref partsource_section
    - \ref partref_section

    \hline
    \section part_section Particles

    These classes in the library \o2p calculate the
    thermodynamic properties of interacting and non-interacting
    quantum and \classical particles.

    The class \ref part is the basic structure for a particle:
    - part::m - mass
    - part::g - degeneracy factor (e.g. \f$2j+1\f$)
    - part::n - number density
    - part::ed - energy density
    - part::pr - pressure
    - part::en - entropy density
    - part::ms - effective mass, i.e. \f$ m^{*} \f$
    - part::nu - effective chemical potential, i.e. \f$ \nu \f$
    - part::inc_rest_mass - True if the rest mass is included
    - part::non_interacting - False if the particle includes interactions

    The data members part::ms and part::nu allow one to specify
    modifications to the mass and the chemical potential due to
    interactions. This allows one to calculate the properties of
    particle due to interactions so long as the basic form of the
    free-particle dispersion relation is unchanged, i.e.
    \f[
    \sqrt{k^2+m^2} - \mu \rightarrow \sqrt{k^2+m^{* 2}} - \nu 
    \f]
    If the particle is non-interacting, then part::nu and part::ms are
    sometimes used by \o2p functions for temporary storage.

    If part::inc_rest_mass is \c true (as is the default in all of the
    classes except \ref nucleus), then all
    functions include the rest mass energy density in the energy
    density, the "mu" functions expect that the rest mass is included
    in part::mu or part::nu as input and the "density" functions
    output part::mu or part::nu including the rest mass.
    
    When part::inc_rest_mass is true, antiparticles are implemented by
    choosing the antiparticle chemical potential to be \f$ - \mu \f$,
    and when inc_rest_mass is false, antiparticles are implemented by
    choosing the chemical potential of the antiparticles to be \f$ -
    \mu - 2 m \f$.

    The thermodynamic identity used to compute the pressure for
    interacting particles is
    \f[
    P = -\varepsilon + s T + \nu n
    \f]
    where part::nu is used. This way, the particle class doesn't need to
    know about the structure of the interactions to ensure that the
    thermodynamic identity is satisfied. Note that in the \o2e
    library, where in the equations of state the normal thermodynamic
    identity is used
    \f[
    P = -\varepsilon + s T + \mu n
    \f]
    Frequently, the interactions which create an effective chemical
    potential which is different than part::mu thus create extra terms in
    the pressure and the energy density for the given equation of
    state.

    At zero temperature, fermions and bosons can be treated exactly in
    the classes \ref fermion and \ref boson. The \ref quark class is a
    descendant of the \ref fermion class which contains extra data
    members for the quark condensate and the contribution to the bag
    constant. The \ref classical classical is a descendant of both
    fermion and boson and calculates everything in the classical
    limit.

    At finite temperature, there are different classes corresponding
    to different approaches to computing the integrals over the
    distribution functions. The approximation scheme from \ref Johns96
    is used in eff_boson, eff_fermion, and eff_quark. An exact method
    employing direct integration of the distribution functions is used
    in rel_boson and rel_fermion, but these are necessarily quite a
    bit slower. 

    The class \ref eff_fermion usually works to within about 1 part in
    \f$ 10^4 \f$, but can be as bad as 1 part in \f$ 10^2 \f$ in some
    more extreme cases. The default settings for \ref rel_fermion give
    an accuracy of at least 1 part in \f$ 10^6 \f$ (and frequently
    better than this). For \ref rel_fermion, some additional accuracy
    may be obtained by decreasing the integration tolerances.
    \comment
    7/6/09 - These tolerances were determined from bm_part2.cpp using
    the mathematica numbers in internal/partmath.cpp. I'm not sure the
    testing in bm_part.cpp is really that useful. 
    \endcomment

    The class nonrel_fermion assumes a non-relativistic dispersion
    relation for fermions. It includes zero-temperature methods and an
    exact method for finite temperatures. The non-relativistic
    integrands are much simpler and \ref nonrel_fermion uses the
    appropriate GSL functions (which are nearly exact) to compute
    them.

    \hline

    <b>Units:</b>

    Factors of \f$ \hbar, c \f$ and \f$ k_B \f$ have been removed
    everywhere, so that mass, energy, and temperature all have the
    same units. Number and entropy densities have units of mass cubed
    (or energy cubed). This particle classes can be used with any
    system of units which is based on powers of one unit, i.e. 
    \f$ [n] = [T]^3 = [m]^3 = [P]^{3/4} = [\varepsilon]^{3/4}\f$, etc.

    \hline

    <b>Derivative information:</b>

    Sometimes it is useful to know derivatives like $ds/dT$ in
    addition to the energy and pressure. There are three classes which
    compute these derivatives for fermions and classical particles.
    The class \ref sn_classical handles the nondegenerate limit, \ref
    sn_fermion handles fermions and \ref sn_nr_fermion handles
    nonrelativistic fermions. These classes compute the derivatives
    \f[
    \left(\frac{d n}{d \mu}\right)_{T}, \quad
    \left(\frac{d n}{d T}\right)_{\mu}, \quad \mathrm{and} \quad
    \left(\frac{d s}{d T}\right)_{\mu}
    \f]
    All other first derivatives of the thermodynamic functions can
    be written in terms of these three. To see how to compute the
    specific heat, for example, see the discussion in the documentation
    of \ref deriv_part. 

    \hline
    \section nuclei_section Atomic nuclei

    \b Nuclei

    Atomic nuclei, class \ref nucleus, are implemented as descendants of
    \ref classical. This class sets the value of nucleus::inc_rest_mass to
    false by default. 

    Nuclear mass formulas are given as children of \ref nuclear_mass.
    The class ame_mass provides the experimental data from \ref Audi95
    or \ref Audi03, the class mnmsk_mass provides the mass formula
    from \ref Moller95, and the class hfb_mass provides the mass
    formula from \ref Goriely02, \ref Samyn04, or \ref Goriely07. A
    simple semi-empirical mass formula is given in \ref
    semi_empirical_mass and this can be fit to experimentally measured
    masses using \ref mass_fit.
    
    The class \ref nuclear_dist provides an generic base
    class for a \collection of several nuclei with an STL-like
    iterator. There are two implementations of this base class,
    simple_dist which provides a simple distribution and 
    \ref full_dist which enumerates all the nuclei for a 
    given mass formula.

    \hline
    \section partsource_section Example source code
    
    \subsection exlist_subsect Example list
    
    - \ref ex_part_sect
    - \ref ex_mass_fit_sect

    \subsection ex_part_sect Particle example

    \dontinclude ex_part.cpp
    \skip Example:
    \until End of example

    \subsection ex_mass_fit_sect Nuclear mass fit example

    \dontinclude ex_mass_fit.cpp
    \skip Example:
    \until End of example

    \hline
    \section partref_section Bibliography

    Some of the references which contain links should direct you to
    the work referred to directly through dx.doi.org.

    \anchor Audi95 Audi95: 
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(95)00445-9">
    G. Audi and A. H. Wapstra</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(95)00445-9}{
    G. Audi and A. H. Wapstra},
    \endlatexonly
    Nucl. Phys. A \b 595 (1995) 409-480.

    \anchor Audi03 Audi03: 
    \htmlonly
    <a href="http://dx.doi.org/10.1016/j.nuclphysa.2003.11.003">
    G.Audi, A. H. Wapstra and C. Thibault</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/j.nuclphysa.2003.11.003}{
    G.Audi, A. H. Wapstra and C. Thibault} ,
    \endlatexonly
    Nucl. Phys. A \b 729 (2003) 337.

    \anchor Callen Callen
    H. B. Callen, "Thermodynamics and an Introduction to
    Thermostatistics.", 2nd edition.

    \anchor Eggleton73 Eggleton73:
    \htmlonly
    <a href="http://adsabs.harvard.edu/abs/1973A%26A....23..325E">
    P. P. Eggleton, J. Faulkner, and B. P. Flannery</a>,
    \endhtmlonly
    \latexonly
    \href{http://adsabs.harvard.edu/abs/1973A%26A....23..325E}{
    P. P. Eggleton, J. Faulkner, and B. P. Flannery},
    \endlatexonly
    Astron. and Astrophys. \b 23 (1973) 325.
    \comment
    As of 3/11/09 a DOI link does not appear to be available
    \endcomment

    \anchor Goriely02 Goriely02:
    \htmlonly
    <a href="http://link.aps.org/doi/10.1103/PhysRevC.66.024326">
    S. Goriely, M. Samyn, P.-H. Heenen, J. M. Pearson, and F. Tondeur</a>,
    \endhtmlonly
    \latexonly
    \href{http://link.aps.org/doi/10.1103/PhysRevC.66.024326}{
    S. Goriely, M. Samyn, P.-H. Heenen, J. M. Pearson, and F. Tondeur},
    \endlatexonly
    Phys. Rev. C \b 66 (2002) 024326. 

    \anchor Goriely07 Goriely07:
    \htmlonly
    <a href="http://link.aps.org/doi/10.1103/PhysRevC.75.064312">
    S. Goriely, M. Samyn, and J. M. Pearson</a>,
    \endhtmlonly
    \latexonly
    \href{http://link.aps.org/doi/10.1103/PhysRevC.75.064312}{
    S. Goriely, M. Samyn, and J. M. Pearson},
    \endlatexonly
    Phys. Rev. C \b 75 (2007) 064312.

    \anchor Johns96 Johns96:
    \htmlonly
    <a href="http://dx.doi.org/10.1086/178212">
    Johns, P.J. Ellis, and J.M. Lattimer</a>, 
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1086/178212}{
    Johns, P.J. Ellis, and J.M. Lattimer}, 
    \endlatexonly
    Astrophys. J. \b 473 (1996) 1020.

    \anchor Landau Landau:
    L. D. Landau, "Statistical Physics: Part 1", 3rd Edition.

    \anchor Moller95 Moller95:
    \htmlonly
    <a href="http://dx.doi.org/10.1006/adnd.1995.1002">
    P. Moller, J.R. Nix, W.D. Myers, and W.J. Switecki</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1006/adnd.1995.1002}{
    P. Moller, J.R. Nix, W.D. Myers, and W.J. Switecki},
    \endlatexonly
    At. Data Nucl. Data Tables \b 59 (1995) 185.

    \anchor Samyn04 Samyn04:
    \htmlonly
    <a href="http://link.aps.org/doi/10.1103/PhysRevC.70.044309">
    M. Samyn, S. Goriely, M. Bender and J. M. Pearson</a>,
    \endhtmlonly
    \latexonly
    \href{http://link.aps.org/doi/10.1103/PhysRevC.70.044309}{
    M. Samyn, S. Goriely, M. Bender and J. M. Pearson},
    \endlatexonly
    Phys. Rev. C \b 70 (2004) 044309.

*/
