/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/* Example: ex_stiff.cpp
   -------------------------------------------------------------------
   An example to demonstrate solving stiff differential equations
*/

#include <o2scl/test_mgr.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/funct.h>
#include <o2scl/ode_funct.h>
#include <o2scl/gsl_astep.h>
#include <o2scl/gsl_bsimp.h>

using namespace std;
using namespace o2scl;

int derivs(double x, size_t nv, const ovector_base &y, 
	   ovector_base &dydx, int &pa) {
  dydx[0]=480.0*y[0]+980.0*y[1];
  dydx[1]=-490.0*y[0]-990.0*y[1];
  return 0;
}

int jac(double x, size_t nv, const ovector_base &y, 
	omatrix_base &dfdy, ovector_base &dfdx, int &pa) {
  dfdy[0][0]=480.0;
  dfdy[0][1]=980.0;
  dfdy[1][0]=-490.0;
  dfdy[1][1]=-990.0;
  dfdx[0]=0.0;
  dfdx[1]=0.0;
  return 0;
}

int main(void) {
  test_mgr t;
  t.set_output_level(1);

  cout.setf(ios::scientific);
  cout.precision(3);
  
  int vp=0;

  // Specification of the differential equations and the Jacobian
  ode_funct_fptr<int,ovector_base> od(derivs);
  ode_jac_funct_fptr<int,ovector_base> oj(jac);
  
  // ------------------------------------------------------------
  // First solve with gsl_bsimp, designed to handle stiff ODEs

  gsl_bsimp<int,ode_funct_fptr<int>,ode_jac_funct_fptr<int> > gb;

  double x1, dx=1.0e-1;
  ovector y1(2), dydx1(2), yout1(2), yerr1(2), dydx_out1(2);

  x1=0.0;
  y1[0]=1.0;
  y1[1]=0.0;

  derivs(x1,2,y1,dydx1,vp);

  for(size_t i=1;i<=40;i++) {
    
    gb.step(x1,dx,2,y1,dydx1,y1,yerr1,dydx1,vp,od,oj);
    x1+=dx;
    
    cout.setf(ios::showpos);
    cout << x1 << " " << y1 << " " << yerr1 << " " 
	 << -exp(-500.0*x1)+2.0*exp(-10.0*x1) << " "
	 << exp(-500.0*x1)-exp(-10.0*x1) << endl;
    cout.unsetf(ios::showpos);

    t.test_rel(y1[0],-exp(-500.0*x1)+2.0*exp(-10.0*x1),3.0e-4,"y0");
    t.test_rel(y1[1],exp(-500.0*x1)-exp(-10.0*x1),3.0e-4,"y1");

  }

  cout << endl;

  // ------------------------------------------------------------
  // Now compare to the traditional adaptive stepper

  gsl_astep<int,ode_funct_fptr<int> > ga;

  double x2;
  ovector y2(2), dydx2(2), yout2(2), yerr2(2), dydx_out2(2);

  x2=0.0;
  y2[0]=1.0;
  y2[1]=0.0;

  derivs(x2,2,y2,dydx2,vp);

  size_t j=0;
  while (x2<4.0) {

    ga.astep(x2,dx,4.0,2,y2,dydx2,yerr2,vp,od);

    if (j%25==0) {
      cout.setf(ios::showpos);
      cout << x2 << " " << y2 << " " << yerr2 << " " 
	   << -exp(-500.0*x2)+2.0*exp(-10.0*x2) << " "
	   << exp(-500.0*x2)-exp(-10.0*x2) << endl;
      cout.unsetf(ios::showpos);
    }
    j++;

  }

  cout << endl;

  t.report();

  return 0;

}
// End of example
