/*
  -------------------------------------------------------------------

  Copyright (C) 2006, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  -------------------------------------------------------------------
*/
#ifndef O2SCL_FORMAT_FLOAT_H
#define O2SCL_FORMAT_FLOAT_H

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <o2scl/err_hnd.h>
#include <o2scl/file_format.h>
#include <o2scl/misc.h>
#include <o2scl/string_conv.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Format a floating point number into a Latex or HTML string
      
      This class formats floating point strings into something 
      useful for HTML or Latex documents.

      The base-10 logarithm of the smallest and largest 
      numbers to be represented without a string of the form
      \verbatim
      $\times 10^{\mathrm{x}}$
      \endverbatim
      can be specified in set_exp_min() and set_exp_max(). The number
      of significant figures can be specified with set_sig_figs() (the
      default is 5). 

      \comment
      To force convert() to adds zeros to the right side of the
      mantissa to guarantee that the requested number of significant
      digits is given, use <tt>set_pad_zeros(true)</tt>.
      \endcomment

      Note that this function does not warn the user if the number
      of significant figures requested is larger than the machine 
      precision.

      The format for a normal number is
      \verbatim
      prefix (sign-string) number suffix
      \endverbatim
      and in scientific notation is
      \verbatim 
      sci-prefix (sci-sign-string) number times-string 
      exp-prefix (exp-sign-string) exponent exp-suffix sci-suffix
      \endverbatim

      \future Implement padding with zeros
      \future Separate out decimal point and make it separate.
      
  */
  class format_float {
    
  protected:

    /// \name Base text settings
    //@{
    /// Prefix (default "")
    std::string prefx;
    /// Suffix (default "")
    std::string suffx;
    /// Sign string (default "-")
    std::string sgn;
    /// Sign string in scientific mode (default "-")
    std::string sci_sgn;
    /// Sign string for exponent in scientific mode (default "-")
    std::string exp_sgn;
    /// Prefix in scientific mode (default "")
    std::string sci_prefx;
    /// Suffix in scientific mode (default "")
    std::string sci_suffx;
    /// Exponent prefix (default "")
    std::string exp_prefx;
    /// Exponent suffix (default "")
    std::string exp_suffx;
    /// Times symbol for scientific mode (default " x ")
    std::string tmes;
    /// String for numbers which are not finite (default "Nan")
    std::string not_finte;
    /// String for zeros (default "0")
    std::string zeros;
    //@}
    
    /// \name Other settings
    //@{
    /// Number of significant figures (default 5)
    size_t sig_fgs;
    /// Lower limit for automatic mode (default -2)
    int ex_mn;
    /// Upper limit for automatic mode (default -2)
    int ex_mx;
    /// If true, pad with zeros (default false)
    bool pad_zeros;
    //@}

  public:

    format_float() {
      prefx="";
      sgn="-";
      suffx="";
      sci_prefx="";
      tmes=" x ";
      exp_prefx="^{";
      exp_sgn="-";
      sci_sgn="-";
      exp_suffx="}";
      sci_suffx="";
      not_finte="Nan";
      zeros="0";
      ex_mn=-2;
      ex_mx=3;
      sig_fgs=5;
      pad_zeros=false;
    }

    /** \brief Set HTML mode
	
	This function is equivalent to the settings:
	\code
	format_float::set_prefix("");
	format_float::set_sign("-");
	format_float::set_suffix("");
	format_float::set_sci_prefix("");
	format_float::set_times(" &times; ");
	format_float::set_exp_prefix("10<sup>");
	format_float::set_exp_sign("-");
	format_float::set_sci_sign("-");
	format_float::set_exp_suffix("</sup>");
	format_float::set_sci_suffix("");
	format_float::set_not_finite("Nan");
	format_float::set_zero("0");
	\endcode
    */
    int html_mode() {
      prefx="";
      sgn="-";
      suffx="";
      sci_prefx="";
      tmes=" &times; ";
      exp_prefx="10<sup>";
      exp_sgn="-";
      sci_sgn="-";
      exp_suffx="</sup>";
      sci_suffx="";
      not_finte="Nan";
      zeros="0";
      return 0;
    }

    /** \brief Set Latex mode

	This function is equivalent to the settings:
	\code
	format_float::set_prefix("");
	format_float::set_sign("$-$");
	format_float::set_suffix("");
	format_float::set_sci_prefix("");
	format_float::set_times(" $\\times ");
	format_float::set_exp_prefix("10^{");
	format_float::set_exp_sign("-");
	format_float::set_sci_sign("$-$");
	format_float::set_exp_suffix("}");
	format_float::set_sci_suffix("");
	format_float::set_not_finite("Nan");
	format_float::set_zero("0");
	\endcode
    */
    int latex_mode() {
      prefx="";
      sgn="$-$";
      suffx="";
      sci_prefx="";
      tmes=" $\\times ";
      exp_prefx="10^{";
      sci_sgn="$-$";
      exp_sgn="-";
      exp_suffx="}$";
      sci_suffx="";
      not_finte="Nan";
      zeros="0";
      return 0;
    }      
    
    /** \name Set text settings

	These are modified by the functions html_mode() and latex_mode()
    */
    //@{
    /// set prefix
    int set_prefix(std::string prefix) {
      prefx=prefix;
      return 0;
    }

    /// Set suffix
    int set_suffix(std::string suffix) {
      suffx=suffix;
      return 0;
    }

    /// Set prefix for scientific notation
    int set_sci_prefix(std::string sci_prefix) {
      sci_prefix=sci_prefx;
      return 0;
    }

    /// Set suffix for scientific notation
    int set_sci_suffix(std::string sci_suffix) {
      sci_suffx=sci_suffix;
      return 0;
    }

    /// Set prefix for exponent
    int set_exp_prefix(std::string exp_prefix) {
      exp_prefx=exp_prefix;
      return 0;
    }

    /// Set suffix for exponent
    int set_exp_suffix(std::string exp_suffix) {
      exp_suffx=exp_suffix;
      return 0;
    }

    /// Set sign
    int set_sign(std::string sign) {
      sgn=sign;
      return 0;
    }

    /// Set sign for exponent
    int set_exp_sign(std::string exp_sign) {
      exp_sgn=exp_sign;
      return 0;
    }

    /// Set sign for scientific notation
    int set_sci_sign(std::string sci_sign) {
      sci_sgn=sci_sign;
      return 0;
    }

    /// Set times
    int set_times(std::string times) {
      tmes=times;
      return 0;
    }

    /// Set zero
    int set_zero(std::string zero) {
      zeros=zero;
      return 0;
    }

    /// Set string for numbers which are not finite
    int set_not_finite(std::string not_finite) {
      not_finte=not_finite;
      return 0;
    }
    //@}

    /** \name Set other settings

	These are not modified by the functions html_mode() and latex_mode()
    */
    //@{
    /// Set the exponent limits
    int set_exp_limits(int min, int max) {
      ex_mn=min;
      ex_mx=max;
      return 0;
    }

    /// Set the number of significant figures
    int set_sig_figs(size_t sig_figs) {
      sig_fgs=sig_figs;
      return 0;
    }

#ifdef O2SCL_NEVER_DEFINED
    /// Set pad zeros
    int set_pad_zeros(bool pad) {
      pad_zeros=pad;
      return 0;
    }
#endif
    //@}

    /** \name Get text settings

	These are modified by the functions html_mode() and latex_mode()
    */
    //@{
    /// Get prefix
    std::string get_prefix() {
      return prefx;
    }
    
    /// Get suffix
    std::string get_suffix() {
      return suffx;
    }
    
    /// Get prefix for scientific notation
    std::string get_sci_prefix() {
      return sci_prefx;
    }
    
    /// Get suffix for scientific notation
    std::string get_sci_suffix() {
      return sci_suffx;
    }
    
    /// Get prefix for exponent
    std::string get_exp_prefix() {
      return exp_prefx;
    }
    
    /// Get suffix for exponent
    std::string get_exp_suffix() {
      return exp_suffx;
    }
    
    /// Get sign
    std::string get_sign() {
      return sgn;
    }
    
    /// Get sign for exponent
    std::string get_exp_sign() {
      return exp_sgn;
    }
    
    /// Get sign for scientific notation
    std::string get_sci_sign() {
      return sci_sgn;
    }
    
    /// Get times
    std::string get_times() {
      return tmes;
    }
    
    /// Get zero
    std::string get_zero() {
      return zeros;
    }
    
    /// Get string for numbers which are not finite
    std::string get_not_finite() {
      return not_finte;
    }
    //@}

    /** \name Get other settings

	These are not modified by the functions html_mode() and latex_mode()
    */
    //@{
    /// Get minimum exponent
    int get_exp_min() {
      return ex_mn;
    }

    /// Get maximum exponent
    int get_exp_max() {
      return ex_mx;
    }
    
    /// Get sig_figs
    size_t get_sig_figs() {
      return sig_fgs;
    }
    
#ifdef O2SCL_NEVER_DEFINED
    /// Get pad_zeros
    bool get_pad_zeros() {
      return pad_zeros;
    }
#endif
    //@}

    /// Convert a floating point number to a string
    std::string convert(double x) {

      // Handle special cases
      if (!finite(x)) return not_finte;
      if (x==0.0) return zeros;

      // Check sign
      int sign=1;
      if (x<0) {
	x=-x; 
	sign=-1;
      }

      // Compute exponent and mantissa separately
      int expo=((int)log10(x)), itmp;
      double mant=x/pow(10.0,expo);
      if (mant<1.0) {
	mant*=10.0;
	expo-=1;
      }
      itmp=(int)(mant*pow(10.0,sig_fgs)+0.5);
      mant=((double)itmp)/pow(10.0,sig_fgs);

      // Begin constructing the string for output
      std::string s;

      if (expo<=ex_mx && expo>=ex_mn) {

	// Normal notation
	
	std::string nums=o2scl::dtos(mant*pow(10.0,expo),sig_fgs,true);
	if (pad_zeros) {
	  while (nums.length()<sig_fgs+1) nums+='0';
	}
	
	if (sign==-1) {
	  s=prefx+sgn+nums+suffx;
	} else {
	  s=prefx+nums+suffx;
	}

      } else {

	// Scientific notation

	std::string nums=o2scl::dtos(mant,sig_fgs,true);
	if (pad_zeros) {
	  while (nums.length()<sig_fgs+1) nums+='0';
	}

	// Compute sign of exponent
	int expo_sign=1;
	if (expo<0) {
	  expo_sign=-1;
	  expo*=-1;
	}
	
	if (sign==-1) {
	  if (expo_sign<0) {
	    s=sci_prefx+sci_sgn+nums+tmes+exp_prefx+
	      exp_sgn+o2scl::itos(expo)+exp_suffx+sci_suffx;
	  } else {
	    s=sci_prefx+sci_sgn+nums+tmes+exp_prefx+
	      o2scl::itos(expo)+exp_suffx+sci_suffx;
	  }
	} else {
	  if (expo_sign<0) {
	    s=sci_prefx+nums+tmes+exp_prefx+
	      exp_sgn+o2scl::itos(expo)+exp_suffx+sci_suffx;
	  } else {
	    s=sci_prefx+nums+tmes+exp_prefx+
	      o2scl::itos(expo)+exp_suffx+sci_suffx;
	  }
	}
	
      }
      
      return s;

    }
    
    
  };

#ifndef DOXYGENP
}
#endif

#endif
