/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_INTERP_H
#define O2SCL_INTERP_H

#include <iostream>
#include <string>

// for gsl_linalg_solve_symm_tridiag for cspline_interp
#include <gsl/gsl_linalg.h>

#include <o2scl/err_hnd.h>
#include <o2scl/collection.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/ovector_rev_tlate.h>
#include <o2scl/search_vec.h>
#include <o2scl/tridiag.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Base low-level interpolation class [abstract base]

      The descendants of this class are intended to be fast
      interpolation routines for increasing functions, leaving the
      some error handling, user-friendliness, and other more 
      complicated improvements for other classes.

      For any pair of vectors x and y into which you would like to
      interpolate, you need to call allocate() and init() first, and then 
      the interpolation functions, and then free(). If the next pair
      of vectors has the same size, then you need only to call init()
      before the next call to an interpolation function. If the vectors
      do not change, then you may call the interpolation functions in
      succession.

      All of the descendants are based on the GSL interpolation 
      routines and give identical results.

      \future These might work for decreasing functions by just
      replacing calls to vector_bsearch_inc() with
      vector_bsearch_dec(). If this is the case, then this should
      be rewritten accordingly. (I think I might have removed the
      acceleration)
  */
  template<class vec_t> class base_interp {

#ifndef DOXYGEN_INTERNAL

  protected:

    /** 
	\brief An internal function to assist in computing the 
	integral for both the cspline and Akima types
    */
    double integ_eval(double ai, double bi, double ci, double di, double xi, 
		      double a, double b) {
      const double r1 = a - xi;
      const double r2 = b - xi;
      const double r12 = r1 + r2;
      const double bterm = 0.5 * bi * r12;
      const double cterm = (1.0 / 3.0) * ci * (r1 * r1 + r2 * r2 + r1 * r2);
      const double dterm = 0.25 * di * r12 * (r1 * r1 + r2 * r2);
      
      return (b - a) * (ai + bterm + cterm + dterm);
    }

#endif

  public:

    // this empty constructor is required since the 
    // default copy constructor is private
    base_interp() {}

    virtual ~base_interp() {}

    /** 
	\brief The minimum size of the vectors to interpolate between

	This needs to be set in the constructor of the children for
	access by the class user
    */
    size_t min_size;

    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) { return 0; }

    /// Free allocated memory
    virtual int free() { return 0; }

    /// Initialize interpolation routine
    virtual int init(const vec_t &x, const vec_t &y, size_t size) {
      return 0;
    }
    
    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &y0)=0;
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x, const vec_t &y, size_t size,
		      double x0, double &dydx)=0;
    
    /** \brief Give the value of the second derivative 
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &d2ydx2)=0;
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x, const vec_t &y, size_t size, 
		      double a, double b, double &result)=0;

    /// Return the type, \c "base_interp".
    virtual const char *type() { return "base_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    base_interp<vec_t>(const base_interp<vec_t> &);
    base_interp<vec_t>& operator=(const base_interp<vec_t>&);

#endif

  };
  
  /** 
      \brief Linear interpolation (GSL)

      Linear interpolation requires no calls to allocate(), free() or
      init(), as there is no internal storage required.
  */
#ifdef DOYXGENP
  template<class vec_t> class linear_interp : public base_interp
#else 
    template<class vec_t> class linear_interp : public base_interp<vec_t> 
#endif
  {
    
  public:
    
    linear_interp() {
      this->min_size=2;
    }

    virtual ~linear_interp() {}

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      double x_lo, x_hi;
      double y_lo, y_hi;
      double dx;
      size_t index;

      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
      
      /* evaluate */
      x_lo = x_array[index];
      x_hi = x_array[index + 1];
      y_lo = y_array[index];
      y_hi = y_array[index + 1];
      dx = x_hi - x_lo;
      if (dx > 0.0) {
	y = y_lo + (x - x_lo) / dx * (y_hi - y_lo);
	return gsl_success;
      } else {
	y = 0.0;
	O2SCL_ERR_RET("Function not increasing in linear_interp::interp().",
		      gsl_einval);
      }
    }
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {
      
      double x_lo, x_hi;
      double y_lo, y_hi;
      double dx;
      double dy;
      size_t index;
      
      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
      
      /* evaluate */
      x_lo = x_array[index];
      x_hi = x_array[index + 1];
      y_lo = y_array[index];
      y_hi = y_array[index + 1];
      dx = x_hi - x_lo;
      dy = y_hi - y_lo;
      if (dx > 0.0) {
	dydx = dy / dx;
	return gsl_success;
      } else {
	dydx = 0.0;
	O2SCL_ERR_RET("Function not increasing in linear_interp::deriv().",
		      gsl_einval);
      }
    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &d2ydx2) {
      d2ydx2=0.0;
      return gsl_success;
    }


    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double a, double b, 
		      double &result) {
      size_t i, index_a, index_b;
      
      index_a=vector_bsearch_inc<vec_t,double>(a,x_array,0,size-1);
      index_b=vector_bsearch_inc<vec_t,double>(b,x_array,0,size-1);
      
      /* endpoints span more than one interval */
      
      result = 0.0;
      
      /* interior intervals */
      for(i=index_a; i<=index_b; i++) {
	const double x_hi = x_array[i + 1];
	const double x_lo = x_array[i];
	const double y_lo = y_array[i];
	const double y_hi = y_array[i + 1];
	const double dx = x_hi - x_lo;
	
	if(dx != 0.0) {
	  if (i == index_a || i == index_b) {
	    double x1 = (i == index_a) ? a : x_lo;
	    double x2 = (i == index_b) ? b : x_hi;
	    const double D = (y_hi-y_lo)/dx;
	    result += (x2-x1) * (y_lo + 0.5*D*((x2-x_lo)+(x1-x_lo)));
	  } else  {
	    result += 0.5 * dx * (y_lo + y_hi);
	  }
	}
      }
      
      return gsl_success;
    }

    /// Return the type, \c "linear_interp".
    virtual const char *type() { return "linear_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    linear_interp<vec_t>(const linear_interp<vec_t> &);
    linear_interp<vec_t>& operator=(const linear_interp<vec_t>&);

#endif

  };
  
  /** 
      \brief Cubic spline interpolation (GSL)

      By default, this uses natural boundary conditions, where
      the second derivative vanishes at each end point. Extrapolation
      effectively assumes that the second derivative is linear outside
      of the endpoints. If the boolean argument to the constructor
      is set to \c true, then periodic boundary conditions are assumed. 
      Alternatively, use \ref cspline_peri_interp for perodic
      bounary conditions.

      \future Could use \o2 's native tridiagonal routines instead 
      of calling the GSL ones.
  */
#ifdef DOYXGENP
  template<class vec_t> class cspline_interp : public base_interp
#else 
    template<class vec_t> class cspline_interp : public base_interp<vec_t> 
#endif
  {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// \name Storage for cubic spline interpolation
    //@{
    double *c;
    double *g;
    double *diag;
    double *offdiag;
    //@}

    /// True for periodic boundary conditions
    bool peri;

    /// Compute coefficients for cubic spline interpolation
    void coeff_calc(const double c_array[], double dy, double dx, 
		    size_t index, double *b, double *c2, double *d) {

      const double c_i = c_array[index];
      const double c_ip1 = c_array[index + 1];
      *b = (dy / dx) - dx * (c_ip1 + 2.0 * c_i) / 3.0;
      *c2 = c_i;
      *d = (c_ip1 - c_i) / (3.0 * dx);
    }

#endif
    
  public:

    /** \brief Create a base interpolation object with natural or
	periodic boundary conditions
    */
    cspline_interp(bool periodic=false) {
      
      peri=periodic;
      if (peri) this->min_size=2;
      else this->min_size=3;
      
    }

    virtual ~cspline_interp() {}
    
    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) {
      
      c = (double *) malloc (size * sizeof (double));
      if (c == 0) {
	O2SCL_ERR2_RET("Failed to allocate space for c",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      g = (double *) malloc (size * sizeof (double));
      if (g == 0) {
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for g",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      diag = (double *) malloc (size * sizeof (double));
      if (diag == 0) {
	std::free(g);
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for diag",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      offdiag = (double *) malloc (size * sizeof (double));
      if (offdiag == 0) {
	std::free(diag);
	std::free(g);
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for offdiag",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      return 0;
    }
    
    /** \brief Initialize interpolation routine
    */
    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {

      if (peri) {

	/// Periodic boundary conditions
	 
	size_t i;
	size_t num_points = size;
	/* Engeln-Mullges + Uhlig "n" */
	size_t max_index = num_points - 1;  
	/* linear system is sys_size x sys_size */
	size_t sys_size = max_index;    
	
	if (sys_size == 2) {
	  /* solve 2x2 system */
	  
	  const double h0 = xa[1] - xa[0];
	  const double h1 = xa[2] - xa[1];
	  const double h2 = xa[3] - xa[2];
	  const double A = 2.0*(h0 + h1);
	  const double B = h0 + h1;
	  double gx[2];
	  double det;
	  
	  gx[0] = 3.0 * ((ya[2] - ya[1]) / h1 - (ya[1] - ya[0]) / h0);
	  gx[1] = 3.0 * ((ya[1] - ya[2]) / h2 - (ya[2] - ya[1]) / h1);
	  
	  det = 3.0 * (h0 + h1) * (h0 + h1);
	  c[1] = ( A * gx[0] - B * gx[1])/det;
	  c[2] = (-B * gx[0] + A * gx[1])/det;
	  c[0] =  c[2];
	  
	  return gsl_success;

	} else {
	  
	  for (i = 0; i < sys_size-1; i++) {
	    const double h_i       = xa[i + 1] - xa[i];
	    const double h_ip1     = xa[i + 2] - xa[i + 1];
	    const double ydiff_i   = ya[i + 1] - ya[i];
	    const double ydiff_ip1 = ya[i + 2] - ya[i + 1];
	    const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
	    const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
	    offdiag[i] = h_ip1;
	    diag[i] = 2.0 * (h_ip1 + h_i);
	    g[i] = 3.0 * (ydiff_ip1 * g_ip1 - ydiff_i * g_i);
	  }
	  
	  i = sys_size - 1;
	  {
	    const double h_i       = xa[i + 1] - xa[i];
	    const double h_ip1     = xa[1] - xa[0];
	    const double ydiff_i   = ya[i + 1] - ya[i];
	    const double ydiff_ip1 = ya[1] - ya[0];
	    const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
	    const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
	    offdiag[i] = h_ip1;
	    diag[i] = 2.0 * (h_ip1 + h_i);
	    g[i] = 3.0 * (ydiff_ip1 * g_ip1 - ydiff_i * g_i);
	  }

	  if (false) {

	    double *cp1=c+1;
	    int status=o2scl_linalg::solve_cyc_tridiag_sym<double *>
	      (diag,offdiag,g,cp1,sys_size);
	    c[0]=c[max_index];
	    return status;

	  } else {

	    gsl_vector_view g_vec = gsl_vector_view_array( g, sys_size);
	    gsl_vector_view diag_vec =
	      gsl_vector_view_array( diag,sys_size);
	    gsl_vector_view offdiag_vec =
	      gsl_vector_view_array( offdiag, sys_size);
	    gsl_vector_view solution_vec =
	      gsl_vector_view_array (( c) + 1, sys_size);
	    int status =
	      gsl_linalg_solve_symm_cyc_tridiag(&diag_vec.vector,
						&offdiag_vec.vector,
						&g_vec.vector,
						&solution_vec.vector);
	    c[0] =  c[max_index];
	    
	    return status;
	  }
	  
	}

      } else {

	/// Natural boundary conditions

	size_t i;
	size_t num_points = size;
	size_t max_index = num_points - 1;
	size_t sys_size = max_index - 1;

	c[0] = 0.0;
	c[max_index] = 0.0;

	for (i = 0; i < sys_size; i++)
	  {
	    const double h_i   = xa[i + 1] - xa[i];
	    const double h_ip1 = xa[i + 2] - xa[i + 1];
	    const double ydiff_i   = ya[i + 1] - ya[i];
	    const double ydiff_ip1 = ya[i + 2] - ya[i + 1];
	    const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
	    const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
	    offdiag[i] = h_ip1;
	    diag[i] = 2.0 * (h_ip1 + h_i);
	    g[i] = 3.0 * (ydiff_ip1 * g_ip1 -  ydiff_i * g_i);
	  }

	if (sys_size == 1) {

	  c[1] =  g[0] /  diag[0];
	  return gsl_success;

	} else {

	  if (false) {

	    double *cp1=c+1;
	    int status=o2scl_linalg::solve_cyc_tridiag_sym<double *>
	      (diag,offdiag,g,cp1,sys_size);
	    return status;

	  } else {

	    gsl_vector_view g_vec = gsl_vector_view_array( g, sys_size);
	    gsl_vector_view diag_vec = gsl_vector_view_array( diag, sys_size);
	    gsl_vector_view offdiag_vec = 
	      gsl_vector_view_array( offdiag, sys_size - 1);
	    gsl_vector_view solution_vec = 
	      gsl_vector_view_array (( c) + 1, sys_size);
	    
	    int status = gsl_linalg_solve_symm_tridiag(&diag_vec.vector, 
						       &offdiag_vec.vector, 
						       &g_vec.vector, 
						       &solution_vec.vector);
	    return status;
	  }
	}
      }
    }
      
    
    /// Free allocated memory
    virtual int free() {
      std::free(c);
      std::free(g);
      std::free(diag);
      std::free(offdiag);

      return 0;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      
      double x_lo, x_hi;
      double dx;
      size_t index;
      
      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
  
      /* evaluate */
      x_hi = x_array[index + 1];
      x_lo = x_array[index];
      dx = x_hi - x_lo;
      if (dx > 0.0) {
	const double y_lo = y_array[index];
	const double y_hi = y_array[index + 1];
	const double dy = y_hi - y_lo;
	double delx = x - x_lo;
	double b_i, c_i, d_i; 
	coeff_calc( c, dy, dx, index, &b_i, &c_i, &d_i);
	y = y_lo + delx * (b_i + delx * (c_i + delx * d_i));
	return gsl_success;
      } else {
	y = 0.0;
	O2SCL_ERR_RET("Function not increasing in cspline_interp::interp().",
		      gsl_einval);
      }
    }


    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {

      double x_lo, x_hi;
      double dx;
      size_t index;
  
      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
  
      /* evaluate */
      x_hi = x_array[index + 1];
      x_lo = x_array[index];
      dx = x_hi - x_lo;
      if (dx > 0.0) {
	const double y_lo = y_array[index];
	const double y_hi = y_array[index + 1];
	const double dy = y_hi - y_lo;
	double delx = x - x_lo;
	double b_i, c_i, d_i; 
	coeff_calc( c, dy, dx, index, &b_i, &c_i, &d_i);
	dydx = b_i + delx * (2.0 * c_i + 3.0 * d_i * delx);
	return gsl_success;
      } else {
	dydx = 0.0;
	O2SCL_ERR_RET("Function not increasing in cspline_interp::deriv().",
		      gsl_einval);
      }
    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &d2ydx2) {

      double x_lo, x_hi;
      double dx;
      size_t index;
  
      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
  
      /* evaluate */
      x_hi = x_array[index + 1];
      x_lo = x_array[index];
      dx = x_hi - x_lo;
      if (dx > 0.0)
	{
	  const double y_lo = y_array[index];
	  const double y_hi = y_array[index + 1];
	  const double dy = y_hi - y_lo;
	  double delx = x - x_lo;
	  double b_i, c_i, d_i;
	  coeff_calc( c, dy, dx, index, &b_i, &c_i, &d_i);
	  d2ydx2 = 2.0 * c_i + 6.0 * d_i * delx;
	  return gsl_success;
	}
      else
	{
	  d2ydx2 = 0.0;
	  return GSL_FAILURE;
	}
    }

  public:
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double a, double b, double &result) {

      size_t i, index_a, index_b;
  
      index_a=vector_bsearch_inc<vec_t,double>(a,x_array,0,size-1);
      index_b=vector_bsearch_inc<vec_t,double>(b,x_array,0,size-1);

      result = 0.0;
  
      /* interior intervals */
      for(i=index_a; i<=index_b; i++) {
	const double x_hi = x_array[i + 1];
	const double x_lo = x_array[i];
	const double y_lo = y_array[i];
	const double y_hi = y_array[i + 1];
	const double dx = x_hi - x_lo;
	const double dy = y_hi - y_lo;
	if(dx != 0.0) {
	  double b_i, c_i, d_i; 
	  coeff_calc( c, dy, dx, i, &b_i, &c_i, &d_i);
      
	  if (i == index_a || i == index_b) {
	    double x1 = (i == index_a) ? a : x_lo;
	    double x2 = (i == index_b) ? b : x_hi;
	    result += this->integ_eval(y_lo, b_i, c_i, d_i, x_lo, x1, x2);
	  } else {
	    result += dx * (y_lo + dx*(0.5*b_i + 
				       dx*(c_i/3.0 + 0.25*d_i*dx)));
	  }
	} else {
	  result = 0.0;
	  O2SCL_ERR2_RET("Integral failed (dx=0) in ",
			 "cspline_interp::integ().",gsl_failure);
	}
      }
  
      return gsl_success;
    }

    /// Return the type, \c "cspline_interp".
    virtual const char *type() { return "cspline_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    cspline_interp<vec_t>(const cspline_interp<vec_t> &);
    cspline_interp<vec_t>& operator=(const cspline_interp<vec_t>&);

#endif

  };
  
  /** 
      \brief Cubic spline interpolation with periodic 
      boundary conditions (GSL)

      This is convenient to allow interpolation objects to
      be supplied as template parameters
  */
#ifdef DOYXGENP
  template<class vec_t> class cspline_peri_interp : public cspline_interp
#else 
    template<class vec_t> class cspline_peri_interp : 
  public cspline_interp<vec_t> 
#endif
  {
    
  public:
    cspline_peri_interp() : cspline_interp<vec_t>(true) {
    }

      virtual ~cspline_peri_interp() {}

      /// Return the type, \c "cspline_peri_interp".
      virtual const char *type() { return "cspline_peri_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

      cspline_peri_interp<vec_t>(const cspline_peri_interp<vec_t> &);
      cspline_peri_interp<vec_t>& operator=(const cspline_peri_interp<vec_t>&);

#endif

  };

  /** 
      \brief Akima spline interpolation (GSL)

      \future It appears that the interp() function below searches
      for indices slightly differently than the original GSL
      eval() function. This should be checked, as it might be
      slightly non-optimal in terms of speed (shouldn't matter for
      the accuracy). 
  */
#ifdef DOYXGENP
  template<class vec_t> class akima_interp : public base_interp
#else 
    template<class vec_t> class akima_interp : public base_interp<vec_t> 
#endif
  {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// \name Storage for Akima spline interpolation
    //@{
    double *b;
    double *c;
    double *d;
    double *um;
    //@}

    /// True for periodic boundary conditions
    bool peri;

    /// For initializing the interpolation
    void akima_calc(const vec_t &x_array, size_t size, double m[]) {
      size_t i;
      
      for (i = 0; i < (size - 1); i++) {

	const double NE = fabs (m[i + 1] - m[i]) + 
	  fabs (m[i - 1] - m[i - 2]);
	
	if (NE == 0.0) {
	  b[i] = m[i];
	  c[i] = 0.0;
	  d[i] = 0.0;
	} else {
	  const double h_i = x_array[i + 1] - x_array[i];
	  const double NE_next = fabs (m[i + 2] - m[i + 1]) + 
	    fabs (m[i] - m[i - 1]);
	  const double alpha_i = fabs (m[i - 1] - m[i - 2]) / NE;
	  double alpha_ip1;
	  double tL_ip1;
	  if (NE_next == 0.0) {
	    tL_ip1 = m[i];
	  } else {
	    alpha_ip1 = fabs (m[i] - m[i - 1]) / NE_next;
	    tL_ip1 = (1.0 - alpha_ip1) * m[i] + alpha_ip1 * m[i + 1];
	  }
	  b[i] = (1.0 - alpha_i) * m[i - 1] + alpha_i * m[i];
	  c[i] = (3.0 * m[i] - 2.0 * b[i] - tL_ip1) / h_i;
	  d[i] = (b[i] + tL_ip1 - 2.0 * m[i]) / (h_i * h_i);
	}
      }
    }
    
#endif
    
  public:

    /** \brief Create a base interpolation object with or without
	periodic boundary conditions
    */
    akima_interp(bool periodic=false) {
      peri=periodic;
      this->min_size=5;
    }

    virtual ~akima_interp() {}
    
    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) {
      
      b = (double *) malloc (size * sizeof (double));
      if (b==0) {
	O2SCL_ERR2_RET("Failed to allocate space for c",
		       " in akima_interp::allocate().",gsl_enomem);
      }
      
      c = (double *) malloc (size * sizeof (double));
      if (c==0) {
	std::free(b);
	O2SCL_ERR2_RET("Failed to allocate space for g",
		       " in akima_interp::allocate().",gsl_enomem);
      }
      
      d = (double *) malloc (size * sizeof (double));
      if (d==0) {
	std::free(b);
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for diag",
		       " in akima_interp::allocate().",gsl_enomem);
      }
      
      um = (double *) malloc ((size+4) * sizeof (double));
      if (um==0) {
	std::free(b);
	std::free(c);
	std::free(d);
	O2SCL_ERR2_RET("Failed to allocate space for offdiag",
		       " in akima_interp::allocate().",gsl_enomem);
      }
      
      return 0;
    }
    
    /// Initialize interpolation routine
    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {
      
      if (peri) {
	
	/// Periodic boundary conditions

	double *m=um+2;
	size_t i;
	for (i = 0; i <= size - 2; i++)
	  {
	    m[i] = (ya[i + 1] - ya[i]) / (xa[i + 1] - xa[i]);
	  }
	
	/* periodic boundary conditions */
	m[-2] = m[size - 1 - 2];
	m[-1] = m[size - 1 - 1];
	m[size - 1] = m[0];
	m[size] = m[1];
	
	akima_calc (xa,size, m);
	
      } else {

	/// Non-periodic boundary conditions

	double *m=um+2;
	size_t i;
	for (i = 0; i <= size - 2; i++) {
	  m[i] = (ya[i + 1] - ya[i]) / (xa[i + 1] - xa[i]);
	}
	
	/* non-periodic boundary conditions */
	m[-2] = 3.0 * m[0] - 2.0 * m[1];
	m[-1] = 2.0 * m[0] - m[1];
	m[size - 1] = 2.0 * m[size - 2] - m[size - 3];
	m[size] = 3.0 * m[size - 2] - 2.0 * m[size - 3];
	
	akima_calc (xa, size, m);

      }
      return 0;
    }
      
    
    /// Free allocated memory
    virtual int free() {
      std::free(b);
      std::free(c);
      std::free(d);
      std::free(um);

      return 0;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      
      size_t index;
      
      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
  
      const double x_lo = x_array[index];
      const double delx = x - x_lo;
      const double bb = b[index];
      const double cc = c[index];
      const double dd = d[index];
      y = y_array[index] + delx * (bb + delx * (cc + dd * delx));
      return gsl_success;
    }


    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {

      size_t index;
  
      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
      double x_lo = x_array[index];
      double delx = x - x_lo;
      double bb = b[index];
      double cc = c[index];
      double dd = d[index];
      dydx = bb + delx * (2.0 * cc + 3.0 * dd * delx);
      return gsl_success;

    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &d2ydx2) {

      size_t index;
  
      index=vector_bsearch_inc<vec_t,double>(x,x_array,0,size-1);
  
      const double x_lo = x_array[index];
      const double delx = x - x_lo;
      const double cc = c[index];
      const double dd = d[index];
      d2ydx2 = 2.0 * cc + 6.0 * dd * delx;
      return gsl_success;

    }

    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double aa, double bb, 
		      double &result) {

      size_t i, index_a, index_b;
  
      index_a=vector_bsearch_inc<vec_t,double>(aa,x_array,0,size-1);
      index_b=vector_bsearch_inc<vec_t,double>(bb,x_array,0,size-1);

      result = 0.0;
  
      /* interior intervals */

      for(i=index_a; i<=index_b; i++) {

	const double x_hi = x_array[i + 1];
	const double x_lo = x_array[i];
	const double y_lo = y_array[i];
	const double dx = x_hi - x_lo;

	if (dx != 0.0) {
	  
	  if (i==index_a || i==index_b) {
	    double x1 = (i==index_a) ? aa : x_lo;
	    double x2 = (i==index_b) ? bb : x_hi;
	    result += this->integ_eval(y_lo,b[i],c[i],d[i],x_lo,x1,x2);
	  } else {
	    result+=dx*(y_lo+dx*(0.5*b[i]+dx*(c[i]/3.0+0.25*d[i]*dx)));
	  }

	} else {
	  result = 0.0;
	  O2SCL_ERR2_RET("Integral failed (dx=0) in ",
			 "akima_interp::integ().",gsl_failure);
	}
      }
  
      return gsl_success;
    }

    /// Return the type, \c "akima_interp".
    virtual const char *type() { return "akima_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    akima_interp<vec_t>(const akima_interp<vec_t> &);
    akima_interp<vec_t>& operator=(const akima_interp<vec_t>&);

#endif

  };
  
  /** 
      \brief Akima spline interpolation with periodic 
      boundary conditions (GSL)
      
      This is convenient to allow interpolation objects to
      be supplied as template parameters
  */
#ifdef DOYXGENP
  template<class vec_t> class akima_peri_interp : public akima_interp
#else 
    template<class vec_t> class akima_peri_interp : 
  public akima_interp<vec_t> 
#endif
  {

  public:

    akima_peri_interp() : akima_interp<vec_t>(true) {
    }
      
      virtual ~akima_peri_interp() {}
    
      /// Return the type, \c "akima_peri_interp".
      virtual const char *type() { return "akima_peri_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

      akima_peri_interp<vec_t>(const akima_peri_interp<vec_t> &);
      akima_peri_interp<vec_t>& operator=(const akima_peri_interp<vec_t>&);

#endif

  };

  /** 
      \brief A base interpolation object manager [abstract base]
  */
  template<class vec_t> class base_interp_mgr {

  public:

    virtual ~base_interp_mgr() {}

    /// Create a new interpolation object
    virtual base_interp<vec_t> *new_interp()=0;

    /// Deallocate an interpolation object
    virtual int free_interp(base_interp<vec_t> *b) {
      delete b;
      return 0;
    }
  };
  
  /** 
      \brief A base interpolation object manager template 
  */
  template<class vec_t, template<class> class interp_t > class def_interp_mgr :
  public base_interp_mgr<vec_t> {
    
  public:

    virtual ~def_interp_mgr() {}

    /// Create a new interpolation object
    virtual base_interp<vec_t> *new_interp() {
      interp_t<vec_t> *b=new interp_t<vec_t>;
      return b;
    }

  };

  /** 
      \brief Interpolation class
      
      This interpolation class is intended to be sufficiently general
      to handle any vector type. Interpolation of \ref ovector like
      objects is performed with the default template parameters, and
      \ref array_interp is provided for simple interpolation on
      C-style arrays.

      The type of interpolation to be performed can be specified using
      the set_type() function or in the constructor.  The default is
      cubic splines with natural boundary conditions.

      The class automatically handles decreasing arrays by converting
      from an object of type \c vec_t to an object of type \c rvec_t.

      While \c vec_t may be any vector type which allows indexing via
      \c [], \c rvec_t must be a vector type which allows indexing
      and has a constructor with one of the two forms
      \verbatim
      rvec_t::rvec_t(vec_t &v);
      rvec_t::rvec_t(vec_t v);
      \endverbatim
      so that o2scl_interp can automatically "reverse" a vector if 
      necessary.
  */
  template<class vec_t=ovector_const_view, 
    class rvec_t=ovector_const_reverse> class o2scl_interp {
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
  
    /// Pointer to base interpolation object
    base_interp<vec_t> *itp;
  
    /// Pointer to base interpolation object for reversed vectors
    base_interp<rvec_t> *ritp;

    /// Pointer to base interpolation manager
    base_interp_mgr<vec_t> *bim1;
  
    /// Pointer to base interpolation manager for reversed vectors
    base_interp_mgr<rvec_t> *bim2;

      
#endif
      
    public:
      
    /** \brief Default base interpolation object (cubic spline with natural 
	boundary conditions)
    */
    def_interp_mgr<vec_t,cspline_interp> dim1;
  
    /** \brief Default base interpolation object for reversed vectors
	(cubic spline with natural boundary conditions)
    */
    def_interp_mgr<rvec_t,cspline_interp> dim2;

    /// Create with base interpolation objects \c it and \c rit
    o2scl_interp(base_interp_mgr<vec_t> &it, base_interp_mgr<rvec_t> &rit) {
      bim1=&it;
      bim2=&rit;

      itp=bim1->new_interp();
      ritp=bim2->new_interp();
    }

    /** \brief Create an interpolator using the default
	cubic spline interpolation
    */
    o2scl_interp() {
      bim1=&dim1;
      bim2=&dim2;

      itp=bim1->new_interp();
      ritp=bim2->new_interp();
    }

    virtual ~o2scl_interp() {
      bim1->free_interp(itp);
      bim2->free_interp(ritp);
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual double interp(const double x0, size_t n, const vec_t &x, 
			  const vec_t &y) 
    {
	
      double ret=0.0;
      
      if (x[0]>x[n-1]) {

	if (n<ritp->min_size) {
	  O2SCL_ERR("Vector size too small in interp().",gsl_edom);
	}
      
	// Decreasing case

	rvec_t ax(x);
	rvec_t ay(y);

	int r1=ritp->allocate(n);
	if (r1==0) {
	  int r2=ritp->init(ax,ay,n);
	  if (r2==0) {
	    ritp->interp(ax,ay,n,x0,ret);
	    ritp->free();
	  }
	}

      } else {

	if (n<itp->min_size) {
	  O2SCL_ERR("Vector size too small in interp().",gsl_edom);
	}
      
	// Increasing case
	  
	int r1=itp->allocate(n);
	if (r1==0) {
	  int r2=itp->init(x,y,n);
	  if (r2==0) {
	    itp->interp(x,y,n,x0,ret);
	    itp->free();
	  }
	}
	  
      }
      
      return ret;
    }		      
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual double deriv(const double x0, size_t n, const vec_t &x, 
			 const vec_t &y) 
    {
      
      double ret=0.0;
      
      if (x[0]>x[n-1]) {
	  
	if (n<ritp->min_size) {
	  O2SCL_ERR("Vector size too small in deriv().",gsl_edom);
	}
      
	rvec_t ax(x);
	rvec_t ay(y);

	int r1=ritp->allocate(n);
	if (r1==0) {
	  int r2=ritp->init(ax,ay,n);
	  if (r2==0) {
	    ritp->deriv(ax,ay,n,x0,ret);
	    ritp->free();
	  }
	}
	  
      } else {

	if (n<itp->min_size) {
	  O2SCL_ERR("Vector size too small in deriv().",gsl_edom);
	}
      
	// Increasing case
	  
	int r1=itp->allocate(n);
	if (r1==0) {
	  int r2=itp->init(x,y,n);
	  if (r2==0) {
	    itp->deriv(x,y,n,x0,ret);
	    itp->free();
	  }
	}
      }
      
      return ret;
    }		      
    
    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual double deriv2(const double x0, size_t n, const vec_t &x, 
			  const vec_t &y) 
    {
      
      double ret=0.0;
      
      if (x[0]>x[n-1]) {
	  
	if (n<ritp->min_size) {
	  O2SCL_ERR("Vector size too small in deriv2().",gsl_edom);
	}
      
	// Decreasing case
	    
	rvec_t ax(x);
	rvec_t ay(y);

	int r1=ritp->allocate(n);
	if (r1==0) {
	  int r2=ritp->init(ax,ay,n);
	  if (r2==0) {
	    ritp->deriv2(ax,ay,n,x0,ret);
	    ritp->free();
	  }
	}

      } else {

	if (n<itp->min_size) {
	  O2SCL_ERR("Vector size too small in deriv2().",gsl_edom);
	}
      
	// Increasing case

	int r1=itp->allocate(n);
	if (r1==0) {
	  int r2=itp->init(x,y,n);
	  if (r2==0) {
	    itp->deriv2(x,y,n,x0,ret);
	    itp->free();
	  }
	}

      }
      
      return ret;
    }		      
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual double integ(const double x1, const double x2, size_t n, 
			 const vec_t &x, const vec_t &y) 
    {
      
      double ret=0.0;
      
      if (x[0]>x[n-1]) {
	  
	if (n<ritp->min_size) {
	  O2SCL_ERR("Vector size too small in integ().",gsl_edom);
	}
      
	// Decreasing case
	    
	rvec_t ax(x);
	rvec_t ay(y);

	int r1=ritp->allocate(n);
	if (r1==0) {
	  int r2=ritp->init(ax,ay,n);
	  if (r2==0) {
	    ritp->integ(ax,ay,n,x1,x2,ret);
	    ritp->free();
	  }
	}

      } else {

	if (n<itp->min_size) {
	  O2SCL_ERR("Vector size too small in integ().",gsl_edom);
	}
      
	// Increasing case
	    
	int r1=itp->allocate(n);
	if (r1==0) {
	  int r2=itp->init(x,y,n);
	  if (r2==0) {
	    itp->integ(x,y,n,x1,x2,ret);
	    itp->free();
	  }
	}

      }
      
      return ret;
    }		      
    
    /// Set base interpolation object
    int set_type(base_interp_mgr<vec_t> &it, base_interp_mgr<rvec_t> &rit) {
      bim1->free_interp(itp);
      bim2->free_interp(ritp);
      bim1=&it;
      bim2=&rit;
      itp=bim1->new_interp();
      ritp=bim2->new_interp();
      return 0;
    }
      
#ifndef DOXYGEN_INTERNAL

    private:

    o2scl_interp<vec_t,rvec_t>(const o2scl_interp<vec_t,rvec_t> &);
    o2scl_interp<vec_t,rvec_t>& operator=(const o2scl_interp<vec_t,rvec_t>&);

#endif

  };
  
  /** 
      \brief Interpolation class for pre-specified vector
      
      This interpolation class is intended to be sufficiently general
      to handle any vector type. Interpolation of \ref ovector like
      objects is performed with the default template parameters, and
      \ref array_interp_vec is provided for simple interpolation on
      C-style arrays.

      The class automatically handles decreasing arrays by copying the
      old array to a reversed version. For several interpolations on
      the same data, copying the initial array can be faster than
      overloading operator[].

      This class does not double check the vector to ensure that 
      all of the intervals for the abcissa are all increasing or
      all decreasing. 

      The type of interpolation to be performed can be specified 
      using the set_type() function. The default is cubic splines
      with natural boundary conditions. 
      
      \todo Need to fix constructor to behave properly if init()
      fails.  It should free the memory and set \c ln to zero.
  */
  template<class vec_t=ovector_const_view, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class o2scl_interp_vec {
      
#ifndef DOXYGEN_INTERNAL
      
    protected:
      
    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /// Base interpolation object
    base_interp<vec_t> *itp;

    /// The interpolation manager
    base_interp_mgr<vec_t> *bim;
      
    /// True if the original array was increasing
    bool inc;

    /// Pointer to the user-specified x vector
    const vec_t *lx;

    /// Pointer to the user-specified x vector
    const vec_t *ly;

    /// Reversed version of x
    alloc_vec_t lrx;

    /// Reversed version of y
    alloc_vec_t lry;

    /// Size of user-specified vectors
    size_t ln;

#endif
      
    public:
    
    /// Create with base interpolation object \c it
    o2scl_interp_vec(base_interp_mgr<vec_t> &it, 
		     size_t n, const vec_t &x, const vec_t &y) {
      bim=&it;
      itp=bim->new_interp();
      ln=0;
	
      if (n<itp->min_size) {
	O2SCL_ERR("Vector size too small in o2scl_interp_vec().",gsl_edom);
      } else {
	size_t iend=n-1;
	  
	if (x[0]>x[iend]) {
	    
	  ao.allocate(lrx,n);
	  ao.allocate(lry,n);

	  for(size_t i=0;i<n;i++) {
	    lrx[iend-i]=x[i];
	    lry[iend-i]=y[i];
	  }
	    
	  int r1=itp->allocate(n);
	  if (r1==0) {
	    int r2=itp->init(lrx,lry,n);
	    if (r2==0) {
	      ln=n;
	      inc=false;
	    } else {
	      ao.free(lrx);
	      ao.free(lry);
	      itp->free();
	    }
	  } else {
	    ao.free(lrx);
	    ao.free(lry);
	  }
	
	} else if (x[0]<=x[iend]) {
	
	  int r1=itp->allocate(n);
	  if (r1==0) {
	    int r2=itp->init(x,y,n);
	    if (r2==0) {
	      ln=n;
	      lx=&x;
	      ly=&y;
	      inc=true;
	    } else {
	      itp->free();
	    }
	  }
	    
	} else {
	  O2SCL_ERR("Vector endpoints equal in o2scl_interp_vec().",
		    gsl_einval);
	}

      }
	
    }
      
    virtual ~o2scl_interp_vec() {
      if (ln>0) {
	itp->free();
	if (inc==false) {
	  ao.free(lrx);
	  ao.free(lry);
	}
      }
      bim->free_interp(itp);
    }
    
    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual double interp(const double x0) {
      double ret=0.0;
      if (ln>0) {
	if (inc) itp->interp(*lx,*ly,ln,x0,ret);
	else itp->interp(lrx,lry,ln,x0,ret);
      }
      return ret;
    }		      
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual double deriv(const double x0) {
      double ret=0.0;
      if (ln>0) {
	if (inc) itp->deriv(*lx,*ly,ln,x0,ret);
	else itp->deriv(lrx,lry,ln,x0,ret);
      }
      return ret;
    }		      
    
    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual double deriv2(const double x0) {
      double ret=0.0;
      if (ln>0) {
	if (inc) itp->deriv2(*lx,*ly,ln,x0,ret);
	else itp->deriv2(lrx,lry,ln,x0,ret);
      }
      return ret;
    }		      
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual double integ(const double x1, const double x2) {
      double ret=0.0;
      if (ln>0) {
	if (inc) itp->integ(*lx,*ly,ln,x1,x2,ret);
	else itp->integ(lrx,lry,ln,x1,x2,ret);
      }
      return ret;
    }		      
    
    /// Set base interpolation object
    int set_type(base_interp_mgr<vec_t> &it) {
      bim=&it;
      return 0;
    }
  
#ifndef DOXYGEN_INTERNAL

    private:
  
    o2scl_interp_vec<vec_t,alloc_vec_t,alloc_t>
    (const o2scl_interp_vec<vec_t,alloc_vec_t,alloc_t> &);
    o2scl_interp_vec<vec_t,alloc_vec_t,alloc_t>& operator=
    (const o2scl_interp_vec<vec_t,alloc_vec_t,alloc_t>&);
  
#endif
  
  };
  
  /** \brief A specialization of o2scl_interp for C-style double arrays
   */
#ifdef DOXYGENP
  template<size_t n> class array_interp :
  public o2scl_interp
#else
    template<size_t n> class array_interp :
  public o2scl_interp<double[n],array_const_reverse<n> >
#endif
    {
    
    public:
    
      /// Create with base interpolation objects \c it and \c rit
      array_interp(base_interp_mgr<double[n]> &it, 
		   base_interp_mgr<array_const_reverse<n> > &rit)
	: o2scl_interp<double[n],array_const_reverse<n> >(it,rit)
	{}
   
	/** \brief Create an interpolator using the default base
	    interpolation objects
	*/
	array_interp() : 
	  o2scl_interp<double[n],array_const_reverse<n> >() {
	}

    };

  /** 
      \brief A specialization of o2scl_interp_vec for C-style arrays
  */
#ifdef DOXYGENP
  template<class arr_t> class array_interp_vec : 
  public o2scl_interp_vec
#else
    template<class arr_t> class array_interp_vec : 
  public o2scl_interp_vec<arr_t,arr_t,array_alloc<arr_t> > 
#endif
    {
    
    public:
    
      /// Create with base interpolation object \c it
      array_interp_vec(base_interp_mgr<arr_t> &it, 
		       size_t nv, const arr_t &x, const arr_t &y) :
	o2scl_interp_vec<arr_t,arr_t,array_alloc<arr_t> >(it,nv,x,y) {
      }
    
    };

#ifndef DOXYGENP
}
#endif

#endif
