/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SEARCH_VEC_H
#define O2SCL_SEARCH_VEC_H

#include <iostream>
#include <string>
#include <o2scl/err_hnd.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/vector.h>

/* \file search_vec.h
   \brief File for definitions for \ref search_vec

   \comment 
   This file isn't included in Doxygen at the moment since there
   isn't anything in this file other than the search_vec classes.
   \endcomment
*/

#ifndef DOXYGENP
namespace o2scl {
#endif


  /** 
      \brief Searching class for monotonic data with caching
      
      A searching class for monotonic vectors. A caching system
      similar to \c gsl_interp_accel is used.
      
      To find the interval containing a value, use find(). If you
      happen to know in advance that the vector is increasing or
      decreasing, then you can use find_inc() or find_dec()
      instead. To ignore the caching and just use straight binary
      search, you can use the functions in \ref vector.h .

      Alternatively, if you just want to find the index with the
      element closest to a specified value, use ordered_lookup(). Note
      that ordered_lookup() is slightly different than \ref
      ovector_tlate::lookup(), since the latter does not assume the
      data is monotonic.
      
      \note The behavior of these functions is undefined if some of
      the user-specified data is not finite or not strictly
      monotonic. Two adjacent data points should not be equal. This
      class does not verify that the user-specified data has these
      properties.
  */
template<class vec_t> class search_vec {

#ifndef DOXYGEN_INTERNAL

  protected:

    /** \brief Storage for the most recent index
	
        \note This is marked mutable to ensure const-correctness is 
	straightforward.
    */
    mutable size_t cache;

    /// Desc
    vec_t *v;
    
    /// Desc
    size_t n;

#endif

  public:

    /// Create a searching object with vector \c x of size \c nn
    search_vec(size_t nn, vec_t &x) {
      if (nn<2) {
	O2SCL_ERR("Vector too small in search_vec::search_vec().",
		      gsl_einval);
      }
      cache=0;
      v=&x;
      n=nn;
    }

    /** \brief Search an increasing or decreasing vector for the
	interval containing <tt>x0</tt>

	This function is identical to find_inc() if the data is
	increasing, and find_dec() if the data is decreasing. 
    */
    size_t find(const double x0) const {
      if ((*v)[0]<(*v)[n-1]) return find_inc(x0);
      return find_dec(x0);
    }

    /** 
	\brief Search an increasing vector for the interval
	containing <tt>x0</tt>

	This function is a cached version of 
	\ref vector_bsearch_inc() . 

	This function is not quite the same as
	<tt>gsl_interp_accel_find()</tt>, because it will return the
	index of the last element in the array if \c x0 is larger than
	the last element. One can call this function with the
	parameter \c n as one less than the vector size in order to go
	back to the behavior of the GSL function. This function,
	unlike the GSL function, does not record cache hits and
	misses.
    */
    size_t find_inc(const double x0) const {
      if (x0<(*v)[cache]) {
	cache=vector_bsearch_inc<vec_t,double>(x0,*v,0,cache);
      } else if (x0>=(*v)[cache+1]) {
	cache=vector_bsearch_inc<vec_t,double>(x0,*v,cache,n-1);
      }
      return cache;
    }
    
    /** \brief Search a decreasing vector for the interval
	containing <tt>x0</tt>

	This function is a cached version of \ref vector_bsearch_dec()
	.  The operation of this function is undefined if the data is
	not strictly monotonic, i.e. if some of the data elements are
	equal. This function will call the error handler if \c n is
	less than 1.
    */
    size_t find_dec(const double x0) const {
      if (x0>(*v)[cache]) {
	cache=vector_bsearch_dec<vec_t,double>(x0,*v,0,cache);
      } else if (x0<(*v)[cache+1]) {
	cache=vector_bsearch_dec<vec_t,double>(x0,*v,cache,n-1);
      }
      return cache;
    }

    /** 
	\brief Find the index of x0 in the ordered array \c x 

	This returns the index i for which x[i] is as close as
	possible to x0 if x[i] is either increasing or decreasing.

	If you have a non-monotonic vector, you can use \ref
	vector_lookup() instead, or if you a non-monotonic \ref
	ovector or \ref uvector object, consider using \ref
	ovector_view_tlate::lookup() or \ref
	uvector_view_tlate::lookup() instead of this function.

	Generally, if there are two adjacent entries with the same
	value, this function will return the entry with the smaller
	index. (This is the same as the convention in \ref
	ovector_view_tlate::lookup() and \ref
	uvector_view_tlate::lookup() ).

	\future This function just uses the <tt>find</tt> functions
	and then adjusts the answer at the end if necessary. It might
	be possible to improve the speed by rewriting this from
	scratch.
    */
    size_t ordered_lookup(const double x0) const {
      if (n<1) {
	O2SCL_ERR2_RET("Not enough data for ",
		       "search_vec::ordered_lookup().",gsl_einval);
      }

      size_t row;

      if ((*v)[0]<=(*v)[n-1]) {

	// Increasing case

	if (x0>=(*v)[n-1]) {
	  row=n-1;
	} else { 
	  row=find_inc(x0);
	  if (row<n-1 && fabs((*v)[row+1]-x0)<fabs((*v)[row]-x0)) row++;
	}
    
      } else {

	// Decreasing case
    
	if (x0<=(*v)[n-1]) {
	  row=n-1;
	} else {
	  row=find_dec(x0);
	  if (row<n-1 && fabs((*v)[row+1]-x0)<fabs((*v)[row]-x0)) row++;
	}
      }

      return row;
    }

  };

  /** \brief An extended search_vec which is allowed to return the last 
      element
   */
  template<class vec_t> class search_vec_ext : 
  public search_vec<vec_t> {

  public:

    /** 
	\brief Desc

	\future This could be rewritten to allow vectors with
	only one element. 
     */
    /// Create a searching object with vector \c x of size \c nn
    search_vec_ext(size_t nn, vec_t &x) : search_vec<vec_t>(nn,x) {
      //if (n<1) {
      //O2SCL_ERR_RET("Vector too small in search_vec_ext::search_vec_ext().",
      //gsl_einval);
      //}
    }

    /** \brief Search an increasing or decreasing vector for the interval
	containing <tt>x0</tt>
    */
    size_t find(const double x0) const {
      if ((*this->v)[0]<(*this->v)[this->n-1]) return find_inc(x0);
      return find_dec(x0);
    }

    /** 
	\brief Search an increasing vector for the interval
	containing <tt>x0</tt>
    */
    size_t find_inc(const double x0) const {
      if (x0<(*this->v)[this->cache]) {
	this->cache=vector_bsearch_inc<vec_t,double>
	  (x0,*this->v,0,this->cache);
      } else if (this->cache<this->n-1 && x0>=(*this->v)[this->cache+1]) {
	this->cache=vector_bsearch_inc<vec_t,double>
	  (x0,*this->v,this->cache,this->n);
      }
      return this->cache;
    }
    
    /** \brief Search a decreasing vector for the interval
	containing <tt>x0</tt>
    */
    size_t find_dec(const double x0) const {
      if (x0>(*this->v)[this->cache]) {
	this->cache=vector_bsearch_dec<vec_t,double>
	  (x0,*this->v,0,this->cache);
      } else if (this->cache<this->n-1 && x0<(*this->v)[this->cache+1]) {
	this->cache=vector_bsearch_dec<vec_t,double>
	  (x0,*this->v,this->cache,this->n);
      }
      return this->cache;
    }

  };

#ifndef DOXYGENP
}
#endif

#endif
