/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TABLE3D_H
#define O2SCL_TABLE3D_H

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cmath>
#include <sstream>
#include <o2scl/misc.h>
#include <o2scl/err_hnd.h>
#include <o2scl/multi_funct.h>
#include <o2scl/collection.h>
#include <o2scl/search_vec.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/interp.h>
#include <o2scl/smart_interp.h>
#include <o2scl/table.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief A data structure containing many slices of two-dimensional
      data points defined on a grid

      \future Improve interpolation and derivative caching
      \future Make a 'const' version of the interpolation functions
      \future Write a 2d histogram functions using 
      multiple applications of search_vec's find_interval().
  */
  class table3d {
    
  public:
  
    /** \brief Create a new 3D \table
     */
    table3d();

    virtual ~table3d();

    /// \name Initialization
    //@{
    /** 
	\brief Initialize the x-y grid
	
	This function will not allow you to redefine the grid when
	there is data in the \table if a grid of a different size was
	already set from a previous call to either set_xy() or
	set_size(). However, you may freely redefine the grid
	after a call to clear_data() or clear_table(). You may change
	individual grid points at any time with set_grid_x() and 
	set_grid_y().

    */
    template<class vec2_t> int set_xy(std::string x_name, size_t nx, 
				      const vec2_t &x, std::string y_name,
				      size_t ny, const vec2_t &y) {
      if (has_slice && (size_set || xy_set) && (nx!=numx || ny!=numy)) {
	O2SCL_ERR_RET("Size cannot be reset in table3d::set_xy().",
		      o2scl::gsl_einval);
      }
      if (xy_set) {
	xval.free();
	yval.free();
      }
      numx=nx;
      numy=ny;
      xname=x_name;
      yname=y_name;
      xval.allocate(nx);
      yval.allocate(ny);
      for(size_t i=0;i<nx;i++) (xval)[i]=x[i];
      for(size_t i=0;i<ny;i++) (yval)[i]=y[i];
      size_set=true;
      xy_set=true;
      return 0;
    }

    /** 
	\brief Initialize \table size

	This function will not allow you to resize the \table if it
	already has data or if the size has already been set with the
	set_xy() function, unless you clear the data with clear_data()
	or the \table with clear_table() first.
    */
    int set_size(size_t nx, size_t ny);
    //@}

    // --------------------------------------------------------
    /// \name On-grid get and set methods 
    //@{

    /** \brief Set element in slice \c name at location <tt>ix,iy</tt> 
	to value \c val
    */
    int set(size_t ix, size_t iy, std::string name, double val);

    /** \brief Set element in slice of index \c z at location
	<tt>ix,iy</tt> to value \c val .
    */
    int set(size_t ix, size_t iy, size_t z, double val);

    /** \brief Get element in slice \c name at location
	<tt>ix,iy</tt>
    */
    double &get(size_t ix, size_t iy, std::string name);

    /** \brief Get element in slice \c name at location
	<tt>ix,iy</tt> (const version)
    */
    const double &get(size_t ix, size_t iy, std::string name) const;

    /** \brief Get element in slice of index \c z at location
	<tt>ix,iy</tt>
    */
    double &get(size_t ix, size_t iy, size_t z);

    /** \brief Get element in slice of index \c z at location
	<tt>ix,iy</tt> (const version)
    */
    const double &get(size_t ix, size_t iy, size_t z) const;
    //@}

    // --------------------------------------------------------
    /** \name Off-grid get and set methods 

	These methods return the value of a slice on the grid
	point nearest to a user-specified location. For
	interpolation into a point off the grid, use 
	\ref table3d::interp().
     */
    //@{

    /** \brief Set element in slice \c name at the nearest location to
	<tt>x,y</tt> to value \c val
    */
    int set_val(double x, double y, std::string name, double val);

    /** \brief Set element in slice of index \c z at the nearest
	location to <tt>x,y</tt> to value \c val
    */
    int set_val(double x, double y, size_t z, double val);
    
    /** \brief Get element in slice \c name at location closest to
	<tt>x,y</tt>
    */
    double &get_val(double x, double y, std::string name);
    
    /** \brief Get element in slice \c name at location closest to
	<tt>x,y</tt>
    */
    const double &get_val(double x, double y, std::string name) const;
    
    /** \brief Get element in slice of index \c z at location closest
	to <tt>x,y</tt>
    */
    double &get_val(double x, double y, size_t z);

    /** \brief Get element in slice of index \c z at location closest
	to <tt>x,y</tt>
    */
    const double &get_val(double x, double y, size_t z) const;

    /** \brief Set elements in the first <tt>nv</tt> slices at the
	nearest location to <tt>x,y</tt> to value \c val
    */
    template<class vec_t> 
      int set_slices(double x, double y, size_t nv, vec_t &vals) {
      size_t ix, iy;
      lookup_x(x,ix);
      lookup_y(y,iy);

      for(size_t i=0;i<nv && i<list.size();i++) {
	list[i][ix][iy]=vals[i];
      }
      return 0;
    }
    
    /** \brief Get the data for every slice at the nearest location to
	<tt>x,y</tt>
    */
    template<class vec_t>
      int get_slices(double x, double y, size_t nv, vec_t &v) {

      size_t ix, iy;

      lookup_x(x,ix);
      lookup_y(y,iy);

      for(size_t i=0;i<nv && i<list.size();i++) {
	v[i]=list[i][ix][iy];
      }

      return 0;
    }

    //@}

    // --------------------------------------------------------
    /// \name Off-grid get and set methods returning nearest point
    //@{

    /** \brief Set element in slice \c name at the nearest location to
	<tt>x,y</tt> to value \c val
    */
    int set_val_ret(double &x, double &y, std::string name, double val);

    /** \brief Set element in slice of index \c z at the nearest
	location to <tt>x,y</tt> to value \c val
    */
    int set_val_ret(double &x, double &y, size_t z, double val);
    
    /** \brief Get element in slice \c name at location closest to
	<tt>x,y</tt>, and also return the corresponding values of \c x
	and \c y
    */
    double &get_val_ret(double &x, double &y, std::string name);
    
    /** \brief Get element in slice \c name at location closest to
	<tt>x,y</tt>, and also return the corresponding values of \c x
	and \c y
    */
    const double &get_val_ret(double &x, double &y, std::string name) const;
    
    /** \brief Get element in slice of index \c z at location closest
	to <tt>x,y</tt>, and also return the corresponding values of
	\c x and \c y
    */
    double &get_val_ret(double &x, double &y, size_t z);

    /** \brief Get element in slice of index \c z at location closest
	to <tt>x,y</tt>, and also return the corresponding values of
	\c x and \c y
    */
    const double &get_val_ret(double &x, double &y, size_t z) const;

    /** \brief Set elements in the first <tt>nv</tt> slices at the
	nearest location to <tt>x,y</tt> to value \c val
    */
    template<class vec_t> 
      int set_slices_ret(double &x, double &y, size_t nv, vec_t &vals) {
      size_t ix, iy;
      lookup_x(x,ix);
      lookup_y(y,iy);
      x=xval[ix];
      y=yval[iy];

      for(size_t i=0;i<nv && i<list.size();i++) {
	list[i][ix][iy]=vals[i];
      }
      return 0;
    }

    /** \brief Get elements in the first <tt>nv</tt> slices at the
	nearest location to <tt>x,y</tt> to value \c val
    */
    template<class vec_t> 
      int get_slices_ret(double &x, double &y, size_t nv, vec_t &vals) {
      size_t ix, iy;
      lookup_x(x,ix);
      lookup_y(y,iy);
      x=xval[ix];
      y=yval[iy];
      
      for(size_t i=0;i<nv && i<list.size();i++) {
	vals[i]=list[i][ix][iy];
      }
      return 0;
    }
    
    //@}

    // --------------------------------------------------------
    /// \name Grid information get and set methods
    //@{

    /// Set x grid point at index \c ix
    int set_grid_x(size_t ix, double val);
    
    /// Set y grid point at index \c iy
    int set_grid_y(size_t iy, double val);
    
    /// Get x grid point at index \c ix
    double get_grid_x(size_t ix);

    /// Get y grid point at index \c iy
    double get_grid_y(size_t iy);

    /// Get the name of the x grid variable
    std::string get_x_name() {
      return xname;
    }

    /// Get the name of the y grid variable
    std::string get_y_name() {
      return yname;
    }

    /// Set the name of the x grid variable
    int set_x_name(std::string name) {
      xname=name;
      return 0;
    }
    
    /// Set the name of the y grid variable
    int set_y_name(std::string name) {
      yname=name;
      return 0;
    }

    /// Get a const reference to the full x grid
    const ovector &get_x_data() const { return xval; }

    /// Get a const reference to the full y grid
    const ovector &get_y_data() const { return yval; }

    //@}

    // --------------------------------------------------------
    /// \name Size get methods
    //@{

    /// Get the size of the slices
    int get_size(size_t &nx, size_t &ny) const;

    /// Get the x size
    int get_nx() const {
      return numx;
    }
    
    /// Get the y size
    int get_ny() const {
      return numy;
    }

    /// Get the number of slices
    int get_nslices() const;

    //@}

    // --------------------------------------------------------
    /// \name Slice manipulation 
    //@{
    
    /// Create a set of new slices specified in the string \c names
    int line_of_names(std::string names);

    /** \brief Returns the name of slice with index \c z
     */
    std::string get_slice_name(size_t z) const;

    /** \brief Add a new slice
     */
    int new_slice(std::string name);

    /** \brief Set all of the values in slice \c name to \c val
     */
    int set_slice_all(std::string name, double val);

    /** 
	\brief Find the index for column named \c name
    */
    int lookup_slice(std::string name, size_t &ix) const;
    
    /** 
	\brief Return true if slice is already present
    */
    bool is_slice(std::string name, int &ix);
    
    /** \brief Rename slice named \c olds to \c news
	
	This is slow since we have to delete the column and re-insert
	it. This process in turn mangles all of the iterators in the
	list.
    */
    int rename_slice(std::string olds, std::string news);
    
    /** \brief Make a new slice named \c dest which is a copy 
	of the slice with name given in \c src. 
     */
    int copy_slice(std::string src, std::string dest);
    
    /** \brief Initialize all values of slice named \c scol to \c val 
     */
    int init_slice(std::string scol, double val);

    /// Return a constant reference to a slice
    const omatrix &get_slice(std::string scol) const;

    /// Return a constant reference to a slice
    const omatrix &get_slice(size_t iz) const;
    
    /** \brief Return a constant reference to all the slice data

	\comment
	This isn't designated const, i.e. as
	const std::vector<omatrix> &get_data() const;
	because it would then have to be 
	const std::vector<const omatrix> &get_data() const;
	\endcomment
     */
    const std::vector<omatrix> &get_data();
    //@}
  
    // --------------------------------------------------------
    /// \name Lookup and search methods
    //@{
  
    /** 
	\brief Look for a value in the x grid
    */
    int lookup_x(double val, size_t &ix) const;
  
    /** 
	\brief Look for a value in the y grid
    */
    int lookup_y(double val, size_t &iy) const;

    /** 
	\brief Look for a value in a specified slice
    */
    int lookup(double val, std::string slice, size_t &ix, size_t &iy) const;
    //@}

    // --------------------------------------------------------
    /// \name Interpolation, differentiation, and integration
    //@{

    /** 
        \brief Specify the base interpolation objects to use
    */
    int set_interp(base_interp_mgr<ovector_const_view> &b1,
		   base_interp_mgr<ovector_const_subvector> &b2);
    
    /** 
	\brief Interpolate \c x and \c y in slice named \c name
	
    */
    double interp(double x, double y, std::string name);

    /** 
	\brief Interpolate the derivative of the data with
	respect to the x grid at point
	\c x and \c y in slice named \c name
    */
    double deriv_x(double x, double y, std::string name);

    /** 
	\brief Interpolate the derivative of the data with
	respect to the y grid at point
	\c x and \c y in slice named \c name
    */
    double deriv_y(double x, double y, std::string name);

    /** 
	\brief Interpolate the mixed second 
	derivative of the data at point
	\c x and \c y in slice named \c name
    */
    double deriv_xy(double x, double y, std::string name);

    /** 
	\brief Interpolate the integral of the data 
	respect to the x grid 
    */
    double integ_x(double x1, double x2, double y, std::string name);

    /** 
	\brief Interpolate the integral of the data 
	respect to the y grid 
    */
    double integ_y(double x, double y1, double y2, std::string name);

    /** \brief Fill a vector of interpolated values from each slice at the
	point <tt>x,y</tt>
     */
    template<class vec_t>
      int interp_slices(double x, double y, size_t nv, vec_t &v) {
      
      for (size_t i=0;i<list.size();i++) {
	std::string name=get_slice_name(i);
	v[i]=interp(x,y,name);
      }

      return 0;
    }

    //@}

    // --------------------------------------------------------
    /// \name Extract 2-dimensional tables
    //@{
    /// Extract a table at a fixed x grid point 
    int extract_x(double x, table &t);
    
    /// Extract a table at a fixed y grid point 
    int extract_y(double y, table &t);
    //@}

    // --------------------------------------------------------
    /// \name Clear methods 
    //@{
    /** \brief Zero the data entries but keep the slice names
	and grid
    */
    int zero_table();

    /** \brief Clear the table and the slice names
     */
    void clear_table();

    /** \brief Remove all of the data by setting the number
	of lines to zero

	This leaves the column names intact and does not remove
	the constants.
    */
    void clear_data();
    //@}

    // --------------------------------------------------------
    /// \name Summary method 
    //@{
    /** 
	\brief Output a summary of the information stored
	
	Outputs the number of constants, the grid information,
	and a list of the slice names
    */
    int summary(std::ostream *out, int ncol=79) const;
    //@}

#ifdef NEVER_DEFINED
    // --------------------------------------------------------
    /// \name Contour lines method
    //@{

    /** \brief Create contour lines from the slice named \c name

	This uses \ref contour to compute contour lines (stored in \c
	clines) from slice \c name given \c nlev contour levels in \c
	levs .
     */
    template<class vec_t> 
      int slice_contours(std::string name, size_t nlev, vec_t &levs,
			 std::vector<contour_line> &clines) {
      size_t z=0;
      int ret=lookup_slice(name,z);
      
      contour co;
      // Note that the contour class stores the matrix data in (y,x)
      // order, not (x,y) so the order has to be reversed here.
      co.set_data(numy,numx,yval,xval,list[z]);
      co.set_levels(nlev,levs);
      co.calc_contours(clines);

      return 0;
    }
    //@}
#endif
    
    // --------------------------------------------------------
    /// \name Manipulating constants
    //@{
    /// Add a constant
    virtual int add_constant(std::string name, double val);

    /// Remove a constant
    virtual int remove_constant(std::string name);
    
    /// Add a constant
    virtual int set_constant(std::string name, double val);

    /// Get a constant
    virtual double get_constant(std::string name);
    //@}

    /// \name Default interpolation objects
    //@{
    def_interp_mgr<ovector_const_view,cspline_interp> dim1;

    def_interp_mgr<ovector_const_subvector,cspline_interp> dim2;
    //@}

    /// Return the type, \c "table3d".
    virtual const char *type() { return "table3d"; }

  protected:

#ifndef DOXYGEN_INTERNAL
    
    /// \name Interpolation data
    //@{
    /// The base interpolation object
    base_interp_mgr< ovector_const_view> *bimp1;

    /// The subvector base interpolation object
    base_interp_mgr< ovector_const_subvector> *bimp2;

    /// The array of sm_interp_vec pointers 
    sm_interp_vec **si;
    /// Matrices for interpolation
    omatrix_col **aci;
    //@}
    
    friend class table3d_io_type;
    
    /// \name Iterator types
    //@{
    typedef std::map<std::string,size_t, string_comp>::iterator map_iter;
    typedef std::map<std::string,size_t, string_comp>::const_iterator 
      map_const_iter;
    //@}
  
    /// \name Data storage
    //@{
    /// The list of constants
    std::map<std::string,double> constants;

    /// The size of the x grid
    size_t numx;

    /// The size of the y grid
    size_t numy;

    /// A tree connecting column names to list indexes
    std::map<std::string,size_t, string_comp> tree;

    /// The name for the x grid
    std::string xname;

    /// The name for the y grid
    std::string yname;

    /// The pointers to the matrices
    std::vector<omatrix> list;

    /// The x grid
    ovector xval;

    /// The y grid
    ovector yval;

    /// True if the grid has been set
    bool xy_set;

    /// True if the size of the grid has been set
    bool size_set;

    /// True if the table has at least one slice
    bool has_slice;
    //@}
  
    /// \name Tree iterator boundaries
    //@{
    /// Return the beginning of the slice tree
    map_iter begin() {return tree.begin();};
    /// Return the end of the slice tree
    map_iter end() {return tree.end();};
    /// Return the beginning of the slice tree
    map_const_iter const_begin() const {return tree.begin();};
    /// Return the end of the slice tree
    map_const_iter const_end() const {return tree.end();};
    //@}

#endif

  };

  class table3d_io_type : public  io_tlate<table3d> {
  public:
    
  table3d_io_type() :  io_tlate<table3d>() {};
      
    virtual const char *type() { return "table3d"; }
      
    virtual int input( cinput *ci,  in_file_format *ins,
		       table3d *at) {

      bool size_set, xy_set;
      int nx, ny, nt, ncons, r1;
      double dtemp;
      std::string name, nmt;

      // Constants
      ins->int_in(ncons,"const");
      for(int i=0;i<ncons;i++) {
	ins->word_in(name);
	ins->double_in(dtemp);
	r1=at->add_constant(name,dtemp);
	if (r1!=0) {
	  O2SCL_ERR_RET("Failed to add constant in table3d_io_type::input.",
			gsl_esanity);
	}
      }

      ins->bool_in(size_set);
      if (size_set) {

	// Grid
	ins->int_in(nx,"nx");
	ins->int_in(ny,"ny");
	ins->bool_in(xy_set);

	if (xy_set) {
	  std::string xname, yname;
	  ins->word_in(xname);
	  ins->word_in(yname);
	  double *xv=new double[nx];
	  double *yv=new double[ny];
	  for(int i=0;i<nx;i++) ins->double_in(xv[i]);
	  for(int i=0;i<ny;i++) ins->double_in(yv[i]);
	  r1=at->set_xy(xname,((size_t)nx),xv,yname,((size_t)ny),yv);
	  if (r1!=0) {
	    O2SCL_ERR_RET("Failed to set grid in table3d_io_type::input.",
			  gsl_esanity);
	  }
	  delete[] xv;
	  delete[] yv;
	}

	ins->int_in(nt,"nt");
	if (nt>0) {

	  // Slice names
	  std::string str="", tmp;
	  for(int i=0;i<nt;i++) {
	    ins->word_in(tmp);
	    str+=tmp;
	    str+=" ";
	  }
	  r1=at->line_of_names(str);
	  if (r1!=0) {
	    O2SCL_ERR_RET("Failed to add names in table3d_io_type::input().",
			  gsl_esanity);
	  }

	  // Data
	  for(int i=0;i<nt;i++) {
	    for(int j=0;j<nx;j++) {
	      for(int k=0;k<ny;k++) {
		ins->double_in(dtemp);
		r1=at->set(j,k,i,dtemp);
		if (r1!=0) {
		  O2SCL_ERR2_RET("Failed to set() in ",
				 "table3d_io_type::input().",gsl_esanity);
		}
	      }
	    }
	  }
	}
      }
      return 0;
    }

    virtual int output(coutput *co, out_file_format *outs,
		       table3d *at) {
      
      // Constants
      outs->int_out(at->constants.size(),"const");
      std::map<std::string,double>::iterator mit;
      for(mit=at->constants.begin();mit!=at->constants.end();mit++) {
	outs->word_out(mit->first);
	outs->double_out(mit->second);
	outs->end_line();
      }

      outs->bool_out(at->size_set);
      if (at->size_set) {

	// Grid 
	outs->int_out(at->numx,"nx");
	outs->int_out(at->numy,"ny");
	outs->bool_out(at->xy_set);
	if (at->xy_set) {
	  outs->end_line();
	  outs->word_out(at->xname);
	  outs->word_out(at->yname);
	  for(size_t i=0;i<at->numx;i++) outs->double_out((at->xval)[i]);
	  for(size_t i=0;i<at->numy;i++) outs->double_out((at->yval)[i]);
	}
	outs->end_line();

	outs->int_out(at->tree.size(),"nt");
	if (at->tree.size()>0) {

	  // Slice names
	  for(size_t i=0;i<at->tree.size();i++) {
	    outs->word_out(at->get_slice_name(i));
	  }
	  outs->end_line();

	  // Data
	  for(size_t i=0;i<at->tree.size();i++) {
	    for(size_t j=0;j<at->numx;j++) {
	      for(size_t k=0;k<at->numy;k++) {
		outs->double_out(at->get(j,k,i));
	      }
	    }
	    outs->end_line();
	  }
	}
    
      }
      return 0;
    }

  };

#ifndef DOXYGENP
}
#endif

#endif
