/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_HADRONIC_EOS_H
#define O2SCL_HADRONIC_EOS_H

#include <iostream>
#include <string>
#include <o2scl/eos.h>
#include <o2scl/part.h>
#include <o2scl/deriv.h>
#include <o2scl/gsl_deriv.h>
#include <o2scl/mroot.h>
#include <o2scl/gsl_mroot_hybrids.h>
#include <o2scl/collection.h>
#include <o2scl/eff_fermion.h>
#include <o2scl/mm_funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Hadronic equation of state [abstract base]

      In the method documentation below, \f$ n \f$ is baryon number density, 
      \f$ \epsilon \f$ is energy density, and \f$ P \f$ is pressure.
    
      \htmlonly
      See more about Svprime in the Mathematica notebook at 
      <a href="hadronic_eos.nb">
      hadronic_eos.nb</a>, and
      <a href="hadronic_eos.ps">
      hadronic_eos.ps</a>.
      \endhtmlonly
      \latexonly
      See more about Svprime in the Mathematica notebook at 
      \begin{verbatim}
      doc/o2scl/extras/hadronic_eos.nb
      doc/o2scl/extras/hadronic_eos.ps
      \end{verbatim}
      \endlatexonly

      \future Could write a function to compute the "symmetry free energy"
      or the "symmetry entropy"
  */
  class hadronic_eos : public eos {
  public:
    
    hadronic_eos();

    virtual ~hadronic_eos() {};

    /// Binding energy
    double eoa;

    /// Compressibility
    double comp;

    /// Symmetry energy
    double esym;
  
    /// Saturation density
    double n0;

    /// Effective mass
    double msom;

    /// Skewness
    double kprime;
    
    /// Symmetry energy derivative
    double sprime;

    /// \name Equation of state
    //@{
    /** 
	\brief Equation of state as a function of the chemical potentials
    */
    virtual int calc_p(fermion &n, fermion &p, thermo &th)=0;

    /** 
	\brief Equation of state as a function of density
    */
    virtual int calc_e(fermion &n, fermion &p, thermo &th)=0;
    //@}

    /// \name Physical properties
    //@{
    /** 
	\brief Calculate compressibility of nuclear matter using calc_e()

	The compression modulus is defined here by: \f$ \chi = -1/V
	(dV/dP) = 1/n (dP/dn)^{-1} \f$ It is customary to use the
	incompressibility modulus \f$ K=9/(n \chi) \f$ .  This is the value
	denoted \c comp  in the code and can be written: \f$ K = 9 n d^2
	\epsilon /(d n^2) = 9 d P / (d n) \f$ .  It is often referred to as
	the "compressibility" and is about 220 MeV at saturation
	density. (Taken from Chabanat, et.  al. NPA 627 (1997) 710.) Note
	that this differs from \f$ K_2 = 9 n^2 d^2(\epsilon/n) / (d n^2)
	 \f$ by \f$ 18 P/n \f$ at any density except the saturation
	density.  
    */
    virtual double fcomp(double nb);

    /** 
	\brief Calculate binding energy using calc_e()

	\c eoa = (energy density/baryon number density-nucleon mass) at 
	 \f$ n=n_0 \f$ . \f$ E_b \approx -16/(\hbar c) \f$ 
    */
    virtual double feoa(double nb, double pf=0.5);

    /** 
	\brief Calculate symmetry energy of matter using calc_e()

	\c esym = 
	\f[
	\left(\frac{1}{2 n}\frac{d^2 \epsilon}{d \delta^2}
	\right)_{n=n_B,\delta=\delta_0}
	\f]

	where \f$ \delta=1-2 x \f$ , \f$ \delta_0=1-2 x \f$ 
	and \f$ x \f$ is the proton fraction
	(for x=0.5 at saturation density, esym \f$ \approx 32/ \hbar c \f$ )
    */
    virtual double fesym(double nb, double pf=0.5);
    
    /** 
	\brief The symmetry energy slope parameter
	
	This returns the value of the "slope parameter" 
	of the symmetry energy
	\f[
	L=3 n_{B} \left(\frac{\partial E_{sym}}{\partial n_{B}}\right)
	\f]
	in inverse Fermis. 

	where \f$ n_B \f$ is the baryon density. This ranges 
	between about zero and 200 MeV for many EOSs. If \c alt_sym
	is false (the default), then \ref fesym() is used to
	compute the symmetry energy, otherwise \ref fesym_diff()
	is used.
    */
    virtual double fesym_slope(double nb, bool alt_sym=false);

    /** \brief Calculate symmetry energy of matter as energy of 
	neutron matter minus the energy of nuclear matter

	This function returns the energy per baryon of neutron matter
	minus the energy per baryon of nuclear matter. This will
	deviate significantly from the results from fesym() only if
	the dependence of the symmetry energy on \f$ \delta \f$ is not
	quadratic.
    */
    virtual double fesym_diff(double nb);

    /** 
	\brief Calculate \f$ S^{\prime} \f$ in matter using calc_e()

	\c sprime = 
	\f[
	\left[n \frac{d}{d n}
	\left(\frac{1}{2 n}\frac{d^2 \epsilon}{d \delta^2}
	\right)\right]_{n=n_B,\delta=\delta_0}
	\f]

	where \f$ \delta=1-2 x \f$ , \f$ \delta_0=1-2 (\mathrm{pf}) \f$ 
	and \f$ x \f$ is the proton fraction
    */
    virtual double fsprime(double nb, double pf=0.5);

    /** 
	\brief Calculate skewness of nuclear matter using calc_e()

	The skewness is defined to be 
	\f$ 27 n^3 d^3 (\epsilon/n)/(d n^3) = 
	27 n^3 d^2 (P/n^2) / (d n^2) \f$ 

	and is denoted 'kprime'.
	This definition seems to be ambiguous for densities other than the 
	saturation density and is not quite analogous to the compressibility.
    */
    virtual double fkprime(double nb);

    /** 
	\brief Calculate reduced neutron effective mass using calc_e()

	Neutron effective mass (n.ms) divided by vacuum mass (n.m) in
	nuclear matter at saturation density.  Note that this simply uses
	the value of n.ms from calc_e(), so that this effective mass could
	be either the Landau or Dirac mass depending on the context. Note
	that this may not be equal to the reduced proton effective mass.  
    */
    virtual double fmsom(double nb, double pf=0.5);

    /** 
	\brief Calculate saturation density using calc_e()

	This function finds the density for which the pressure vanishes
	in matter with \f$ n_n=n_p \f$ .

	\f$ n_0 \f$ = baryon number density at which 
	\f$ P=0, n_0 \approx 0.16 \f$ 
    */
    virtual double fn0(double protfrac, double &leoa);

    /// Calculates all of the properties at the saturation density
    virtual int saturation();
    //@}

    /** 
	\brief Calculate coefficients for \gradient \part of Hamiltonian

	\note This is still somewhat experimental.

	We want the \gradient \part of the Hamiltonian in the form
	\f[
	{\cal H}_{\mathrm{grad}} = \frac{1}{2} \sum_{i=n,p}
	\sum_{j=n,p} Q_{ij}
	\vec{\nabla} n_i \cdot \vec{\nabla} n_j
	\f]

	The expression for the \gradient terms from \ref Pethick95 is 
	\f{eqnarray*}
	{\cal H}_{\mathrm{grad}}&=&-\frac{1}{4}
	\left(2 P_1 + P_{1;f}-P_{2;f}\right) 
	\nonumber \\
	&& +\frac{1}{2} \left( Q_1+Q_2 \right) 
	\left(n_n \nabla^2 n_n+n_p \nabla^2 n_p\right) \nonumber \\
	&& + \frac{1}{4}\left( Q_1-Q_2 \right) 
	\left[\left(\nabla n_n\right)^2+\left(\nabla n_p\right)^2
	\right] \nonumber \\
	&& + \frac{1}{2} \frac{d Q_2}{d n} 
	\left( n_n \nabla n_n + n_p \nabla n_p \right) \cdot \nabla n
	\f}
	This can be rewritten
	\f{eqnarray*}
	{\cal H}_{\mathrm{grad}}&=&\frac{1}{2}\left(\nabla n\right)^2
	\left[ \frac{3}{2} P_1+n \frac{d P_1}{d n}\right] \nonumber \\
	&& - \frac{3}{4} \left[ \left( \nabla n_n\right)^2 + 
	\left( \nabla n_p \right)^2 \right] \nonumber \\
	&& -\frac{1}{2} \left[ \right] \cdot \nabla n \frac{d Q_1}{d n}
	\nonumber \\ && - \frac{1}{4} \left( \nabla n\right)^2 P_2
	- \frac{1}{4} \left[ \left( \nabla n_n\right)^2 +
	\left( \nabla n_p \right)^2 \right] Q_2
	\f}
	or
	\f{eqnarray*}
	{\cal H}_{\mathrm{grad}}&=&\frac{1}{4} \left( \nabla n\right)^2
	\left[3 P_1 + 2 n \frac{d P_1}{d n}-P_2\right] \nonumber \\
	&& - \frac{1}{4} \left( 3 Q_1+Q_2 \right)
	\left[ \left( \nabla n_n\right)^2 + 
	\left( \nabla n_p \right)^2 \right] \nonumber \\
	&& - \frac{1}{2} \frac{d Q_1}{d n}
	\left[ n_n \nabla n_n + n_p \nabla n_p \right]
	\cdot \nabla n 
	\f}
	or
	\f{eqnarray*}
	{\cal H}_{\mathrm{grad}}&=&\frac{1}{4} \left( \nabla n\right)^2
	\left[3 P_1 + 2 n \frac{d P_1}{d n}-P_2\right] \nonumber \\
	&& - \frac{1}{4} \left( 3 Q_1+Q_2 \right)
	\left[ \left( \nabla n_n\right)^2 + 
	\left( \nabla n_p \right)^2 \right] \nonumber \\
	&& - \frac{1}{2} \frac{d Q_1}{d n}
	\left[ n_n \left( \nabla n_n \right)^2 +
	n_p \left( \nabla n_p \right)^2 + n \nabla n_n \cdot
	\nabla n_p \right]
	\f}

	Generally, for Skyrme-like interactions
	\f{eqnarray*}
	P_i &=& \frac{1}{4} t_i \left(1+\frac{1}{2} x_i \right) \nonumber \\
	Q_i &=& \frac{1}{4} t_i \left(\frac{1}{2} + x_i \right) \, .
	\f}
	for \f$ i=1,2 \f$ .

	This function uses the assumption \f$ x_1=x_2=0 \f$ to 
	calculate \f$ t_1 \f$ and \f$ t_2 \f$ from the neutron
	and proton effective masses assuming the Skyrme form. The
	values of \f$ Q_{ij} \f$ and their derivatives are then computed.

	The functions set_n_and_p() and set_thermo() will be called by
	gradient_qij(), to facilitate the use of the \c n, \c p, and
	\c th parameters.
       
    */
    int gradient_qij(fermion &n, fermion &p, thermo &th, 
		     double &qnn, double &qnp, double &qpp, 
		     double &dqnndnn, double &dqnndnp,
		     double &dqnpdnn, double &dqnpdnp,
		     double &dqppdnn, double &dqppdnp);
    
    /// \name Functions for calculating physical properties
    //@{
    /** 
	\brief Calculate pressure of nuclear matter
	as a function of baryon density

        Used by fcomp(), <tt>pa</tt> is unused.
    */
    double calc_pressure(double nb, int &pa);

    /** \brief Calculate pressure / baryon density squared in nuclear
	matter as a function of baryon density
	
	Used by fkprime(), <tt>pa</tt> is unused.
    */
    double calc_press_on2(double nb, int &pa);

    /** 
	\brief Calculate energy density as a function of 'delta' 

        Used by fesym(), <tt>pa</tt> is unused.
    */
    double calc_edensity(double delta, int &pa);

    /** 
	\brief Calculate symmetry energy as a function of 'delta'

        Used by fsprime(), <tt>pa</tt> is unused.
    */
    double calc_esym(double nb, int &pa);

    /** 
	\brief Return the symmetry energy at density \c nb.
	
	Used by fesym_slope().
    */
    double calc_esym(double nb, bool &alt);

    /** 
	\brief Solve for zero pressure as a function of baryon density
     
        Used by fn0().
    */
    int saturation_matter_e(double x, double &y, int &pa);
    //@}

    /// \name Other functions
    //@{
    /** 
	\brief Nucleonic matter from calc_p()
    */
    int nuc_matter_p(size_t nv, const ovector_base &x, ovector_base &y, 
		     double *&pa);
    
    /** 
	\brief Nucleonic matter from calc_e()
    */
    int nuc_matter_e(size_t nv, const ovector_base &x, ovector_base &y, 
		     double *&pa);
    //@}

    /// \name Set auxilliary objects
    //@{
    /** \brief Set class mroot object for use in calculating chemical
	potentials from densities 
    */
    virtual int set_mroot(mroot<double *,mm_funct<double *> > &mr);

    /** \brief Set class mroot object for use calculating saturation density
     */
    virtual int set_sat_root(root<int,funct<int> > &mr);

    /// Set \ref deriv object to use to find saturation properties
    virtual int set_sat_deriv(deriv<int,funct<int> > &de);

    /** 
	\brief Set the second \ref deriv object to use to find 
	saturation properties

	Computing the slope of the symmetry energy at the saturation
	density requires two derivative objects, because it has to
	take an isospin derivative and a density derivative. Thus this
	second \ref deriv object is used in the function
	fesym_slope().
    */
    virtual int set_sat_deriv2(deriv<bool,funct<bool> > &de);
    
    /// Set neutron and proton 
    virtual int set_n_and_p(fermion &n, fermion &p);
    //@}

    /// The defaut neutron
    fermion def_neutron;

    /// The defaut proton
    fermion def_proton;

    /// \name Default solvers and derivative classes
    //@{
    /** 
	\brief The default object for derivatives
	
	The value of gsl_deriv::h is set to \f$ 10^{-3} \f$ in 
	the hadronic_eos constructor.
    */
    gsl_deriv<int,funct<int> > def_deriv;
    
    /** 
	\brief The second default object for derivatives
	
	The value of gsl_deriv::h is set to \f$ 10^{-3} \f$ in 
	the hadronic_eos constructor.
    */
    gsl_deriv<bool,funct<bool> > def_deriv2;

    /** 
	\brief The default solver

	Used by calc_e() to solve nuc_matter_p() (2 variables) and by
	calc_p() to solve nuc_matter_e() (2 variables).
    */
    gsl_mroot_hybrids<double *,mm_funct<double *> > def_mroot;
    
    /** \brief The default solver for calculating the saturation 
	density

	Used by fn0() (which is called by saturation()) to solve
	saturation_matter_e() (1 variable).
    */
    cern_mroot_root<int,funct<int> > def_sat_root;
    //@}
    
    /// Return string denoting type ("hadronic_eos")
    virtual const char *type() { return "hadronic_eos"; }

#ifndef DOXYGEN_INTERNAL

  protected:

    /// Compute t1 for gradient_qij().
    double t1_fun(double barn, int &vp);
    
    /// Compute t2 for gradient_qij().
    double t2_fun(double barn, int &vp);
    
    /// The EOS solver
    mroot<double *,mm_funct<double *> > *eos_mroot;
    
    /// The solver to compute saturation properties
    root<int,funct<int> > *sat_root;

    /// The derivative object for saturation properties
    deriv<int,funct<int> > *sat_deriv;

    /// The second derivative object for saturation properties
    deriv<bool,funct<bool> > *sat_deriv2;

    /// The neutron object
    fermion *neutron;

    /// The proton object
    fermion *proton;

    /// Temporary proton fraction
    double proton_frac;
    
    /// Temporary baryon number
    double n_baryon;
    
#endif
    
  };

  /// A hadronic EOS based on a function of the densities [abstract base]
  class hadronic_eos_eden : public hadronic_eos {
  public:

    /** 
	\brief Equation of state as a function of density
    */
    virtual int calc_e(fermion &n, fermion &p, thermo &th)=0;
    
    /** 
	\brief Equation of state as a function of the chemical potentials
    */
    virtual int calc_p(fermion &n, fermion &p, thermo &th);

  };

  /** \brief A hadronic EOS based on a function of the chemical 
      potentials [abstract base]
  */
  class hadronic_eos_pres : public hadronic_eos {
  public:

    /** 
	\brief Equation of state as a function of the chemical potentials
    */
    virtual int calc_p(fermion &n, fermion &p, thermo &th)=0;
    
    /** 
	\brief Equation of state as a function of density
    */
    virtual int calc_e(fermion &n, fermion &p, thermo &th);

  };

  /// A finite temperature hadronic EOS
  class hadronic_eos_temp : public hadronic_eos {

#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// The neutron object
    fermion_T *neutron_T;

    /// The proton object
    fermion_T *proton_T;

    /// The temperature
    double lT;

    /// Solve for nuclear matter at finite temperature given density
    int nuc_matter_temp_e(size_t nv, const ovector_base &x, 
			  ovector_base &y, double *&pa);
    
    /// Solve for nuclear matter at finite temperature given mu
    int nuc_matter_temp_p(size_t nv, const ovector_base &x, 
			  ovector_base &y, double *&pa);
    
#endif

  public:

    /// The defaut neutron
    eff_fermion def_neutron_T;

    /// The defaut proton
    eff_fermion def_proton_T;

    hadronic_eos_temp() {
      def_neutron_T.init(939.0/o2scl_const::hc_mev_fm,2.0);
      def_proton_T.init(939.0/o2scl_const::hc_mev_fm,2.0);
      def_neutron_T.non_interacting=false;
      def_proton_T.non_interacting=false;
      neutron_T=&def_neutron_T;
      proton_T=&def_proton_T;
    }

    virtual ~hadronic_eos_temp() {}

    /// Set neutron and proton 
    virtual int set_n_and_p_T(fermion_T &n, fermion_T &p);

    /** 
	\brief Equation of state as a function of density
    */
    virtual int calc_e(fermion &n, fermion &p, thermo &th)=0;
    
    /** 
	\brief Equation of state as a function of densities at 
	finite temperature
    */
    virtual int calc_temp_e(fermion_T &n, fermion_T &p, double T, 
			    thermo &th)=0;

    /** 
	\brief Equation of state as a function of the chemical potentials
    */
    virtual int calc_p(fermion &n, fermion &p, thermo &th)=0;

    /** 
	\brief Equation of state as a function of the chemical potentials
	at finite temperature
    */
    virtual int calc_temp_p(fermion_T &n, fermion_T &p, double T, 
			    thermo &th)=0;
    
  };

  /** \brief  A hadronic EOS at finite temperature
      based on a function of the densities [abstract base]
  */
  class hadronic_eos_temp_eden : public hadronic_eos_temp {
  public:

    /** 
	\brief Equation of state as a function of density
    */
    virtual int calc_e(fermion &n, fermion &p, thermo &th)=0;
    
    /** 
	\brief Equation of state as a function of densities at 
	finite temperature
    */
    virtual int calc_temp_e(fermion_T &n, fermion_T &p, double T, 
			    thermo &th)=0;

    /** 
	\brief Equation of state as a function of the chemical potentials
    */
    virtual int calc_p(fermion &n, fermion &p, thermo &th);

    /** 
	\brief Equation of state as a function of the chemical potentials
	at finite temperature
    */
    virtual int calc_temp_p(fermion_T &n, fermion_T &p, double T, 
			    thermo &th);
    
  };

  /** \brief A hadronic EOS at finite temperature based on a function
      of the chemical potentials [abstract base]
  */
  class hadronic_eos_temp_pres : public hadronic_eos_temp {
  public:

    /** 
	\brief Equation of state as a function of the chemical potentials
    */
    virtual int calc_p(fermion &n, fermion &p, thermo &th)=0;

    /** 
	\brief Equation of state as a function of the chemical potentials
	at finite temperature
    */
    virtual int calc_temp_p(fermion_T &n, fermion_T &p, double T, 
			    thermo &th)=0;
    
    /** 
	\brief Equation of state as a function of density
    */
    virtual int calc_e(fermion &n, fermion &p, thermo &th);

    /** 
	\brief Equation of state as a function of densities at 
	finite temperature
    */
    virtual int calc_temp_e(fermion_T &n, fermion_T &p, double T, 
			    thermo &th);
  };

#ifndef DOXYGENP
}
#endif

#endif

