/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_INTE_H
#define O2SCL_INTE_H

#include <cmath>
#include <o2scl/err_hnd.h>
#include <o2scl/funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Base integration class [abstract base]

      \future It might be useful to have all of the integration
      classes report the number of function evaluations used
      in addition to the number of iterations which were taken
  */
  template<class param_t, class func_t=funct<param_t> > class inte {
  
    public:
  
    inte() {
      tolf=1.0e-8;
      tolx=1.0e-8;
      verbose=0;
      interror=0.0;
      err_nonconv=true;
      last_conv=0;
    }

    virtual ~inte() {}

    /** \brief Verbosity
     */
    int verbose;

    /// The most recent number of iterations taken
    int last_iter;
  
    /** \brief The maximum relative uncertainty 
	in the value of the integral (default \f$ 10^{-8} \f$)
    */
    double tolf;
    
    /** \brief The maximum absolute uncertainty 
	in the value of the integral (default \f$ 10^{-8} \f$)
    */
    double tolx;
  
    /// If true, call the error handler if the routine does not "converge"
    bool err_nonconv;
  
    /** \brief Zero if last call to integ() or integ_err() converged
	
	This is particularly useful if err_nonconv is false to test
	if the last integration call converged.
    */
    int last_conv;
    
    /** \brief Integrate function \c func from \c a to \c b.
     */
    virtual double integ(func_t &func, double a, double b, param_t &pa)=0;
  
    /** \brief Integrate function \c func from \c a to \c b and place
	the result in \c res and the error in \c err
    */
    virtual int integ_err(func_t &func, double a, double b, 
			  param_t &pa, double &res, double &err)=0;
  
    /** \brief Return the error in the result from the last call to 
	integ()
      
	This will quietly return zero if no integrations have been
	performed or if the integrator does not compute an error.
    */
    double get_error() { return interror; }
  
    /// Return string denoting type ("inte")
    virtual const char *type() { return "inte"; }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// The uncertainty for the last integration computation
    double interror;
  
#endif
  
  };

#ifndef DOXYGENP
}
#endif

#endif




