/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_LINEAR_SOLVER_H
#define O2SCL_LINEAR_SOLVER_H

#include <gsl/gsl_linalg.h>

#include <o2scl/ovector_tlate.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/permutation.h>
#include <o2scl/lu.h>
#include <o2scl/qr.h>
#include <o2scl/hh.h>
#include <o2scl/array.h>

namespace o2scl_linalg {

  /** 
      \brief A generic solver for the linear system \f$ A x = b \f$
      [abstract base]

      A generic solver for dense linear systems.

      Those writing production level code should consider calling
      LAPACK directly using \o2 objects as described in the \ref
      linalg_section section of the User's Guide.

      \future The test code uses a Hilbert matrix, which is known
      to be ill-conditioned, especially for the larger sizes. This
      should probably be changed.
  */
  template<class vec_t=o2scl::ovector_base, class mat_t=o2scl::omatrix_base> 
    class linear_solver {
  public:

    virtual ~linear_solver() {}

    /// Solve square linear system \f$ A x = b \f$ of size \c n
    virtual int solve(size_t n, mat_t &a, vec_t &b, vec_t &x)=0;
  };

  /// Generic linear solver using LU decomposition
  template <class vec_t=o2scl::ovector_base, class mat_t=o2scl::omatrix_base> 
    class linear_solver_lu : 
  public linear_solver<vec_t, mat_t> {
    
  public:
    
    /// Solve square linear system \f$ A x = b \f$ of size \c n
    virtual int solve(size_t n, mat_t &A, 
		      vec_t &b, vec_t &x) {
      int sig;
      o2scl::permutation p(n);
      LU_decomp(n,A,p,sig);
      LU_solve(n,A,p,b,x);
      return 0;
    };
    
    virtual ~linear_solver_lu() {}
    
  };

  /// Generic linear solver using QR decomposition
  template <class vec_t=o2scl::ovector_base, class mat_t=o2scl::omatrix_base> 
    class linear_solver_qr : 
  public linear_solver<vec_t, mat_t> {
    
  public:
    
    /// Solve square linear system \f$ A x = b \f$ of size \c n
    virtual int solve(size_t n, mat_t &A, 
		      vec_t &b, vec_t &x) {
      double *tau=o2scl::new_array<double>(n);
      QR_decomp(n,n,A,tau);
      QR_solve(n,A,tau,b,x);
      o2scl::delete_array(tau);
      return 0;
    };
    
    virtual ~linear_solver_qr() {}
    
  };
  
  /// Generic Householder linear solver 
  template <class vec_t=o2scl::ovector_base, class mat_t=o2scl::omatrix_base> 
    class linear_solver_hh : 
  public linear_solver<vec_t, mat_t> {
    
  public:
    
    /// Solve square linear system \f$ A x = b \f$ of size \c n
    virtual int solve(size_t n, mat_t &A, 
		      vec_t &b, vec_t &x) {
      return HH_solve(n,A,b,x);
    };
    
    virtual ~linear_solver_hh() {}
    
  };
  
  /// GSL solver by LU decomposition
#ifdef DOXYGENP
  class gsl_solver_LU : public linear_solver 
#else
    class gsl_solver_LU : public linear_solver<o2scl::ovector_base,
    o2scl::omatrix_base> 
#endif
      {
	
      public:
	
	/// Solve square linear system \f$ A x = b \f$ of size \c n
	virtual int solve(size_t n, o2scl::omatrix_base &A, 
			  o2scl::ovector_base &b,
			  o2scl::ovector_base &x) {
	  gsl_permutation *P=gsl_permutation_alloc(n);
	  int s;
	  gsl_linalg_LU_decomp((gsl_matrix *)&A,P,&s);
	  gsl_linalg_LU_solve((gsl_matrix *)&A,P,(gsl_vector *)&b,
			      (gsl_vector *)&x);
	  gsl_permutation_free(P);
      
	  return 0;
	};
    
	virtual ~gsl_solver_LU() {}
    
      };

  /// GSL solver by QR decomposition
#ifdef DOXYGENP
  class gsl_solver_QR : public linear_solver 
#else
    class gsl_solver_QR : public linear_solver<o2scl::ovector_base,
    o2scl::omatrix_base> 
#endif
      {
    
      public:
    
	/// Solve square linear system \f$ A x = b \f$ of size \c n
	virtual int solve(size_t n, o2scl::omatrix_base &A, 
			  o2scl::ovector_base &b,
			  o2scl::ovector_base &x) {
	  o2scl::ovector tau(n);
	  gsl_linalg_QR_decomp((gsl_matrix *)&A,(gsl_vector *)&tau);
	  gsl_linalg_QR_solve((gsl_matrix *)&A,(gsl_vector *)&tau,
			      (gsl_vector *)&b,(gsl_vector *)&x);
	  return 0;
	};
    
	virtual ~gsl_solver_QR() {}
    
      };

  /// GSL Householder solver
#ifdef DOXYGENP
  class gsl_solver_HH : public linear_solver 
#else
    class gsl_solver_HH : public linear_solver<o2scl::ovector_base,
    o2scl::omatrix_base> 
#endif
      {
    
      public:
    
	/// Solve square linear system \f$ A x = b \f$ of size \c n
	virtual int solve(size_t n, o2scl::omatrix_base &A, 
			  o2scl::ovector_base &b,
			  o2scl::ovector_base &x) {
	  return gsl_linalg_HH_solve((gsl_matrix *)(&A),(gsl_vector *)(&b),
				     (gsl_vector *)(&x));
	};
    
	virtual ~gsl_solver_HH() {}
    
      };

}

#endif
