/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/test_mgr.h>
#include <o2scl/multi_funct.h>
#include <o2scl/composite_inte.h>
#include <o2scl/gsl_inte_qng.h>
#include <o2scl/gsl_vegas.h>

/// For M_PI
#include <gsl/gsl_math.h>

using namespace std;
using namespace o2scl;

double test_fun(size_t nv, const ovector_base &x, int &vp) {
  // We make a small shift by 0.1 to avoid integrands which
  // are small everywhere
  double y=1.0/(1.0-cos(x[0])*cos(x[1])*cos(x[2]))/M_PI/M_PI/M_PI+0.1;
  return y;
}

double g(double *k, size_t dim, void *params) {
  // We make a small shift by 0.1 to avoid integrands which
  // are small everywhere
  return 1.0/(1.0-cos(k[0])*cos(k[1])*cos(k[2]))/M_PI/M_PI/M_PI+0.1;
}

int main(void) {
  test_mgr t;
  t.set_output_level(1);
 
  cout.setf(ios::scientific);

  double exact = 1.3932039296;
  double res1, res2, res3;

  // GSL version
  {
    double err;

    gsl_monte_function G = {&g,3,0};

    double xl[3]={0,0,0};
    double xu[3]={M_PI,M_PI,M_PI};
    
    const gsl_rng_type *T;
    gsl_rng *r;
    
    gsl_rng_env_setup();
    
    T=gsl_rng_default;
    r=gsl_rng_alloc(T);
    
    gsl_monte_vegas_state *s=gsl_monte_vegas_alloc(3);

    s->verbose=-1;
    
    gsl_monte_vegas_integrate(&G,xl,xu,3,100000,r,s,&res1,&err);
    res1-=0.1*pow(M_PI,3.0);
    t.test_rel(res1,exact,err*10.0,"GSL");
    
    gsl_monte_vegas_free (s);
  }

  // O2SCL version
  {
    double err;

    int vpx=0;
    gsl_vegas<int ,multi_funct<int> > gm;
    ovector a(3), b(3);
    a.set_all(0.0);
    b.set_all(M_PI);

    multi_funct_fptr_noerr<int> tf(test_fun);

    //gm.verbose=3;

    gm.n_points=100000;
    gm.minteg_err(tf,3,a,b,vpx,res2,err);
    res2-=0.1*pow(M_PI,3.0);

    cout << res2 << " " << exact << " " << (res2-exact)/err << endl;
    t.test_rel(res2,exact,err*10.0,"O2SCL");
    t.test_rel(res1,res2,1.0e-9,"O2SCL vs. GSL");
  }

  // Direct integration with composite_inte
  {
    int vpx=0;

    /// The integrator
    typedef composite_inte<int ,multi_funct<int>,
      ovector_base,ovector,ovector_alloc > ci3_t;
    ci3_t ci3;
    
    /// The individual integration objects
    gsl_inte_qng<size_t,funct<size_t> > gl4a;
    gsl_inte_qng<size_t,funct<size_t> > gl4b;
    gsl_inte_qng<size_t,funct<size_t> > gl4c;
    
    gl4a.err_nonconv=false;
    gl4b.err_nonconv=false;

    /// Set the 1-d objects in the integrator
    inte<size_t,funct<size_t> > *ptrs3[3]=
      {&gl4a,&gl4b,&gl4c};
    ci3.set_oned_inte(gl4a,0);
    ci3.set_oned_inte(gl4b,1);
    ci3.set_oned_inte(gl4c,2);
  
    ovector a(3), b(3);
    a[0]=0.0;
    a[1]=0.0;
    a[2]=0.0;
    b[0]=M_PI;
    b[1]=M_PI;
    b[2]=M_PI;

    multi_funct_fptr_noerr<int> tf(test_fun);
    
    res3=ci3.minteg(tf,3,a,b,vpx);
    res3-=0.1*pow(M_PI,3.0);

    t.test_rel(res3,res2,1.0e-3,"monte carlo vs. direct");
  }

  t.report();
 
  return 0;
}


