/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_MMIN_CONP_H
#define O2SCL_GSL_MMIN_CONP_H

#include <gsl/gsl_blas.h>
#include <gsl/gsl_multimin.h>
#include <o2scl/gsl_mmin_conf.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Multidimensional minimization by the Polak-Ribiere
      conjugate gradient algorithm (GSL)

      The functions mmin() and mmin_de() minimize a given function
      until the gradientis smaller than the value of multi_min::tolf
      (which defaults to \f$ 10^{-4} \f$ ).

      See an example for the usage of this class
      in \ref ex_mmin_sect .

      \future A bit of needless copying is required in the function
      wrapper to convert from \c gsl_vector to the templated vector
      type. This can be fixed.
      
  */
  template<class param_t, class func_t=multi_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc, 
    class dfunc_t=grad_funct<param_t,ovector_base>, 
    class auto_grad_t=gradient<param_t,func_t,ovector_base>,
    class def_auto_grad_t=simple_grad<param_t,func_t,ovector_base> > 
    class gsl_mmin_conp : 
  public gsl_mmin_conf<param_t,func_t,vec_t,alloc_vec_t,alloc_t,dfunc_t,
    auto_grad_t,def_auto_grad_t> {
    
    public:
    
    /// Perform an iteration
    virtual int iterate() {
      
      if (this->dim==0) {
	O2SCL_ERR_RET("Memory not allocated in iterate().",gsl_efailed);
      }
      
      gsl_vector *x=this->ugx;
      gsl_vector *gradient=this->ugg;
      gsl_vector *dx=this->udx;
      
      double fa = this->it_min, fb, fc;
      double dir;
      double stepa = 0.0, stepb, stepc=this->step;
      
      double g1norm;
      double pg;
      
      if (this->pnorm == 0.0 || this->g0norm == 0.0) {
	gsl_vector_set_zero(dx);
	this->last_conv=gsl_enoprog;
	O2SCL_CONV2_RET("Either pnorm or g0norm vanished in ",
			"in gsl_mmin_conp::iterate().",
			  gsl_enoprog,this->err_nonconv);
	return 0;
      }
      
      /* Determine which direction is downhill, +p or -p */
      
      gsl_blas_ddot (this->p, gradient, &pg);
      
      dir = (pg >= 0.0) ? +1.0 : -1.0;
      
      /* Compute new trial point at x_c= x - step * p, where p is the
	 current direction */
      
      take_step (x, this->p, stepc, dir / this->pnorm, this->x1, dx);
      
      /* Evaluate function and gradient at new point xc */
      
      for(size_t i=0;i<this->dim;i++) {
	this->avt5[i]=gsl_vector_get(this->x1,i);
      }
      (*this->func)(this->dim,this->avt5,fc,*this->params);
      
      if (fc < fa) {

	/* Success, reduced the function value */
	this->step = stepc * 2.0;
	this->it_min = fc;
	gsl_vector_memcpy (x, this->x1);

	for(size_t i=0;i<this->dim;i++) {
	  this->avt6[i]=gsl_vector_get(this->x1,i);
	}
	if (this->grad_given) {
	  (*this->grad)(this->dim,this->avt6,this->avt7,*this->params);
	} else {
	  (*this->agrad)(this->dim,this->avt6,this->avt7,*this->params);
	}
	for(size_t i=0;i<this->dim;i++) {
	  gsl_vector_set(gradient,i,this->avt7[i]);
	}

	return gsl_success;
      }
      
      /* Do a line minimisation in the region (xa,fa) (xc,fc) to find an
	 intermediate (xb,fb) satisifying fa > fb < fc.  Choose an initial
	 xb based on parabolic interpolation */
      
      this->intermediate_point(x, this->p, dir / this->pnorm, pg,
			       stepa, stepc, fa, fc, this->x1, this->dx1,
			       gradient, &stepb, &fb);
      
      if (stepb == 0.0) {
	this->last_conv=gsl_enoprog;
	O2SCL_CONV2_RET("Variable stepb vanished in ",
			"in gsl_mmin_conp::iterate().",
			gsl_enoprog,this->err_nonconv);
      }
      
      this->minimize(x,this->p,dir / this->pnorm,stepa,stepb,stepc, fa, 
		     fb, fc, this->tol, this->x1, this->dx1, this->x2, 
		     dx, gradient, &(this->step), &(this->it_min), &g1norm);
      
      gsl_vector_memcpy (x,this->x2);
      
      /* Choose a new conjugate direction for the next step */
      
      this->iter = (this->iter + 1) % x->size;
      
      if (this->iter == 0) {
	gsl_vector_memcpy (this->p, gradient);
	this->pnorm = g1norm;
      } else {
	/* p' = g1 - beta * p */
	double g0g1, beta;
	
	/* g0' = g0 - g1 */
	gsl_blas_daxpy (-1.0, gradient, this->g0); 
	/* g1g0 = (g0-g1).g1 */
	gsl_blas_ddot(this->g0, gradient, &g0g1);  
	/* beta = -((g1 - g0).g1)/(g0.g0) */
	beta = g0g1 / (this->g0norm*this->g0norm);       
	
	gsl_blas_dscal (-beta, this->p);
	gsl_blas_daxpy (1.0, gradient, this->p);
	this->pnorm = gsl_blas_dnrm2 (this->p);
      }
      
      this->g0norm = g1norm;
      gsl_vector_memcpy (this->g0, gradient);
      
      return gsl_success;
    }
    
    /// Return string denoting type("gsl_mmin_conp")
    virtual const char *type() { return "gsl_mmin_conp";}
    
  };

  /**
     \brief An array version of gsl_mmin_conp
     
     \comment
     Doxygen has had trouble parsing this template, so I have
     simplified it slightly for the documentation below.
     \endcomment
  */
#ifdef DOXYGENP
  template<class param_t, size_t nv> class gsl_mmin_conp_array :
  public gsl_mmin_conp<param_t,func_t,vec_t,alloc_vec_t,
    alloc_t,dfunc_t,auto_grad_t,def_auto_grad_t> {
  };
#else
  template<class param_t, size_t nv> class gsl_mmin_conp_array :
  public gsl_mmin_conp<param_t,multi_vfunct<param_t,nv>,
    double[nv],double[nv],array_alloc<double[nv]>,grad_vfunct<param_t,nv>,
    gradient_array<param_t,multi_vfunct<param_t,nv>,nv>,
    simple_grad_array<param_t,multi_vfunct<param_t,nv>,nv> > {
  };
#endif
  
  
#ifndef DOXYGENP
}
#endif

#endif
