/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/* multimin/simplex2.c
 * 
 * Copyright (C) 2007 Brian Gough
 * Copyright (C) 2002 Tuomo Keskitalo, Ivo Alxneit
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */
#ifndef O2SCL_GSL_MMIN_SIMP2_H
#define O2SCL_GSL_MMIN_SIMP2_H

#include <gsl/gsl_multimin.h>
#include <o2scl/multi_min.h>
#include <o2scl/cblas.h>
#include <o2scl/vec_stats.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Multidimensional minimization by the Simplex method (v2) (GSL)

      This class minimizes a function using Nelder and Mead's Simplex
      algorithm. A simplex in a N-dimensional space is definted as a
      set of N+1 points which describe an N-dimensional volume
      surrounding the minimum. The algorithm proceeds by shifting
      the simplex points until the simplex is sufficiently small and
      thus the minimum is known with sufficient accuracy. 

      For the earlier method used in GSL, see \ref gsl_mmin_simp .

      This class has a high-level interface using mmin(),
      mmin_twovec() or mmin_simplex() which automatically performs the
      memory allocation and minimization, or a GSL-like interface
      using allocate(), free(), interate() and set() or set_simplex().

      The simplex can be completely specified by the user (see
      mmin_simplex() and set_simplex()). Alternatively, the simplex is
      automatically specified given initial guess \f$ x_j \f$ and a
      step size vector \f$ s_k \f$ for \f$ 0\leq k<n_s \f$. The
      simplex \f$ p_{ij} \f$ with \f$ 0\leq i\leq n \f$ and \f$ 0\leq
      j<n \f$ is chosen with \f$ p_{0j} = x_j \f$ and
      \f{eqnarray*}
      p_{i+1,j} &=& x_j\quad\mathrm{for}\quad i\neq j \\
      p_{i+1,j} &=& x_j+s_{j~\mathrm{mod}~n_s}\quad\mathrm{for}\quad i=j
      \f}
      for \f$ 0<i<n \f$. The step size vector \f$ s \f$ is set by the
      set_step() member function. The prsence of \f$ \mathrm{mod} \f$
      in the recipe above just indicates that elements of the step
      size vector are automatically re-used if there are less step
      sizes than dimensions in the minimization problem.

      See an example for the usage of this class
      in \ref ex_mmin_sect .

      Based on \ref Nelder65 .

      \note I have had problems with this algorithm not converging,
      and these difficulties appear to persist in the original GSL
      version. The original simplex algorithm in the class \ref
      gsl_mmin_simp will sometimes converge in these cases.
      
  */
#ifdef DOXYGENP
  template<class param_t,class func_t=multi_funct<param_t>,
    class vec_t=ovector_base, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc> class gsl_mmin_simp2 :
  public multi_min
#else
    template<class param_t,class func_t=multi_funct<param_t>,
    class vec_t=ovector_base, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc> class gsl_mmin_simp2 :
  public multi_min<param_t,func_t,func_t,vec_t>
#endif
    {
    
#ifndef DOXYGEN_INTERNAL
    
      protected:

      /// An array of n+1 vectors containing the simplex
      alloc_vec_t *x1;
      /**
	 \brief The n+1 function values at the simplex points

	 \comment
	 This can't be the template type because it has to be 
	 a different size (n+1) rather than n.
	 \endcomment
      */
      ovector y1;
      /// Workspace vector 1
      alloc_vec_t ws1;
      /// Workspace vector 2
      alloc_vec_t ws2;
      /// Workspace vector 3
      alloc_vec_t ws3;
      /// Center of simplex
      alloc_vec_t center;
      /// Desc
      alloc_vec_t delta;
      /// Distance of vector from center
      alloc_vec_t xmc;
      /// Squared simplex size
      double S2;
      /// Desc
      long count;
    
      /// Compute the center of the simplex 
      int compute_center() {
	size_t i, j;
	double val;
      
	for (j = 0; j < dim; j++) {
	  val = 0.0;
	  for (i = 0; i < dim+1; i++) {
	    val += x1[i][j];
	  }
	  val /= dim+1;
	  center[j]=val;
	}
      
	return 0;
      }
    
      /** 
	  \brief Compute the size of the simplex
	
	  Calculates simplex size as average sum of length of vectors
	  from simplex center to corner points:
	
	  \f$ (1/n) \sum || y - y_{\mathrm{middlepoint}} || \f$
      */
      double compute_size() {
	
	size_t P=dim+1;
	double ss=0.0;
	
	for(size_t i=0;i<P;i++) {
	  for(size_t j=0;j<dim;j++) ws1[j]=x1[i][j];
	  o2scl_cblas::daxpy(dim,-1.0,center,ws1);
	  double t=o2scl_cblas::dnrm2(dim,ws1);
	  ss += t*t;
	}
	
	/* Store squared size in the state */
	S2=ss/P;
	return sqrt(ss/(double)(P));
      }

      /** 
	  \brief Move a corner of a simplex

	  Moves a simplex corner scaled by coeff (negative value
	  represents mirroring by the middle point of the "other" corner
	  points) and gives new corner in xc and function value at xc 
	  in \c newval.
      */
      virtual int try_corner_move(const double coeff, size_t corner, 
				  vec_t &xc, func_t &f, size_t nvar, 
				  param_t &pa, double &newval) {
	
	size_t P=dim+1;
	
	/* 
	   xc = (1-coeff)*(P/(P-1)) * center(all) + 
	   ((P*coeff-1)/(P-1))*x_corner 
	*/
	double alpha=(1.0-coeff)*P/dim;
	double beta=(P*coeff-1.0)/dim;
	
	for(size_t j=0;j<dim;j++) {
	  xc[j]=center[j]*alpha;
	  xc[j]+=x1[corner][j]*beta;
	}
	
	return f(nvar,xc,newval,pa);
      }

      /// Desc
      virtual int update_point(size_t i, vec_t &xx, double val) {

	const size_t P=dim+1;

	/* Compute delta = x - x_orig */
	for(size_t j=0;j<dim;j++) {
	  delta[j]=xx[j];
	  delta[j]-=x1[i][j];
	}

	/* Compute xmc = x_orig - c */
	for(size_t j=0;j<dim;j++) {
	  xmc[j]=x1[i][j];
	  xmc[j]-=center[j];
	}
	
	/* Update size: S2' = S2 + (2/P) * 
	   (x_orig - c).delta + (P-1)*(delta/P)^2 
	*/
	double d=o2scl_cblas::dnrm2(dim,delta);
	double xmcd=o2scl_cblas::ddot(dim,xmc,delta);
	S2 += (2.0 / P) * xmcd + ((P - 1.0) / P) * (d * d / P);

	/* Update center:  c' = c + (x - x_orig) / P */
	double alpha=1.0/P;
	for(size_t j=0;j<dim;j++) {
	  center[j]-=alpha*x1[i][j];
	  center[j]+=alpha*xx[j];
	}
	
	for(size_t j=0;j<dim;j++) {
	  x1[i][j]=xx[j];
	}
	y1[i]=val;

	return 0;
      }
      
      /** 
	  \brief Contract the simplex towards the best point
	  
	  Function contracts the simplex in respect to best valued
	  corner. All corners besides the best corner are moved.
	
	  \comment 
	  The GSL version requires a vector for workspace
	  named 'xc' which is not required here.
	  \endcomment
      */
      virtual int contract_by_best(size_t best, func_t &f, 
				   size_t nvar, param_t &pa) {
	
	/* Function contracts the simplex in respect to best valued
	   corner. That is, all corners besides the best corner are
	   moved. (This function is rarely called in practice, since
	   it is the last choice, hence not optimised - BJG) 
	*/
	size_t i,j, it;
	double newval;
	bool failed;

	int status=gsl_success;
	
	for (i=0;i<dim+1;i++) {
	
	  if (i!=best) {
	  
	    for (j=0;j<dim;j++) {
	      x1[i][j]=0.5*(x1[i][j]+x1[best][j]);
	    }
	    int ret=f(nvar,x1[i],y1[i],pa);
	    if (!finite(y1[i])) {
	      O2SCL_ERR2_RET("Function not finite in ",
			     "gsl_mmin_simp2::contract_by_best().",
			     gsl_ebadfunc);
	    }

	    if (avoid_nonzero==true && ret!=0) {
	      std::cout << "Found problem in contract." << std::endl;

	      // copy the old point to ws3
	      for (j=0;j<dim;j++) {
		ws3[j]=2.0*x1[i][j]-x1[best][j];
	      }
	    
	      // Try (21*best+20*xold)/41, (22*best+19*xold)/41, ...,
	      // (40*best+xold)/41 to see if we can find a contraction
	      // that works
	      for(it=0;ret!=0 && it<20;it++) {
		for (j=0;j<dim;j++) {
		  x1[i][j]=((20-it)*ws3[j]+(it+21)*x1[best][j])/41.0;
		}
		ret=f(nvar,x1[i],y1[i],pa);
		std::cout << "it, ret, x: " << it << " " << ret << " "
			  << x << std::endl;
	      }
	      if (ret!=0) {
		this->last_conv=gsl_efailed;
		O2SCL_CONV2_RET("Failed to find suitable contraction ",
				"in gsl_mmin_simp2::contract_by_best().",
				gsl_efailed,this->err_nonconv);
	      }
	    }
	  }
	}
	
	compute_center();
	size=compute_size();
	
	return gsl_success;
      }

      /// Number of variables to be minimized over
      size_t dim;

      /// Function
      func_t *func;

      /// Parameters
      param_t *params;

      /// True if set() has been called
      bool set_called;

      /// Vector of step sizes
      ovector step_vec;

      /// Vector allocator
      alloc_t ao;

      /**
	 \brief If true, try to automatically avoid regions where the 
	 function returns a non-zero value (default false)

	 \note This option doesn't work yet, so I've made the variable
	 protected to prevent the user from changing it.
      */
      bool avoid_nonzero;

#endif

      public:

      gsl_mmin_simp2() {
	dim=0;
	print_simplex=0;
	step_vec.allocate(1);
	step_vec[0]=1.0;
	avoid_nonzero=false;
	count=0;
      }
    
      virtual ~gsl_mmin_simp2() {
	free();
	step_vec.free();
      }

      /// Set the step sizes for each independent variable
      template<class vec2_t> int set_step(size_t nv, vec2_t &step) {
	if (nv>0) {
	  step_vec.free();
	  step_vec.allocate(nv);
	  for(size_t i=0;i<nv;i++) step_vec[i]=step[i];
	}
	return 0;
      }

      /// Size of current simplex computed by iterate()
      double size;

      /// Present minimum vector computed by iterate()
      alloc_vec_t x;

      /// Function value at minimum computed by iterate()
      double fval;

      /** 
	  \brief Print simplex information in print_iter() (default 0)
	
	  If this is 1 and \ref verbose is greater than 0, 
	  then print_iter() will print the function values at all 
	  the simplex points. If this is 2 and \ref verbose is
	  greater than 0, then print_iter() will print the 
	  simplex coordinates in addition to the function values.
      */
      int print_simplex;

      /** \brief Calculate the minimum \c min of \c func w.r.t the
	  array \c x of size \c nvar.
      */
      virtual int mmin(size_t nn, vec_t &xx, double &fmin, param_t &pa,
		       func_t &ufunc) {

	if (nn==0) {
	  O2SCL_ERR2_RET("Tried to minimize over zero variables ",
			 " in gsl_mmin_simp2::mmin().",gsl_einval);
	}

	this->last_conv=0;
	int ret=0,status,iter=0;
      
	allocate(nn);

	alloc_vec_t ss;
	ao.allocate(ss,nn);
	for (size_t is=0;is<nn;is++) ss[is]=step_vec[is % step_vec.size()];
	ret=set(ufunc,pa,nn,xx,ss);
	ao.free(ss);

	if(ret!=0) {
	  free();
	  return ret;
	}
  
	do {
	  iter++;
	  
	  status=iterate();
	  if(status) break;
	  
	  if(this->verbose>0) {
	    print_iter(nn,x,x1,fval,iter,size,this->tolx,
		       "gsl_mmin_simp2");
	  }
    
	  status=gsl_multimin_test_size(size,this->tolx);
	  
	} while(status == GSL_CONTINUE && iter<this->ntrial);
	
	for (size_t i=0;i<nn;i++) xx[i]=x[i];
	fmin=fval;
  
	free();
	this->last_ntrial=iter;
	
	if(iter>=this->ntrial) {
	  this->last_conv=gsl_emaxiter;
	  O2SCL_CONV_RET("Exceeded max # of iterations in mmin_simp2::mmin().",
			 gsl_emaxiter,this->err_nonconv);
	}

	return status;
      }
    
      /** \brief Calculate the minimum \c min of \c func w.r.t the
	  array \c x of size \c nvar, using \c xx and \c xx2 to specify
	  the simplex
      */
      virtual int mmin_twovec(size_t nn, vec_t &xx, vec_t &xx2, double &fmin, 
			      param_t &pa, func_t &ufunc) {
      
	this->last_conv=0;
	int ret=0,i,status,iter=0;
      
	allocate(nn);

	alloc_vec_t ss;
	ao.allocate(ss,nn);

	for (size_t is=0;is<nn;is++) ss[is]=xx2[is]-xx[is];
	ret=set(ufunc,pa,nn,xx,ss);
	ao.free(ss);
      
	if(ret!=0) {
	  free();
	  return ret;
	}
  
	do {
	  iter++;
	  
	  status=iterate();
	  if(status) break;

	  if(this->verbose>0) {
	    print_iter(nn,x,x1,fval,iter,size,this->tolx,
		       "gsl_mmin_simp2");
	  }
	
	  status=gsl_multimin_test_size(size,this->tolx);
	  
	} while(status == GSL_CONTINUE && iter<this->ntrial);
	
	for (i=0;i<((int)nn);i++) xx[i]=x[i];
	fmin=fval;
  
	free();
	this->last_ntrial=iter;

	if(iter>=this->ntrial) {
	  this->last_conv=gsl_emaxiter;
	  O2SCL_CONV2_RET("Exceeded max number of iterations in ",
			  "gsl_mmin_simp2::mmin_twovec().",
			  gsl_emaxiter,this->err_nonconv);
	}

	return status;
      }
    
      /** \brief Calculate the minimum \c min of \c func w.r.t the
	  array \c x of size \c nvar, given an initial simplex
      */
      template<class mat_t> 
      int mmin_simplex(size_t nn, mat_t &sx, double &fmin, 
		       param_t &pa, func_t &ufunc) {
      
	int ret=0,i,status,iter=0;
      
	allocate(nn);
	
	ret=set_simplex(ufunc,pa,sx);
	if(ret!=0) {
	  free();
	  return ret;
	}
  
	do {
	  iter++;
	  
	  status=iterate();
	  if(status) break;

	  if(this->verbose>0) {
	    print_iter(nn,x,x1,fval,iter,size,this->tolx,
		       "gsl_mmin_simp2");
	  }
    
	  status=gsl_multimin_test_size(size,this->tolx);
	  
	} while(status == GSL_CONTINUE && iter<this->ntrial);
	
	for (i=0;i<((int)nn);i++) sx[0][i]=x[i];
	fmin=fval;
  
	free();
	this->last_ntrial=iter;

	if(iter>=this->ntrial) {
	  O2SCL_ERR_RET
	    ("Exceeded max # of iterations in gsl_mmin_simp2::mmin_twovec().",
	     gsl_emaxiter);
	}

	return status;
      }
    
      /// Allocate the memory
      virtual int allocate(size_t n) {
	
	if(dim!=0) free();
	set_called=false;
	dim=n;

	count=0;
	
	ao.allocate(x,n);
	x1=new alloc_vec_t[n+1];
	for(size_t i=0;i<n+1;i++) ao.allocate(x1[i],n);
	y1.allocate(n+1);
	ao.allocate(ws1,n);
	ao.allocate(ws2,n);
	ao.allocate(ws3,n);
	ao.allocate(center,n);
	ao.allocate(delta,n);
	ao.allocate(xmc,n);

	return gsl_success;
      }
    
      /// Free the allocated memory
      virtual int free() {

	if (dim>0) {
	  for(size_t i=0;i<dim+1;i++) ao.free(x1[i]);
	  delete[] x1;
	  y1.free();
	  ao.free(ws1);
	  ao.free(ws2);
	  ao.free(ws3);
	  ao.free(center);
	  ao.free(delta);
	  ao.free(xmc);
	  ao.free(x);
	}

	dim=0;

	return 0;
      }

      /// Set the function and initial guess
      virtual int set(func_t &ufunc, param_t &pa, size_t n, vec_t &ax, 
		      vec_t &step_size) {
	size_t i;
      
	if(dim!=n) allocate(n);
	params=&pa;
	func=&ufunc;

	// Copy initial guess to x
	for (i=0;i<dim;i++) x[i]=ax[i];
      
	// first point is the original x0 
      
	int ret=ufunc(dim,ax,y1[0],pa);
	if (!finite(y1[0])) {
	  O2SCL_ERR2_RET("Function not finite in ",
			 "gsl_mmin_simp2::set().",gsl_ebadfunc);
	}
	for(i=0;i<dim;i++) x1[0][i]=ax[i];
  
	/* following points are initialized to x0+step_size */
      
	for (i=1;i<dim+1;i++) {
	  for(size_t j=0;j<dim;j++) x1[i][j]=x[j];
	  x1[i][i-1]=x1[i][i-1]+step_size[i-1];
	  ufunc(dim,x1[i],y1[i],pa);
	}
 
	/* Initialize simplex size */
      
	compute_center();
	size=compute_size();
	
	set_called=true;
	count++;

	return gsl_success;
      }

      /// Set the function and initial simplex
      template<class mat_t> 
      int set_simplex(func_t &ufunc, param_t &pa, mat_t &sx) {

	if(dim==0) {
	  O2SCL_ERR_RET("Memory not allocated in gsl_mmin_simp2::set().",
			gsl_ebadlen);
	}
      
	params=&pa;
	func=&ufunc;

	for(size_t i=0;i<dim+1;i++) {
	  for(size_t j=0;j<dim;j++) {
	    x1[i][j]=sx[i][j];
	  }
	  ufunc(dim,x1[i],y1[i],pa);
	  if (!finite(y1[i])) {
	    O2SCL_ERR2_RET("Function not finite in ",
			   "gsl_mmin_simp2::set_simplex().",gsl_ebadfunc);
	  }
	}
	
	/* Initialize simplex size */
	
	compute_center();
	size=compute_size();
	
	set_called=true;
 
	return gsl_success;
      }

      /// Perform an iteration
      virtual int iterate() {

	size_t n=dim+1;
	size_t i;
	size_t lo, hi, s_hi;
	double dhi,ds_hi,dlo;
	int status;
	double val,val2;
 
	/* get index of highest, second highest and lowest point */

	lo=hi=0;
	dlo=dhi=y1[0];
	s_hi=1;
	ds_hi=y1[1];
      
	for (i=1;i<n;i++) {
	  val=y1[i];
	  if(val<dlo) {
	    dlo=val;
	    lo=i;
	  } else if (val > dhi) {
	    ds_hi=dhi;
	    s_hi=hi;
	    dhi=val;
	    hi=i;
	  } else if(val > ds_hi) {
	    ds_hi=val;
	    s_hi=i;
	  }
	}

	/* reflect the highest value */
	
	int ret1=try_corner_move(-1.0,hi,ws1,*func,dim,*params,val);

	if (avoid_nonzero && ret1!=0) {
	  std::cout << "Found problem move1: " << std::endl;
	  for (size_t it=0;it<20 && ret1!=0;it++) {
	    ret1=try_corner_move(-1.0+((double)it)/10.0,hi,ws1,
				 *func,dim,*params,val);
	    std::cout << "it,ret: " << it << " " << ret1 << std::endl;
	  }
	  if (ret1!=0) {
	    O2SCL_ERR2_RET("Failed to move corner (1) in ",
			   "gsl_mmin_simp2::iterate().",gsl_efailed);
	  }
	}
      
	if (gsl_finite(val) && val<y1[lo]) {

	  /* reflected point becomes lowest point,try expansion */
	
	  int ret2=try_corner_move(-2.0,hi,ws2,*func,dim,*params,val2);

	  if (avoid_nonzero && ret2!=0) {
	    std::cout << "Found problem move2: " << std::endl;
	    for (size_t it=0;it<20 && ret2!=0;it++) {
	      ret2=try_corner_move(-2.0+((double)it)/6.7,hi,ws2,
				   *func,dim,*params,val2);
	      std::cout << "it,ret: " << it << " " << ret2 << std::endl;
	    }
	    if (ret2!=0) {
	      O2SCL_ERR2_RET("Failed to move corner (2) in ",
			     "gsl_mmin_simp2::iterate().",gsl_efailed);
	    }
	  }

	  if (gsl_finite(val2) && val2<y1[lo]) {
	    update_point(hi,ws2,val2);
	  } else {
	    update_point(hi,ws1,val);
	  }
	  
	} else if (!gsl_finite(val) || val > y1[s_hi]) {
	
	  /* reflection does not improve things enough */
	
	  if (gsl_finite(val) && val <= y1[hi]) {
	    
	    /* if trial point is better than highest point,replace
	       highest point */
	    
	    update_point(hi,ws1,val);
	  }
      
	  /* try one dimensional contraction */
	
	  int ret3=try_corner_move(0.5,hi,ws2,*func,dim,*params,val2);

	  if (avoid_nonzero && ret3!=0) {
	    std::cout << "Found problem move3: " << std::endl;
	    for (size_t it=0;it<20 && ret3!=0;it++) {
	      ret3=try_corner_move(0.025*((double)(19-it)),hi,ws2,
				   *func,dim,*params,val2);
	      std::cout << "it,ret: " << it << " " << ret3 << std::endl;
	    }
	    if (ret3!=0) {
	      O2SCL_ERR2_RET("Failed to move corner (2) in ",
			     "gsl_mmin_simp2::iterate().",gsl_efailed);
	    }
	  }
	  
	  if (gsl_finite(val2) && val2 <= y1[hi]) {

	    update_point(hi,ws2,val2);

	  } else {

	    /* contract the whole simplex in respect to the best point */
	    status=contract_by_best(lo,*func,dim,*params);
	    if(status != 0) {
	      O2SCL_ERR("Function contract_by_best() failed in iterate().",
			gsl_efailed);
	    }
	    
	  }

	} else {

	  /* trial point is better than second highest point.
	     Replace highest point by it */
	  
	  update_point(hi,ws1,val);
	}
  
	/* return lowest point of simplex as x */
      
	lo=vector_min_index(dim+1,y1,val);
	for(i=0;i<dim;i++) x[i]=x1[lo][i];
	fval=y1[lo];
      
	/* Update simplex size */
	
	if (S2 > 0) {
	  size=sqrt(S2);
	} else {
	  /* recompute if accumulated error has made size invalid */
	  size=compute_size();
	}

	return gsl_success;
      }
      
      /** 
	  \brief Print out iteration information.
	  
	  Depending on the value of the variable verbose, this prints out
	  the iteration information. If verbose=0, then no information is
	  printed, while if verbose>1, then after each iteration, the
	  present values of x and y are output to std::cout along with the
	  iteration number. If verbose>=2 then each iteration waits for a
	  character.
      */
      virtual int print_iter(size_t nv, vec_t &xx, alloc_vec_t *simp,
			     double y, int iter,
			     double value, double limit,
			     std::string comment) {
	
	if (this->verbose<=0) return 0;
	
	size_t i;
	char ch;
      
	(*this->outs) << comment << " Iteration: " << iter << std::endl;
	text_out_file outsf(this->outs,79);
	outsf.word_out("x:");
	for(i=0;i<nv;i++) outsf.double_out(xx[i]);
	outsf.end_line();
	if (print_simplex>0) {
	  if (print_simplex==2) {
	    (*this->outs) << "Simplex Values:" << std::endl;
	    for(i=0;i<nv+1;i++) {
	      outsf.int_out(i);
	      outsf.char_out(':');
	      for(size_t j=0;j<nv;j++) {
		outsf.double_out(simp[i][j]);
	      }
	      outsf.char_out(':');
	      outsf.double_out(y1[i]);
	      outsf.end_line();
	    }
	  } else {
	    outsf.word_out("Simplex Values:");
	    for(i=0;i<nv+1;i++) {
	      outsf.double_out(y1[i]);
	    }
	    outsf.end_line();
	  }
	}
	(*this->outs) << "y: " << y << " Val: " << value << " Lim: " 
	<< limit << std::endl;
	if (this->verbose>1) {
	  (*this->outs) << "Press a key and type enter to continue. ";
	  (*this->ins) >> ch;
	}
	
	return 0;
      }

      /// Return string denoting type("gsl_mmin_simp2")
      virtual const char *type() { return "gsl_mmin_simp2";}

    };
  
#ifndef DOXYGENP
}
#endif

#endif
