/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_NONADAPT_STEP_H
#define O2SCL_NONADAPT_STEP_H

#include <o2scl/adapt_step.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_odeiv.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief An non-adaptive stepper implementation of \ref adapt_step.

      This class simply calls the specified ODE stepper without any
      attempt to modify the size of the step. It is primarily useful
      to allow for simple comparisons between adaptive and
      non-adaptive solution or to proceed with a solution in places
      where an adaptive stepper is failing.

      To modify the ODE stepper which is used, use the function
      adapt_step::set_step().
  */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class nonadapt_step : 
  public adapt_step<param_t,func_t,vec_t,alloc_vec_t,alloc_t> {

#ifndef DOXYGEN_INTERNAL

    protected:

    /// The allocated vector size
    size_t msize;

    /** 
	\brief Internal storage for dydx
      
	\comment
	This isn't named dydx so that we can name function arguments
	dydx in astep_derivs().
	\endcomment
    */
    alloc_vec_t dydx_int;

    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;
      
#endif
    
    public:
    
    nonadapt_step()  {
      msize=0;
    }
      
    virtual ~nonadapt_step() {
      if (msize>0) {
	ao.free(dydx_int);
      }
    }

    /** 
	\brief Make an adaptive integration step of the system 
	\c derivs 
	  
	This attempts to take a step of size \c h from the point \c x of
	an \c n-dimensional system \c derivs starting with \c y. On
	exit, \c x and \c y contain the new values at the end of the
	step, \c h contains the size of the step, \c dydx_out contains
	the derivative at the end of the step, and \c yerr contains the
	estimated error at the end of the step.

	If the base stepper returns a non-zero value, that value is
	passed on as the return value of <tt>astep()</tt>, though the
	input \c y may still have been modified by the base stepper.

    */
    virtual int astep(double &x, double &h, double xlimit,
		      size_t n, vec_t &y, vec_t &u_dydx_out,
		      vec_t &yerr, param_t &pa, 
		      func_t &derivs) {

      if (n!=msize) {
	ao.allocate(dydx_int,n);
      }
      
      derivs(x,n,y,dydx_int,pa);
      if (h>0) {
	if (x+h>xlimit) h=xlimit-x;
      } else {
	if (x+h<xlimit) h=xlimit-x;
      }
      int ret=this->stepp->step(x,h,n,y,dydx_int,y,yerr,
				u_dydx_out,pa,derivs);
      x+=h;

      return ret;
    }
      
    /** \brief Make an adaptive integration step of the system 
	\c derivs

	This function performs an adaptive integration step with the
	\c n-dimensional system \c derivs and parameter \c pa. It
	Begins at \c x with initial stepsize \c h, ensuring that the
	step goes no farther than \c xlimit. At the end of the step,
	the size of the step taken is \c h and the new value of \c x
	is in \c x_out. Initially, the function values and
	derivatives should be specified in \c y and \c dydx. The
	function values, derivatives, and the error at the end of
	the step are given in \c yout, \c yerr, and \c dydx_out.
	Unlike in \c odestep objects, the objects \c y, \c yout, 
	\c dydx, and \c dydx_out must all be distinct.

	If the base stepper returns a non-zero value, that value is
	passed on as the return value of <tt>astep()</tt>, though the
	output parameters may still have been modified by the
	base stepper.
    */
    virtual int astep_full(double x, double &h, double xlimit, double &x_out,
			   size_t n, vec_t &y, vec_t &dydx, 
			   vec_t &yout2, vec_t &yerr, vec_t &dydx_out, 
			   param_t &pa, func_t &derivs) {
    
      if (h>0) {
	if (x+h>xlimit) h=xlimit-x;
      } else {
	if (x+h<xlimit) h=xlimit-x;
      }
      int ret=this->stepp->step(x,h,n,y,dydx,yout2,yerr,dydx_out,pa,derivs);
      x_out=x+h;
    
      return ret;
    }

    /** 
	\brief Make an adaptive integration step of the system 
	\c derivs with derivatives
	  
	This attempts to take a step of size \c h from the point \c x of
	an \c n-dimensional system \c derivs starting with \c y and
	given the initial derivatives \c dydx. On exit, \c x, \c y and
	\c dydx contain the new values at the end of the step, \c h
	contains the size of the step, \c dydx contains the derivative
	at the end of the step, and \c yerr contains the estimated error
	at the end of the step.

	If the base stepper returns a non-zero value, that value is
	passed on as the return value of <tt>astep()</tt>, though the
	inputs \c y and \c dydx may still have been modified by the
	base stepper.
    */
    virtual int astep_derivs(double &x, double &h, double xlimit,
			     size_t n, vec_t &y, vec_t &dydx, 
			     vec_t &yerr, param_t &pa, 
			     func_t &derivs) {
      
      if (h>0) {
	if (x+h>xlimit) h=xlimit-x;
      } else {
	if (x+h<xlimit) h=xlimit-x;
      }
      int ret=this->stepp->step(x,h,n,y,dydx,y,yerr,dydx,pa,derivs);
      x+=h;

      return ret;
    }
  
  };
  
#ifndef DOXYGENP
}
#endif

#endif
