/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_BV_SOLVE_H
#define O2SCL_ODE_BV_SOLVE_H
#include <string>
#include <o2scl/collection.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/adapt_step.h>
#include <o2scl/gsl_astep.h>
#include <o2scl/ode_iv_solve.h>
#include <o2scl/gsl_mroot_hybrids.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Solve boundary-value ODE problems with constant 
      boundary conditions [abstract base]
  */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_base, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc,
    class vec_int_t=ovector_int_base> class ode_bv_solve {

    public:

    ode_bv_solve() {
      verbose=0;
    }

    virtual ~ode_bv_solve() {}
    
    /** 
	\brief Solve the boundary-value problem
    */
    virtual int solve(double x0, double x1, double h, size_t n,
		      vec_t &ystart, vec_t &yend, vec_int_t &index,
		      param_t &pa, func_t &derivs)=0;
    
    /** \name Values for the index array */
    //@{
    /// Unknown on both the left and right boundaries
    static const int unk=0;
    /// Known on the right boundary
    static const int right=1;
    /// Known on the left boundary
    static const int left=2;
    /// Known on both the left and right boundaries
    static const int both=3;
    //@}

    /// Set output level
    int verbose;
    
  };
  
  /** 
      \brief Solve boundary-value ODE problems by shooting

      \future Create a solution table as in ode_iv_solve
  */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc, class vec_int_t=ovector_int_base> 
    class ode_bv_shoot : 
  public ode_bv_solve<param_t,func_t,vec_t,alloc_vec_t,alloc_t,vec_int_t> {

    public:
    
    ode_bv_shoot() {
      oisp=&def_ois;
      mrootp=&def_mroot;
    }
    
    virtual ~ode_bv_shoot() {}
    
    /** 
	\brief Solve the boundary-value problem
    */
    virtual int solve(double x0, double x1, double h, size_t n,
		      vec_t &ystart, vec_t &yend, vec_int_t &index,
		      param_t &pa, func_t &derivs) {
      
      // Make copies of the inputs for later access
      this->l_index=&index;
      this->l_ystart=&ystart;
      this->l_yend=&yend;
      this->l_x0=x0;
      this->l_x1=x1;
      this->l_h=h;
      this->l_derivs=&derivs;
      this->l_n=n;
  
      int nsolve=0, nsolve2=0, j;

      // Count the number of variables we'll need to solve for:
      for (size_t i=0;i<n;i++) {
	if (index[i]<2) nsolve++;
	if (index[i]%2==1) nsolve2++;
      }

      // Make sure that the boundary conditions make sense:
      if (nsolve!=nsolve2) {
	O2SCL_ERR2_RET("Incorrect boundary conditions in ",
		      "ode_bv_shoot::solve()",gsl_einval);
      } 

      // Make space for the solution
      ovector sx(nsolve), sy(nsolve);

      // Copy initial guesses from ystart
      j=0;
      for (size_t i=0;i<n;i++) {
	if (index[i]<2) {
	  sx[j]=ystart[i];
	  j++;
	}
      }
  
      // Solve
      mm_funct_mfptr<ode_bv_shoot<param_t,func_t,vec_t,alloc_vec_t,
      alloc_t,vec_int_t>,param_t > 
      mfm(this,&ode_bv_shoot<param_t,func_t,vec_t,
	  alloc_vec_t,alloc_t,vec_int_t>::solve_fun);
      int ret=this->mrootp->msolve(nsolve,sx,pa,mfm);
      solve_fun(nsolve,sx,sy,pa);

      // Copy the solution back to ystart and yend
      j=0;
      for (size_t i=0;i<n;i++) {
	if (index[i]<2) {
	  ystart[i]=sx[j];
	  j++;
	}
      }
  
      if (ret!=0) {
	O2SCL_ERR("Solver failed in ode_bv_shoot::solve().",ret);
      }
      return ret;
    }
    
    /** 
	\brief Set initial value solver
    */
    int set_iv(ode_iv_solve<param_t,func_t,vec_t,alloc_vec_t,alloc_t> &ois) {
      oisp=&ois;
      return 0;
    }
    
    /** \brief Set the equation solver */
    int set_mroot(mroot<param_t,mm_funct<param_t> > &root) {
      mrootp=&root;
      return 0;
    }

    /// The default initial value solver
    ode_iv_solve<param_t,func_t,vec_t,alloc_vec_t,alloc_t> def_ois;

    /// The default equation solver
    gsl_mroot_hybrids<param_t,mm_funct<param_t> > def_mroot;

#ifndef DOXYGEN_INTERNAL

    protected:
    
    /// The solver for the initial value problem
    ode_iv_solve<param_t,func_t,vec_t,alloc_vec_t,alloc_t> *oisp;

    /// The equation solver
    mroot<param_t,mm_funct<param_t> > *mrootp;

    /// The index defining the boundary conditions
    vec_int_t *l_index;

    /// Storage for the starting vector
    vec_t *l_ystart;

    /// Storage for the ending vector
    vec_t *l_yend;

    /// Storage for the starting point
    double l_x0;

    /// Storage for the ending abcissa
    double l_x1;

    /// Storage for the stepsize
    double l_h;

    /// The functions to integrate
    func_t *l_derivs;

    /// The number of functions
    size_t l_n;
    
    /// The shooting function to be solved by the multidimensional solver
    int solve_fun(size_t nv, const vec_t &sx, vec_t &sy, 
		  param_t &pa) {
      int j;
      ovector y(this->l_n), y2(this->l_n);
      
      // Create leftmost point from combination of 
      // starting array and proposed solution
      j=0;
      for(size_t i=0;i<this->l_n;i++) {
	if ((*this->l_index)[i]<2) {
	  y[i]=sx[j];
	  j++;
	} else {
	  y[i]=(*this->l_ystart)[i];
	}
      }
  
      // Shoot across
      this->oisp->solve_final_value(this->l_x0,this->l_x1,this->l_h,
				    this->l_n,y,y2,pa,*this->l_derivs);
  
      j=0;
      for(size_t i=0;i<this->l_n;i++) {
	if ((*this->l_index)[i]%2==1) {
	  // Construct the equations from the rightmost point
	  if ((*this->l_yend)[i]==0.0) {
	    sy[j]=y2[i];
	  } else {
	    sy[j]=(y2[i]-(*this->l_yend)[i])/(*this->l_yend)[i];
	  }
	  j++;
	} else {
	  // Otherwise copy the final values from y2 to *l_yend
	  (*this->l_yend)[i]=y2[i];
	}
      }
  
      return 0;
    }

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
