/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_IT_SOLVE_H
#define O2SCL_ODE_IT_SOLVE_H

#include <iostream>

#include <o2scl/misc.h>
#include <o2scl/array.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/test_mgr.h>
#include <o2scl/linear_solver.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /// Function class for ode_it_solve
  template<class param_t, class vec_t=ovector_base> class ode_it_funct {
  
  public:
  
  ode_it_funct() {
  }
  
  virtual ~ode_it_funct() {}
  
  /// Using \c x and \c y, return the value of function number \c ieq 
  virtual double operator()(size_t ieq, double x, vec_t &y, param_t &pa) {
    return 0.0;
  }

  };

  /// Function pointer for ode_it_solve
  template<class param_t, class vec_t=ovector_base> 
    class ode_it_funct_fptr : public ode_it_funct<param_t,vec_t> {

  public:

  virtual ~ode_it_funct_fptr() {}

  /// Create using a function pointer
  ode_it_funct_fptr(double (*fp)(size_t, double, vec_t &, param_t &)) {
    fptr=fp;
  }

  /// Using \c x and \c y, return the value of function number \c ieq 
  virtual double operator()(size_t ieq, double x, vec_t &y, param_t &pa) {
    return fptr(ieq,x,y,pa);
  }

  protected:

  ode_it_funct_fptr() {}

  /// The function pointer
  double (*fptr)(size_t ieq, double x, vec_t &y, param_t &pa);
  };

  /// Member function pointer for ode_it_solve
  template<class tclass, class param_t, class vec_t=ovector_base>
    class ode_it_funct_mfptr : public ode_it_funct<param_t,vec_t> {
  public:

  virtual ~ode_it_funct_mfptr() {}

  /// Create using a class instance and member function
  ode_it_funct_mfptr(tclass *tp, 
		     double (tclass::*fp)(size_t, double, 
					  vec_t &, param_t &)) {
    tptr=tp;
    fptr=fp;
  }

  /// Using \c x and \c y, return the value of function number \c ieq 
  virtual double operator()(size_t ieq, double x, vec_t &y, param_t &pa) {
    return (*tptr.*fptr)(ieq,x,y,pa);
  }

  protected:

  ode_it_funct_mfptr() {}

  /// The class pointer
  tclass *tptr;

  /// The member function pointer
  double (tclass::*fptr)(size_t ieq, double x, vec_t &y, param_t &pa);

  };

  /** \brief ODE solver using a generic linear solver to solve 
      finite-difference equations

      \note This class does not work with all matrix and vector
      types because using operator[][] would require it to be
      inefficient for use with sparse matrices.

      \future Create a GSL-like set() and iterate() interface
      \future Implement as a child of ode_bv_solve ?
      \future Max and average tolerance?
  */
  template <class param_t, class func_t=ode_it_funct<param_t>, 
    class vec_t=ovector_base, class mat_t=omatrix_base, 
    class matrix_row_t=omatrix_row, class solver_vec_t=ovector_base, 
    class solver_mat_t=omatrix_base> class ode_it_solve {
    
  public:
  
  ode_it_solve() {
    h=1.0e-4;
    niter=30;
    tolf=1.0e-8;
    verbose=0;
    solver=&def_solver;
    alpha=1.0;
  }

  virtual ~ode_it_solve() {}

  /// Set level of output (default 0)
  int verbose;

  /** \brief Stepsize for finite differencing (default \f$ 10^{-4} \f$)
   */
  double h;

  /// Tolerance (default \f$ 10^{-8} \f$)
  double tolf;
  
  /// Maximum number of iterations (default 30)
  size_t niter;
  
  /// Set the linear solver
  int set_solver(o2scl_linalg::linear_solver<solver_vec_t,solver_mat_t> &ls) {
    solver=&ls;
    return 0;
  }
    
  /// Size of correction to apply (default 1.0)
  double alpha;

  /** \brief Solve \c derivs with boundary conditions \c left and 
      \c right

      Given a grid of size \c n_grid and \c n_eq differential equations,
      solve them by relaxation. The grid is specified in \c x, which
      is a vector of size \c n_grid. The differential equations are
      given in \c derivs, the boundary conditions on the left hand
      side in \c left, and the boundary conditions on the right hand
      side in \c right. The number of boundary conditions on the left
      hand side is \c nb_left, and the number of boundary conditions on
      the right hand side should be <tt>n_eq-nb_left</tt>. The initial
      guess for the solution, a matrix of size <tt>[n_grid][n_eq]</tt>
      should be given in \c y. Upon success, \c y will contain an
      approximate solution of the differential equations. The matrix
      \c mat is workspace of size <tt>[n_grid*n_eq][n_grid*n_eq]</tt>, and
      the vectors \c rhs and \c y are workspace of size
      <tt>[n_grid*n_eq]</tt>.

  */
  int solve(size_t n_grid, size_t n_eq, size_t nb_left, vec_t &x, 
	    mat_t &y, func_t &derivs, func_t &left, func_t &right,
	    solver_mat_t &mat, solver_vec_t &rhs, solver_vec_t &dy,
	    param_t &pa) {

    // Store the functions for simple derivatives
    fd=&derivs;
    fl=&left;
    fr=&right;
    
    // Variable index
    size_t ix;

    // Number of RHS boundary conditions
    size_t nb_right=n_eq-nb_left;

    // Number of variables
    size_t nvars=n_grid*n_eq;
    
    bool done=false;
    for(size_t it=0;done==false && it<niter;it++) {
      
      ix=0;
      
      mat.set_all(0.0);

      // Construct the entries corresponding to the LHS boundary. 
      // This makes the first nb_left rows of the matrix.
      for(size_t i=0;i<nb_left;i++) {
	matrix_row_t yk(y,0);
	rhs[ix]=-left(i,x[0],yk,pa);
	for(size_t j=0;j<n_eq;j++) {
	  int rxa=mat.set(ix,j,fd_left(i,j,x[0],yk,pa));
	}
	ix++;
      }

      // Construct the matrix entries for the internal points
      // This loop adds n_grid-1 sets of n_eq rows
      for(size_t k=0;k<n_grid-1;k++) {
	size_t kp1=k+1;
	double tx=(x[kp1]+x[k])/2.0;
	double dx=x[kp1]-x[k];
	matrix_row_t yk(y,k);
	matrix_row_t ykp1(y,k+1);
	
	for(size_t i=0;i<n_eq;i++) {
	  
	  rhs[ix]=y[k][i]-y[kp1][i]+(x[kp1]-x[k])*
	    (derivs(i,tx,ykp1,pa)+derivs(i,tx,yk,pa))/2.0;
	  
	  size_t lhs=k*n_eq;
	  for(size_t j=0;j<n_eq;j++) {
	    int rxb=mat.set(ix,lhs+j,-fd_derivs(i,j,tx,yk,pa)*dx/2.0);
	    int rxc=mat.set(ix,lhs+j+n_eq,-fd_derivs(i,j,tx,ykp1,pa)*dx/2.0);
	    if (i==j) {
	      int rxd=mat.set(ix,lhs+j,mat(ix,lhs+j)-1.0);
	      int rxe=mat.set(ix,lhs+j+n_eq,mat(ix,lhs+j+n_eq)+1.0);
	    }
	  }

	  ix++;

	}
      }
      
      // Construct the entries corresponding to the RHS boundary
      // This makes the last nb_right rows of the matrix.
      for(size_t i=0;i<nb_right;i++) {
	matrix_row_t ylast(y,n_grid-1);
	size_t lhs=n_eq*(n_grid-1);
	
	rhs[ix]=-right(i,x[n_grid-1],ylast,pa);
	
	for(size_t j=0;j<n_eq;j++) {
	  int rxf=mat.set(ix,lhs+j,fd_right(i,j,x[n_grid-1],ylast,pa));
	}
	  
	ix++;

      }
      
      // Compute correction by calling the linear solver

      if (verbose>3) {
	std::cout << "Matrix: " << std::endl;
	for(size_t i=0;i<nvars;i++) {
	  for(size_t j=0;j<nvars;j++) {
	    std::cout << mat[i][j] << " ";
	  }
	  std::cout << std::endl;
	}
	std::cout << "Deviations:" << std::endl;
	for(size_t i=0;i<nvars;i++) {
	  std::cout << rhs[i] << std::endl;
	}
      }
	
      int ret=solver->solve(ix,mat,rhs,dy);

      if (verbose>3) {
	std::cout << "Corrections:" << std::endl;
	for(size_t i=0;i<nvars;i++) {
	  std::cout << dy[i] << std::endl;
	}
	std::cout << "Press a key and press enter to continue: " 
		  << std::endl;
	char ch;
	std::cin >> ch;
      }
      
      // Apply correction and compute its size

      double res=0.0;
      ix=0;

      for(size_t igrid=0;igrid<n_grid;igrid++) {
	for(size_t ieq=0;ieq<n_eq;ieq++) {
	  y[igrid][ieq]+=alpha*dy[ix];
	  res+=dy[ix]*dy[ix];
	  ix++;
	}
      }

      if (verbose>0) {
	// Since we're in the o2scl namespace, we explicitly
	// specify std::sqrt() here
	std::cout << "ode_it_solve: " << it << " " << std::sqrt(res) << " " 
		  << tolf << std::endl;
	if (verbose>1) {
	  char ch;
	  std::cout << "Press a key and type enter to continue. ";
	  std::cin >> ch;
	}
      }
      
      // If the correction has become small enough, we're done
      if (std::sqrt(res)<=tolf) done=true;
    }

    if (done==false) {
      O2SCL_ERR_RET("Exceeded number of iterations in solve().",
		    o2scl::gsl_emaxiter);
    }

    return 0;
  }
  
  /// Default linear solver
  o2scl_linalg::linear_solver_hh<solver_vec_t,solver_mat_t> def_solver;
  
  protected:
  
  /// \name Storage for functions
  //@{
  ode_it_funct<param_t,vec_t> *fl, *fr, *fd;
  //@}

  /// Solver
  o2scl_linalg::linear_solver<solver_vec_t,solver_mat_t> *solver;

  /** \brief Compute the derivatives of the LHS boundary conditions

      This function computes \f$ \partial f_{left,\mathrm{ieq}} / \partial
      y_{\mathrm{ivar}} \f$
  */
  virtual double fd_left(size_t ieq, size_t ivar, double x, vec_t &y, 
			 param_t &pa) {
    double ret, dydx;
    
    y[ivar]+=h;
    ret=(*fl)(ieq,x,y,pa);
    
    y[ivar]-=h;
    ret-=(*fl)(ieq,x,y,pa);
    
    ret/=h;
    return ret;
  }
  
  /** \brief Compute the derivatives of the RHS boundary conditions
	
      This function computes \f$ \partial f_{right,\mathrm{ieq}} / \partial
      y_{\mathrm{ivar}} \f$
  */
  virtual double fd_right(size_t ieq, size_t ivar, double x, vec_t &y, 
			  param_t &pa) {
    double ret, dydx;
    
    y[ivar]+=h;
    ret=(*fr)(ieq,x,y,pa);
    
    y[ivar]-=h;
    ret-=(*fr)(ieq,x,y,pa);
    
    ret/=h;
    return ret;
  }
  
  /** \brief Compute the finite-differenced part of the 
      differential equations

      This function computes \f$ \partial f_{\mathrm{ieq}} / \partial
      y_{\mathrm{ivar}} \f$
  */
  virtual double fd_derivs(size_t ieq, size_t ivar, double x, vec_t &y, 
			   param_t &pa) {
    double ret, dydx;
    
    y[ivar]+=h;
    ret=(*fd)(ieq,x,y,pa);
    
    y[ivar]-=h;
    ret-=(*fd)(ieq,x,y,pa);
    
    ret/=h;
    
    return ret;
  }
  
  };

#ifndef DOXYGENP
}
#endif

#endif
