/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_IV_SOLVE_H
#define O2SCL_ODE_IV_SOLVE_H

#include <string>
#include <o2scl/collection.h>
#include <o2scl/adapt_step.h>
#include <o2scl/gsl_astep.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Solve an initial-value ODE problems given an adaptive ODE
      stepper
  */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class ode_iv_solve {
   
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
  /// Memory allocator
  alloc_t ao;

  /// The adaptive stepper
  adapt_step<param_t,func_t,vec_t,alloc_vec_t,alloc_t> *astp;
    
  /// Print out iteration information
  virtual int print_iter(double x, size_t nv, vec_t &y) {
    std::cout << type() << " x: " << x << " y: ";
    for(size_t i=0;i<nv;i++) std::cout << y[i] << " ";
    std::cout << std::endl;
    if (verbose>1) {
      char ch;
      std::cin >> ch;
    }
    return 0;
  }
  
#endif

  public:
      
  ode_iv_solve() {
    verbose=0;
    ntrial=1000;
    nsteps_out=10;
    astp=&gsl_astp;
    exit_on_fail=true;
  }
      
  virtual ~ode_iv_solve() {}

  /// Set the adaptive stepper to use
  int set_adapt_step(adapt_step<param_t,func_t,vec_t,
		     alloc_vec_t,alloc_t> &as) {
    astp=&as;
    return 0;
  }

  /** 
      \brief Solve the initial-value problem and output a table 

      Initially, \c x_sol should be a vector of size \c n_sol, and \c
      y_sol, \c dydx_sol, and \c yerr_sol should be matrices with size
      \c [n_sol][n].  On exit, \c n_sol will will be the size of the
      solution table, less than or equal to the original value of \c
      n_sol. This function avoids performing extra calls to the
      adaptive stepper, and the table will be approximately evenly
      spaced.

      This template function works with any matrix class \c mat_t
      which can be accessed using successive applications of
      operator[] and which has an associated class \c mat_row2_t
      which returns a row of a matrix of type \c mat_t as
      an object with type \c vec_t. 

      If \ref verbose is greater than zero, The solution at each
      internal point will be written to \c std::cout.  If \ref verbose
      is greater than one, a character will be required after each
      point.

      \comment
      10/23/09 - Some copying can be avoided by removing the "yerrl"
      and "dydx2" temporary vectors, for example.
      \endcomment
  */
  template<class mat_t, class mat_row2_t>
  int solve_table(double x0, double x1, double h, size_t n,
		  vec_t &ystart, size_t &n_sol, vec_t &x_sol, 
		  mat_t &y_sol, mat_t &dydx_sol, mat_t &yerr_sol, 
		  param_t &pa, func_t &derivs) {
    
    int ret=0;
    int first_ret=0;
    size_t nmax=n_sol;
    nsteps=0;

    // Stepsize for next verbose output. Use nsteps_out-1 instead of
    // nsteps_out since the first point is always output below.
    double dx_verb=(x1-x0)/((double)(nsteps_out-1));
    // Stepsize for next table entry
    double dx_tab=(x1-x0)/((double)(n_sol-1));

    double x_verb=x0+dx_verb, x_tab=x0+dx_tab;
    
    // Output first point 
    if (verbose>0) {
      print_iter(x0,n,ystart);
      if (verbose>1) {
	char ch;
	std::cin >> ch;
      }
    }

    // Allocate space for errors and derivatives
    alloc_vec_t yerrl, dydx, dydx2;
    ao.allocate(yerrl,n);
    ao.allocate(dydx,n);
    ao.allocate(dydx2,n);

    // Initial derivative evaulation
    derivs(x0,n,ystart,dydx,pa);

    // Add first point to table
    x_sol[0]=x0;
    for(size_t j=0;j<n;j++) {
      y_sol[0][j]=ystart[j];
      dydx_sol[0][j]=dydx[j];
      yerr_sol[0][j]=0.0;
    }

    // Copy first point to table again for first step
    size_t icurr=1;
    x_sol[1]=x0;
    for(size_t j=0;j<n;j++) y_sol[1][j]=ystart[j];

    double xnext;

    bool done=false;
    while (done==false) {

      // Use ystart as temporary storage for the end of the current
      // adaptive step
      
      mat_row2_t yrow(y_sol,icurr);
      ret=astp->astep_full(x0,h,x1,xnext,n,yrow,dydx,ystart,yerrl,
			   dydx2,pa,derivs);
      
      if (ret!=0) {
	if (exit_on_fail) {
	  n_sol=icurr+1;

	  ao.free(dydx);
	  ao.free(dydx2);
	  ao.free(yerrl);

	  // call error handler
	  O2SCL_ERR2_RET("Adaptive stepper returned non-zero in ",
			 "ode_iv_solve::solve_table().",gsl_efailed);
	} else if (first_ret==0) {
	  first_ret=ret;
	}
      }
      
      // Update step counter and abort if necessary
      nsteps++;
      if (nsteps>ntrial) {
	ao.free(dydx);
	ao.free(dydx2);
	ao.free(yerrl);

	O2SCL_ERR2_RET("Too many steps required in ",
		       "ode_iv_solve::solve_table().",gsl_emaxiter);
      }
      
      // If we've made enough progress, do verbose output
      if (xnext>=x_verb && verbose>0) {
	print_iter(xnext,n,ystart);
	if (verbose>1) {
	  char ch;
	  std::cin >> ch;
	}
	x_verb+=dx_verb;
      }

      // If we're at the end
      if (xnext>=x1) {

	// Exit the loop
	done=true;

	// Store the final entry in the table
	x_sol[icurr]=xnext;
	for(size_t j=0;j<n;j++) {
	  y_sol[icurr][j]=ystart[j];
	  dydx_sol[icurr][j]=dydx2[j];
	  yerr_sol[icurr][j]=yerrl[j];
	}

	// Update the table size
	n_sol=icurr+1;

      } else {

	if (xnext>=x_tab && icurr!=nmax-1) {
	  
	  // If we've made enough progress, store an entry in the table
	  x_sol[icurr]=xnext;
	  for(size_t j=0;j<n;j++) {
	    y_sol[icurr][j]=ystart[j];
	    dydx_sol[icurr][j]=dydx2[j];
	    yerr_sol[icurr][j]=yerrl[j];
	    
	    // Also prepare for the next step
	    y_sol[icurr+1][j]=ystart[j];
	    dydx[j]=dydx2[j];
	  }
	  x_tab+=dx_tab;
	  
	  // Update x0 and the current and next indicies
	  x0=xnext;
	  icurr++;
	  
	} else {

	  // Otherwise, just prepare for the next step
	  x0=xnext;
	  for(size_t j=0;j<n;j++) {
	    dydx[j]=dydx2[j];
	    y_sol[icurr][j]=ystart[j];
	  }
	  
	}
	
      }
      
    }
    
    ao.free(dydx);
    ao.free(dydx2);
    ao.free(yerrl);

    return first_ret;
  }

  /** \brief Solve the initial-value problem from \c x0 to \c x1 over
      a grid storing derivatives

      Initially, \c xsol should be an array of size \c nsol, and \c
      ysol should be a \c omatrix of size \c [nsol][n]. This function
      never takes a step larger than the grid size.

      If \ref verbose is greater than zero, The solution at each grid
      point will be written to \c std::cout. If \ref verbose is
      greater than one, a character will be required after each point.
  */
  template<class alloc_mat_t, class mat_row_t, class mat_alloc_t>
  int solve_grid(double x0, double x1, double h, size_t n,
		 vec_t &ystart, size_t nsol, vec_t &xsol, 
		 alloc_mat_t &ysol, alloc_mat_t &dydx_sol, 
		 param_t &pa, func_t &derivs) {
    int ret=0;
    mat_alloc_t amat;
    alloc_mat_t err_sol;
    amat.allocate(err_sol,nsol,n);
    solve_grid<alloc_mat_t,mat_row_t>(x0,x1,h,n,ystart,nsol,xsol,ysol,
				dydx_sol,err_sol,pa,derivs);
    amat.free(err_sol,nsol);
    return ret;
  }

  /** \brief Solve the initial-value problem from \c x0 to \c x1 over
      a grid storing derivatives

      Initially, \c xsol should be an array of size \c nsol, and \c
      ysol should be a \c omatrix of size \c [nsol][n]. This function
      never takes a step larger than the grid size.

      If \ref verbose is greater than zero, The solution at each grid
      point will be written to \c std::cout. If \ref verbose is
      greater than one, a character will be required after each point.
  */
  template<class mat_t, class mat_row_t>
  int solve_grid(double x0, double x1, double h, size_t n,
		 vec_t &ystart, size_t nsol, vec_t &xsol, 
		 mat_t &ysol, mat_t &dydx_sol, mat_t &err_sol,
		 param_t &pa, func_t &derivs) {
    
    double x=x0, xnext;
    int ret=0, first_ret=0;
    nsteps=0;
      
    // Copy the initial point to the first row
    xsol[0]=x0;
    for(size_t j=0;j<n;j++) ysol[0][j]=ystart[j];

    // Verbose output for the first row
    if (verbose>0) print_iter(xsol[0],n,ystart);

    // Create space for the initial point of each step
    vec_t &y_tmp=ystart;
    alloc_vec_t dydx_tmp;
    ao.allocate(dydx_tmp,n);

    // Evalulate the first derivative
    derivs(x0,n,y_tmp,dydx_tmp,pa);

    for(size_t i=1;i<nsol && ret==0;i++) {
	  
      xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
	    
      mat_row_t y_row(ysol,i);
      mat_row_t dydx_row(dydx_sol,i);
      mat_row_t yerr_row(err_sol,i);

      // Step until we reach the next grid point
      bool done=false;
      while(done==false && ret==0) {
	
	ret=astp->astep_full(x,h,xnext,xsol[i],n,y_tmp,dydx_tmp,
			     y_row,yerr_row,dydx_row,pa,derivs);
	nsteps++;
	if (ret!=0) {
	  if (exit_on_fail) {
	    ao.free(dydx_tmp);
	    O2SCL_ERR2_RET("Adaptive stepper failed in ",
			   "ode_iv_solve::solve_grid()",ret);
	  } else if (first_ret!=0) {
	    first_ret=ret;
	  }
	}
	
	if (nsteps>ntrial) {
	  ao.free(dydx_tmp);
	  O2SCL_ERR2_RET("Too many steps required in ",
			 "ode_iv_solve::solve_grid().",gsl_emaxiter);
	}

	// Copy the results of the last step to the starting
	// point for the next step
	x=xsol[i];
	for(size_t j=0;j<n;j++) {
	  y_tmp[j]=y_row[j];
	  dydx_tmp[j]=dydx_row[j];
	}

	// Decide if we have reached the grid point
	if (x1>x0) {
	  if (x>=xnext) done=true;
	} else {
	  if (x<=xnext) done=true;
	}
	    
      }
      
      if (verbose>0) print_iter(xsol[i],n,dydx_row);

    }
    
    ao.free(dydx_tmp);
    
    return first_ret;
  }

  /** \brief Solve the initial-value problem 
      to get the final value 

      If \ref verbose is greater than zero, The solution at less
      than or approximately equal to \ref nsteps_out points will be
      written to \c std::cout. If \ref verbose is greater than one,
      a character will be required after each selected point.
  */
  int solve_final_value(double x0, double x1, double h, size_t n,
			vec_t &ystart, vec_t &yend, 
			param_t &pa, func_t &derivs) {
    
    alloc_vec_t dydx, dydx2, yerr;
    ao.allocate(dydx,n);
    ao.allocate(dydx2,n);
    ao.allocate(yerr,n);
    int ret=solve_final_value(x0,x1,h,n,ystart,dydx,yend,yerr,
			      dydx2,pa,derivs);
    ao.free(dydx);
    ao.free(dydx2);
    ao.free(yerr);
    return ret;
  }
      
  /** \brief Solve the initial-value problem 
      to get the final value 

      If \ref verbose is greater than zero, The solution at less
      than or approximately equal to \ref nsteps_out points will be
      written to \c std::cout. If \ref verbose is greater than one,
      a character will be required after each selected point.
  */
  int solve_final_value(double x0, double x1, double h, size_t n,
			vec_t &ystart, vec_t &yend, vec_t &yerr,
			param_t &pa, func_t &derivs) {
    
    alloc_vec_t dydx, dydx2;
    ao.allocate(dydx,n);
    ao.allocate(dydx2,n);
    int ret=solve_final_value(x0,x1,h,n,ystart,dydx,yend,yerr,
			      dydx2,pa,derivs);
    ao.free(dydx);
    ao.free(dydx2);
    return ret;
  }
      
  /** \brief Solve the initial-value problem 
      to get the final value and derivative

      If \ref verbose is greater than zero, The solution at less
      than or approximately equal to \ref nsteps_out points will be
      written to \c std::cout. If \ref verbose is greater than one,
      a character will be required after each selected point.

      The solution fails if more than \ref ntrial steps are required.

      This function will also fail if <tt>x1>x0</tt> and <tt>h<0</tt>
      or if <tt>x1<x0</tt> but <tt>h>0</tt>.
  */
  int solve_final_value(double x0, double x1, double h, size_t n,
			vec_t &ystart, vec_t &dydx_start, vec_t &yend, 
			vec_t &yerr, vec_t &dydx_end,
			param_t &pa, func_t &derivs) {
    
    if ((x1>x0 && h<=0.0) || (x0>x1 && h>=0.0)) {
      O2SCL_ERR_RET("Interval ordering and sign of h don't match.",
		    gsl_einval);
    }


    double x=x0, xverb=0.0, dxverb=0.0, xnext;
    int ret=0, first_ret=0;

    nsteps=0;

    // Compute stepsize for verbose output
    if (verbose>0) {
      print_iter(x0,n,ystart);
      if (verbose>1) {
	char ch;
	std::cin >> ch;
      }
      if (x1>x0) {
	dxverb=(x1-x0)/((double)nsteps_out);
	xverb=x0+dxverb;
      } else {
	dxverb=(x0-x1)/((double)nsteps_out);
	xverb=x0-dxverb;
      }
    }

    // Use yend as workspace
    for(size_t i=0;i<n;i++) yend[i]=ystart[i];

    // Compute initial derivative
    derivs(x,n,yend,dydx_start,pa);
    for(size_t i=0;i<n;i++) dydx_end[i]=dydx_start[i];
      
    bool done=false;
    while (done==false) {
	
      // Take a step of the first type
      ret=astp->astep_full(x,h,x1,xnext,n,ystart,dydx_start,yend,yerr,
			   dydx_end,pa,derivs);

      if (ret!=0) {
	if (exit_on_fail) {
	  O2SCL_ERR2_RET("Adaptive stepper failed in ",
			 "ode_iv_solve::solve_grid()",ret);
	} else if (first_ret!=0) {
	  first_ret=ret;
	}
      }

      // Print out verbose info 
      if (verbose>0 && xnext>xverb) {
	print_iter(xnext,n,yend);
	if (verbose>1) {
	  char ch;
	  std::cin >> ch;
	}
	xverb+=dxverb;
      }

      // Check number of iterations
      nsteps++;
      if (nsteps>ntrial) {
	O2SCL_ERR2_RET("Exceeded max number of iterations in ",
		       "ode_iv_solve::solve_final_value().",
		       gsl_emaxiter);
      }

      if (ret!=0) {
	done=true;
      } else {
	if (x1>x0) {
	  if (xnext>=x1) done=true;
	} else {
	  if (xnext<=x1) done=true;
	}
      }

      if (done==false) {

	// Take a step of the second type
	ret=astp->astep_full(xnext,h,x1,x,n,yend,dydx_end,ystart,yerr,
			     dydx_start,pa,derivs);
	
	if (ret!=0) {
	  if (exit_on_fail) {
	    O2SCL_ERR2_RET("Adaptive stepper failed in ",
			   "ode_iv_solve::solve_grid()",ret);
	  } else if (first_ret!=0) {
	    first_ret=ret;
	  }
	}

	// Print out verbose info 
	if (verbose>0 && x>xverb) {
	  print_iter(x,n,ystart);
	  if (verbose>1) {
	    char ch;
	    std::cin >> ch;
	  }
	  xverb+=dxverb;
	}

	// Check number of iterations
	nsteps++;
	if (nsteps>ntrial) {
	  O2SCL_ERR2_RET("Exceeded max number of iterations in ",
			 "ode_iv_solve::solve_final_value().",
			 gsl_emaxiter);
	}

	if (ret!=0) {
	  done=true;
	} else {
	  if (x1>x0) {
	    if (x>=x1) {
	      done=true;
	      for(size_t j=0;j<n;j++) {
		yend[j]=ystart[j];
		dydx_end[j]=dydx_start[j];
	      }
	    }
	  } else {
	    if (x<=x1) {
	      done=true;
	      for(size_t j=0;j<n;j++) {
		yend[j]=ystart[j];
		dydx_end[j]=dydx_start[j];
	      }
	    }
	  }
	}

      }

      // End of while loop
    }
	
    // Print out final verbose info
    if (verbose>0) {
      print_iter(x1,n,yend);
      if (verbose>1) {
	char ch;
	std::cin >> ch;
      }
    }
	
    return first_ret;
  }
      
  /// Set output level
  int verbose;
      
  /** 
      \brief Number of output points if \ref verbose is greater 
      than zero (default 10)

      This is used in functions solve_table() and solve_final_value()
      to control how often steps are output when verbose is greater
      than zero.
  */
  size_t nsteps_out;

  /// Maximum number of applications of the adaptive stepper (default 1000)
  size_t ntrial;

  /// Number of adaptive steps employed
  size_t nsteps;

  /** 
      \brief If true, stop the solution if the adaptive stepper fails
      
      If this is false, then failures in the adaptive stepper are
      ignored.
  */
  bool exit_on_fail;

  /// The default adaptive stepper
  gsl_astep<param_t,func_t,vec_t,alloc_vec_t,alloc_t> gsl_astp;

  /// Return the type, \c "ode_iv_solve".
  virtual const char *type() { return "ode_iv_solve"; }
  
  };

#ifndef DOXYGENP
}
#endif

#endif
