/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/convert_units.h>

using namespace std;
using namespace o2scl;

/** \brief Return the value \c val after converting using units \c
    from and \c to
*/
double convert_units::convert(std::string from, std::string to, double val) {

  // Remove whitespace
  remove_whitespace(from);
  remove_whitespace(to);

  // Look in cache for conversion
  std::string both=from+","+to;
  miter m3=mcache.find(both);
  if (m3!=mcache.end()) {
    return val*m3->second.c;
  }

  // Look in cache for reverse conversion
  std::string both2=to+","+from;
  m3=mcache.find(both2);
  if (m3!=mcache.end()) {
    return val/m3->second.c;
  }

  // Look for combined conversions
  for(miter m=mcache.begin();m!=mcache.end();m++) {
    if (m->second.f==from) {
      std::string b=m->second.t+","+to;
      miter m2=mcache.find(b);
      if (m2!=mcache.end()) {
	if (verbose>0) {
	  std::cout << "Using conversions: " << m->second.f << " , "
		    << m->second.t << " " << m->second.c << std::endl;
	  std::cout << " (1)          and: " << m2->second.f << " , "
		    << m2->second.t << " " << m2->second.c << std::endl;
	}
	return val*m->second.c*m2->second.c;
      }
      std::string b2=to+","+m->second.t;
      miter m4=mcache.find(b2);
      if (m4!=mcache.end()) {
	if (verbose>0) {
	  std::cout << "Using conversions: " << m->second.f << " , "
		    << m->second.t << std::endl;
	  std::cout << " (2)          and: " << m4->second.f << " , "
		    << m4->second.t << std::endl;
	}
	return val*m->second.c/m4->second.c;
      }
    } else if (m->second.t==from) {
      std::string b=m->second.f+","+to;
      miter m2=mcache.find(b);
      if (m2!=mcache.end()) {
	if (verbose>0) {
	  std::cout << "Using conversions: " << m->second.f << " , "
		    << m->second.t << std::endl;
	  std::cout << " (3)          and: " << m2->second.f << " , "
		    << m2->second.t << std::endl;
	}
	return val/m->second.c*m2->second.c;
      }
    } else if (m->second.f==to) {
      std::string b=m->second.t+","+from;
      miter m2=mcache.find(b);
      if (m2!=mcache.end()) {
	if (verbose>0) {
	  std::cout << "Using conversions: " << m->second.f << " , "
		    << m->second.t << std::endl;
	  std::cout << " (4)          and: " << m2->second.f << " , "
		    << m2->second.t << std::endl;
	}
	return val/m->second.c/m2->second.c;
      }
      std::string b2=from+","+m->second.t;
      miter m4=mcache.find(b2);
      if (m4!=mcache.end()) {
	if (verbose>0) {
	  std::cout << "Using conversions: " << m->second.f << " , "
		    << m->second.t << std::endl;
	  std::cout << " (5)          and: " << m4->second.f << " , "
		    << m4->second.t << std::endl;
	}
	return val/m->second.c*m4->second.c;
      }
    } else if (m->second.t==to) {
      std::string b=m->second.f+","+from;
      miter m2=mcache.find(b);
      if (m2!=mcache.end()) {
	if (verbose>0) {
	  std::cout << "Using conversions: " << m->second.f << " , "
		    << m->second.t << std::endl;
	  std::cout << " (6)          and: " << m2->second.f << " , "
		    << m2->second.t << std::endl;
	}
	return val*m->second.c/m2->second.c;
      }
    }
  }

  if (use_gnu_units) {
    
    // Run the GNU 'units' command
    std::string cmd=units_cmd_string+" '"+from+"' '"+to+"'";
    if (verbose>0) std::cout << "cmd: " << cmd << std::endl;
    
    FILE *ps_pipe=popen(cmd.c_str(),"r");
    if (err_on_fail && !ps_pipe) {
      O2SCL_ERR("Pipe could not be opened in convert_units::convert().",
		gsl_efailed);
      return 0.0;
    }

    char line1[80];
    int size=80;

    fgets(line1,80,ps_pipe);
    if (verbose>0) std::cout << "output: " << line1 << std::endl;

    // Read the output from the 'units' command and compute the 
    // conversion factor
    std::string s=line1, t1, t2;
    std::istringstream *ins=new std::istringstream(s);
    (*ins) >> t1 >> t2;
    delete ins;
    double conv=stod(t2);
      
    // Cleanup
    if (err_on_fail && pclose(ps_pipe)!=0) {
      O2SCL_ERR("Pipe could not be closed in convert_units::convert().",
		gsl_efailed);
      return 0.0;
    }
      
    // Add the newly computed conversion to the table
    unit_t ut;
    ut.f=from;
    ut.t=to;
    ut.c=conv;
    mcache.insert(make_pair(both,ut));
      
    return conv*val;
  }
      
  if (err_on_fail) {
    O2SCL_ERR("Conversion not found in convert_units::convert().",
	      gsl_efailed);
  }

  return 0.0;
}

int convert_units::remove_cache(std::string from, std::string to) {

  // Remove whitespace
  remove_whitespace(from);
  remove_whitespace(to);
  std::string both=from+","+to;

  miter m3=mcache.find(both);
  if (m3!=mcache.end()) {
    mcache.erase(m3);
    return gsl_success;
  }
  
  if (err_on_fail) {
    O2SCL_ERR((((string)"Conversion ")+from+" -> "+to+
	       " not found in convert_units::remove_cache().").c_str(),
	      gsl_enotfound);
  }

  return gsl_enotfound;
}

int convert_units::insert_cache
(std::string from, std::string to, double conv) {

  // Remove whitespace
  remove_whitespace(from);
  remove_whitespace(to);
  
  if (err_on_fail && 
      (from.find(',')!=std::string::npos || 
       to.find(',')!=std::string::npos)) {
    O2SCL_ERR_RET("Units cannot contain comma in insert_cache()",
		  gsl_efailed);
  }

  unit_t ut;
  ut.f=from;
  ut.t=to;
  ut.c=conv;
  std::string both=from+","+to;

  // If it's already inside the cache, just update the conversion
  // value
  miter m3=mcache.find(both);
  if (m3!=mcache.end()) {
    m3->second=ut;
  }

  mcache.insert(make_pair(both,ut));
  return 0;
}
    
/// Print the present unit cache
int convert_units::print_cache() {
  miter m;
  if (mcache.size()==0) {
    cout << "No units in cache." << endl;
  }
  std::cout << "Unit cache: " << std::endl;
  for (m=mcache.begin();m!=mcache.end();m++) {
    std::cout.setf(std::ios::left);
    std::cout.width(25);
    std::cout << m->second.f << " ";
    std::cout.width(25);
    std::cout << m->second.t << " ";
    std::cout.unsetf(std::ios::left);
    std::cout.precision(10);
    std::cout << m->second.c << std::endl;
  }
  return 0;
}

/// Add conversion factors for energy equivalents
int convert_units::energy_conv() {
      
  // Entries derived from http://physics.nist.gov/cuu/Constants.
  // The NIST website uses 'h' to convert instead of hbar,
  // so we have to add a lot of factors of 2 pi
      
  insert_cache("J","m^-1",5.03411747e24*2.0*o2scl_const::pi);
  insert_cache("J","km^-1",5.03411747e27*2.0*o2scl_const::pi);
  insert_cache("J","cm^-1",5.03411747e22*2.0*o2scl_const::pi);
  insert_cache("J","fm^-1",5.03411747e9*2.0*o2scl_const::pi);
  insert_cache("J","s^-1",1.509190450e33*2.0*o2scl_const::pi);
  insert_cache("J","K",1.0e7/gsl_cgs::boltzmann);
  insert_cache("J","eV",6.24150965e18);
  insert_cache("J","MeV",6.24150965e12);
  insert_cache("J","u",6.70053641e9);
  insert_cache("J","erg",1.0e7);
  insert_cache("J","kg",1.0/gsl_mks::speed_of_light/
	       gsl_mks::speed_of_light);
  insert_cache("J","g",1.0e3/gsl_mks::speed_of_light/
	       gsl_mks::speed_of_light);
      
  return 0;
}
