/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_POLY_H
#define O2SCL_POLY_H

/** \file poly.h
    \brief Classes for solving polynomials

    \warning 
    We should be careful about using pow() in functions using
    complex<double> since pow(((complex<double>)0.0),3.0) returns
    (nan,nan). Instead, we should use pow(((complex<double>)0.0),3)
    which takes an integer for the second argument. The sqrt() function,
    always succeeds i.e. sqrt(((complex<double>)0.0))=0.0

    \future The quartics are tested only for a4=1, which should
    probably be generalized.
    
*/

#include <iostream>
#include <complex>
#include <gsl/gsl_math.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_poly.h>
#include <o2scl/constants.h>
#include <o2scl/err_hnd.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Solve a quadratic polynomial with real coefficients and 
      real roots [abstract base]
  */
  class quadratic_real {
  public:

    virtual ~quadratic_real() {}

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ .
    */
    virtual int solve_r(const double a2, const double b2, const double c2, 
			double &x1, double &x2)=0;

    /// Return a string denoting the type ("quadratic_real")
    const char *type() { return "quadratic_real"; }
  };

  /** 
      \brief Solve a quadratic polynomial with real coefficients and 
      complex roots [abstract base]
  */
  class quadratic_real_coeff : public quadratic_real {

  public:

    virtual ~quadratic_real_coeff() {}

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ .
    */
    virtual int solve_r(const double a2, const double b2, const double c2, 
			double &x1, double &x2);

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two complex solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ 
    */
    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, 
			 std::complex<double> &x2)=0;

    /// Return a string denoting the type ("quadratic_real_coeff")
    const char *type() { return "quadratic_real_coeff"; }
  };

  /** 
      \brief Solve a quadratic polynomial with complex coefficients and 
      complex roots [abstract base]
  */
  class quadratic_complex : public quadratic_real_coeff {
  public:

    virtual ~quadratic_complex() {}

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ .
    */
    virtual int solve_r(const double a2, const double b2, const double c2, 
			double &x1, double &x2);

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two complex solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ 
    */
    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, std::complex<double> &x2);

    /** \brief Solves the complex polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two complex solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ 
    */
    virtual int solve_c(const std::complex<double> a2, 
			const std::complex<double> b2, 
			const std::complex<double> c2, 
			std::complex<double> &x1, 
			std::complex<double> &x2)=0;

    /// Return a string denoting the type ("quadratic_complex")
    const char *type() { return "quadratic_complex"; }
  };  

  /** 
      \brief Solve a cubic polynomial with real coefficients and real roots
      [abstract base]
  */
  class cubic_real {
  public:

    virtual ~cubic_real() {}

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the three 
	solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_r(const double a3, const double b3, const double c3, 
			const double d3, double &x1, double &x2, 
			double &x3)=0;

    /// Return a string denoting the type ("cubic_real")
    const char *type() { return "cubic_real"; }
  };

  /** 
      \brief Solve a cubic polynomial with real coefficients and 
      complex roots [abstract base]
  */
  class cubic_real_coeff : public cubic_real {

  public:

    virtual ~cubic_real_coeff() {}

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the three 
	solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_r(const double a3, const double b3, const double c3, 
			const double d3, double &x1, double &x2, double &x3);

    /** \brief Solves the polynomial 
	 \f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ 
	giving the real solution \f$ x=x_1 \f$ and two complex solutions 
	 \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, std::complex<double> &x2,
			 std::complex<double> &x3)=0;

    /// Return a string denoting the type ("cubic_real_coeff")
    const char *type() { return "cubic_real_coeff"; }
  };

  /** 
      \brief Solve a cubic polynomial with complex coefficients and 
      complex roots [abstract base]
  */
  class cubic_complex : public cubic_real_coeff {

  public:

    virtual ~cubic_complex() {}

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the three 
	solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_r(const double a3, const double b3, const double c3, 
			const double d3, double &x1, double &x2, double &x3);

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the real 
	solution \f$ x=x_1 \f$ and two complex solutions 
	\f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, std::complex<double> &x2,
			 std::complex<double> &x3);
    
    /** \brief Solves the complex polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ 
	giving the three complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and 
	\f$ x=x_3 \f$ .
    */
    virtual int solve_c(const std::complex<double> a3, 
			const std::complex<double> b3, 
			const std::complex<double> c3, 
			const std::complex<double> d3, 
			std::complex<double> &x1, std::complex<double> &x2, 
			std::complex<double> &x3)=0;

    /// Return a string denoting the type ("cubic_complex")
    const char *type() { return "cubic_complex"; }
  };

  /** 
      \brief Solve a quartic polynomial with real coefficients and 
      real roots [abstract base]
  */
  class quartic_real {

  public:

    virtual ~quartic_real() {}
    
    /** \brief Solves the polynomial 
	\f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	\f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, 
			double &x1, double &x2, 
			double &x3, double &x4)=0;

    /// Return a string denoting the type ("quartic_real")
    const char *type() { return "quartic_real"; }
  };

  /** 
      \brief Solve a quartic polynomial with real coefficients and 
      complex roots [abstract base]
  */
  class quartic_real_coeff : public quartic_real {

  public:

    virtual ~quartic_real_coeff() {}

    /** \brief Solves the polynomial 
	\f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	\f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, double &x3, double &x4);

    /**  \brief Solves the polynomial 
	 \f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	 giving the four complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	 \f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, 
			 std::complex<double> &x4)=0;

    /// Return a string denoting the type ("quartic_real_coeff")
    const char *type() { return "quartic_real_coeff"; }
  };

  /** 
      \brief Solve a quartic polynomial with complex coefficients and 
      complex roots [abstract base]
  */
  class quartic_complex : public quartic_real_coeff {

  public:

    virtual ~quartic_complex() {}

    /** \brief Solves the polynomial 
	\f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	\f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, 
			double &x3, double &x4);

    /**  \brief Solves the polynomial 
	 \f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	 giving the four complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	 \f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, std::complex<double> &x4);

    /** \brief Solves the complex polynomial 
	 \f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	 \f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_c(const std::complex<double> a4, 
			const std::complex<double> b4, 
			const std::complex<double> c4, 
			const std::complex<double> d4, 
			const std::complex<double> e4, 
			std::complex<double> &x1, 
			std::complex<double> &x2, std::complex<double> &x3,
			std::complex<double> &x4)=0;

    /// Return a string denoting the type ("quartic_complex")
    const char *type() { return "quartic_complex"; }
  };

  /** \brief Solve a general polynomial with real
      coefficients and complex roots [abstract base]
  */
  class poly_real_coeff : public quadratic_real_coeff,
    public cubic_real_coeff, public quartic_real_coeff {

  public:
    
    virtual ~poly_real_coeff() {}
    
    /** 
	\brief Solve the n-th order polynomial
	
        The coefficients are stored in co[], with the leading coefficient
	as co[0] and the constant term as co[n]. The roots are returned
	in ro[0],...,ro[n-1].
    */
    virtual int solve_rc(int n, const double co[], 
			 std::complex<double> ro[])=0;

    /// Return a string denoting the type ("poly_real_coeff")
    const char *type() { return "poly_real_coeff"; }
  };

  /** \brief Solve a general polynomial with complex
      coefficients [abstract base]
  */
  class poly_complex : public quadratic_complex,
    public cubic_complex, public quartic_complex {

    public:

    virtual ~poly_complex() {}
    
    /** 
	\brief Solve the n-th order polynomial
	
        The coefficients are stored in co[], with the leading coefficient
	as co[0] and the constant term as co[n]. The roots are returned
	in ro[0],...,ro[n-1].
    */
    virtual int solve_c(int n, const std::complex<double> co[], 
			std::complex<double> ro[])=0;
    
    /// Polish the roots 
    virtual int polish_c(int n, const std::complex<double> co[],
			 std::complex<double> *ro)=0;

    /// Return a string denoting the type ("poly_complex")
    const char *type() { return "poly_complex"; }
  };

  /** 
      \brief Solve a cubic with real coefficients and complex roots 
      (CERNLIB)

      This follows the original Fortran code except in the function
      rrteq3(), the roots are returned in x[0], x[1], and x[2] instead
      of x[1], x[2], and x[3]. (The arrays y and z were already
      zero-indexed in the CERN routine.) Similar to the original, in
      the case of complex roots, x[0] is the real root and x[1] and
      x[2] contain the real and imaginary parts of the complex roots.

      Another small change is that the discriminant for the resolvent
      cubic is evaluated slightly differently in order to improve the
      properties in the case where the roots are not of order unity.
      The default CERNLIB behavior can be restored by setting
      improve_scale to \c false.
  */
  class cern_cubic_real_coeff : public cubic_real_coeff {

  public:

    cern_cubic_real_coeff() {
      eps=1.0e-6;
      delta=1.0e-15;
      improve_scale=true;
    }

    /// Numerical tolerance (default \f$ 10^{-6} \f$)
    double eps;

    /// Numerical tolerance (default \f$ 10^{-15} \f$)
    double delta;

    /// Improve algorithm for bad-scaled roots (default true)
    bool improve_scale;

    virtual ~cern_cubic_real_coeff() {}

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the real 
	solution \f$ x=x_1 \f$ and two complex solutions 
	\f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, 
			 std::complex<double> &x2, std::complex<double> &x3);

    /** \brief The CERNLIB-like interface
    */	
    virtual int rrteq3(double r, double s, double t, double x[], double &d);

    /// Return a string denoting the type ("cern_cubic_real_coeff")
    const char *type() { return "cern_cubic_real_coeff"; }
  };

  /** 
      \brief Solve a quartic with real coefficients and complex roots (CERNLIB)
  */
  class cern_quartic_real_coeff : public quartic_real_coeff {

  public:

    virtual ~cern_quartic_real_coeff() {}

    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, std::complex<double> &x4);

    /// The CERNLIB-like interface
    virtual int rrteq4(double a, double b, double c, double d, 
		       std::complex<double> z[], double &dc, 
		       int &mt);

    /// Return a string denoting the type ("cern_quartic_real_coeff")
    const char *type() { return "cern_quartic_real_coeff"; }

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The object to solve for the associated cubic
    cern_cubic_real_coeff cub_obj;

#endif

  };

  /** 
      \brief Solve a quadratic with real coefficients and complex roots (GSL)
  */
  class gsl_quadratic_real_coeff : public quadratic_real_coeff {

  public:

    virtual ~gsl_quadratic_real_coeff() {}

    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, std::complex<double> &x2);

    /// Return a string denoting the type ("gsl_quadratic_real_coeff")
    const char *type() { return "gsl_quadratic_real_coeff"; }

  };

  /** 
      \brief Solve a cubic with real coefficients and complex roots (GSL)
  */
  class gsl_cubic_real_coeff : public cubic_real_coeff {

  public:

    virtual ~gsl_cubic_real_coeff() {}

    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, 
			 std::complex<double> &x2, 
			 std::complex<double> &x3);

    /// Return a string denoting the type ("gsl_cubic_real_coeff")
    const char *type() { return "gsl_cubic_real_coeff"; }

    /** 
	\brief An alternative to \c gsl_poly_complex_solve_cubic()
	
	This is an alternative to the function
	<tt>gsl_poly_complex_solve_cubic()</tt> with some small
	corrections to ensure finite values for some cubics. See
	<tt>src/other/poly_ts.cpp</tt> for more.
    */
    int gsl_poly_complex_solve_cubic2(double a, double b, double c, 
				      gsl_complex *z0, gsl_complex *z1, 
				      gsl_complex *z2);
    
  };

  /** 
      \brief Solve a quartic with real coefficients and real roots (GSL)

      This class internally uses the GSL functions to solve the
      resolvent cubic and associated quadratics, while
      \ref gsl_quartic_real2 contains explicit code to solve
      them instead.

      \future Optimize value of \c cube_root_tol and compare
      more clearly to \ref gsl_quartic_real2
  */
  class gsl_quartic_real : public quartic_real {
    
  public:
    
    gsl_quartic_real() {
      cube_root_tol=1.0e-7;
    }

    virtual ~gsl_quartic_real() {}

    /** 
	\brief A tolerance for determining the proper cube root 
	(default \f$ 10^{-4} \f$ )
    */
    double cube_root_tol;

    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, 
			double &x3, double &x4);

    /// Return a string denoting the type ("gsl_quartic_real")
    const char *type() { return "gsl_quartic_real"; }

  };

  /** 
      \brief Solve a quartic with real coefficients and real roots (GSL)

      This class directly solves 
      resolvent cubic and associated quadratics without using 
      the GSL functions (as done in \ref gsl_quartic_real).
      
      \future Optimize value of \c cube_root_tol and compare
      more clearly to \ref gsl_quartic_real
  */
  class gsl_quartic_real2 : public quartic_real {

  public:

    gsl_quartic_real2() {
      cube_root_tol=1.0e-7;
    }

    virtual ~gsl_quartic_real2() {}

    /** 
	\brief A tolerance for determining the proper cube root 
	(default \f$ 10^{-7} \f$ )
    */
    double cube_root_tol;

    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, 
			double &x3, double &x4);

    /// Return a string denoting the type ("gsl_quartic_real2")
    const char *type() { return "gsl_quartic_real2"; }
  };

  /** 
      \brief Solve a general polynomial with real coefficients (GSL)
  */
  class gsl_poly_real_coeff : public poly_real_coeff {

  public:

    gsl_poly_real_coeff();

    virtual ~gsl_poly_real_coeff();

    virtual int solve_rc(int n, const double co[], 
			 std::complex<double> ro[]);

    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, 
			 std::complex<double> &x2, 
			 std::complex<double> &x3);

    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, 
			 std::complex<double> &x2);

    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, std::complex<double> &x4);

    /// Return a string denoting the type ("gsl_poly_real_coeff")
    const char *type() { return "gsl_poly_real_coeff"; }

  protected:

#ifndef DOXYGEN_INTERNAL

    /// Workspace for quadratic polynomials
    gsl_poly_complex_workspace *w2;

    /// Workspace for cubic polynomials
    gsl_poly_complex_workspace *w3;

    /// Workspace for quartic polynomials
    gsl_poly_complex_workspace *w4;

    /// Workspace for general polynomials
    gsl_poly_complex_workspace *wgen;

    /// The size of the workspace \ref wgen
    int gen_size;

#endif

  };

  /** 
      \brief Solve a quadratic with complex coefficients and complex roots
  */
  class quadratic_std_complex : public quadratic_complex {

  public:

    virtual ~quadratic_std_complex() {}

    virtual int solve_c(const std::complex<double> a2, 
			const std::complex<double> b2, 
			const std::complex<double> c2, 
			std::complex<double> &x1, std::complex<double> &x2);

    /// Return a string denoting the type ("quadratic_std_complex")
    const char *type() { return "quadratic_std_complex"; }
  };

  /** 
      \brief Solve a cubic with complex coefficients and complex roots
  */
  class cubic_std_complex : public cubic_complex {

  public:

    virtual ~cubic_std_complex() {}

    virtual int solve_c(const std::complex<double> a3, 
			const std::complex<double> b3, 
			const std::complex<double> c3, 
			const std::complex<double> d3, 
			std::complex<double> &x1, std::complex<double> &x2, 
			std::complex<double> &x3);

    /// Return a string denoting the type ("cubic_std_complex")
    const char *type() { return "cubic_std_complex"; }
  };

  /** 
      \brief Solve a quartic with real coefficients and real roots
  */
  class simple_quartic_real : public quartic_real {

  public:

    simple_quartic_real() {
      cube_root_tol=1.0e-6;
    }

    virtual ~simple_quartic_real() {}

    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, double &x3, double &x4);

    /// Return a string denoting the type ("simple_quartic_real")
    const char *type() { return "simple_quartic_real"; }

    /** 
	\brief A tolerance for determining the proper cube root 
	(default \f$ 10^{-6} \f$ )
    */
    double cube_root_tol;
  };
  
  /** 
      \brief Solve a quartic with complex coefficients and complex roots
  */
  class simple_quartic_complex : public quartic_complex {

  public:

    virtual ~simple_quartic_complex() {}

    virtual int solve_c(const std::complex<double> a4, 
			const std::complex<double> b4, 
			const std::complex<double> c4, 
			const std::complex<double> d4, 
			const std::complex<double> e4, 
			std::complex<double> &x1, 
			std::complex<double> &x2, std::complex<double> &x3,
			std::complex<double> &x4);

    /// Return a string denoting the type ("simple_quartic_complex")
    const char *type() { return "simple_quartic_complex"; }

#ifndef DOXYGENP

  protected:

    /// The object to solve for the associated cubic
    cubic_std_complex cub_obj;
    
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
