/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TWOD_INTP_H
#define O2SCL_TWOD_INTP_H

#include <iostream>
#include <string>
#include <vector>
#include <o2scl/smart_interp.h>
#include <o2scl/omatrix_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Two-dimensional interpolation class 

      This class implements two-dimensional interpolation. Derivatives
      and integrals along both x- and y-directions can be computed.

      The storage of the matrix to be specified in the function
      set_data() and this function is designed to follow the format:
      \f[
      \begin{array}{cccc}
      & x_0 & x_1 & x_2 \\
      y_0 & M_{00} & M_{01} & M_{02} \\
      y_1 & M_{10} & M_{11} & M_{12} \\
      y_2 & M_{20} & M_{21} & M_{22}
      \end{array}
      \f]
      thus the matrix should be \c M[i][j] where \c i is the y index
      and \c j is the x index. (See also the discussion in the
      User's guide in the section called \ref rowcol_subsect.)

      The function set_data(), does not copy the data, it 
      stores pointers to the data. If the data is modified, then the
      function reset_interp() must be called to reset the
      interpolation information with the original pointer information.
      The storage for the data, including the arrays \c x_grid and \c
      y_grid are all managed by the user.

      By default, cubic spline interpolation with natural boundary
      conditions is used. This can be changed with the set_interp()
      function.

      There is an example for the usage of this class given
      in <tt>examples/ex_twod_intp.cpp</tt>.

      \future Could also include mixed second/first derivatives:
      deriv_xxy and deriv_xyy.
      \future Implement an improved caching system in case, for example
      \c xfirst is true and the last interpolation used the same
      value of \c x.

      \comment

      There's a whole list of things to think about here:

      1. One could
      generalize the matrix and vector types. This would demand
      adding subvector types for the interpolation, etc. 

      2. One could copy the user data instead of storing pointers to it.

      3. One could include get() and set() methods, but one would
      have to be careful about when to reset the interpolation.

      4. xy order vs. yx order for the matrix

      5/16/08 - There has been some talk about GSL implementing 2d
      interpolation, and I think the table3d class will be much more
      useful in the end. Thus, I think it's probably best to delay
      working on this for awhile until GSL 2d interpolation or table3d
      are more settled.

      \endcomment

  */
  class twod_intp {
  public:
    
    twod_intp();

    virtual ~twod_intp();
    
    /** 
	\brief Initialize the data for the 2-dimensional interpolation
	
        The interpolation type (passed directly to int_type) is
	specified in \c int_type and the data is specified in
	\c data. The data should be arranged so that the first
	array index is the y-value (the "row") and the second array
	index is the x-value (the "column"). The arrays \c xfun 
	and \c yfun specify the two independent
	variables. \c xfun should be an array of length
	\c nx , and should be an array of length \c ny. The
	array \c data should be a two-dimensional array of size
	[ny][nx].

	If \c x_first is true, then set_data() creates interpolation
	objects for each of the rows. Calls to interp() then uses
	these to create a column at the specified value of \c x. An
	interpolation object is created at this column to find the
	value of the function at the specified value \c y. If \c
	x_first is false, the opposite strategy is employed. These two
	options may give slightly different results. 
	
    */
    int set_data(size_t n_x, size_t n_y, ovector &x_grid,
		 ovector &y_grid, omatrix &data, 
		 bool x_first=true);
    
    /** 
	\brief Reset the stored interpolation since the data has changed
	
        This will return an error if the set_data() has not been called
    */
    int reset_interp();

    /** \brief Perform the 2-d interpolation 
     */
    double interp(double x, double y);

    /** \brief Compute the partial derivative in the x-direction
     */
    double deriv_x(double x, double y);

    /** \brief Compute the partial second derivative in the x-direction
     */
    double deriv2_x(double x, double y);

    /** \brief Compute the integral in the x-direction between x=x0
	and x=x1
    */
    double integ_x(double x0, double x1, double y);

    /** \brief Compute the partial derivative in the y-direction
     */
    double deriv_y(double x, double y);

    /** \brief Compute the partial second derivative in the y-direction
     */
    double deriv2_y(double x, double y);

    /** \brief Compute the integral in the y-direction between y=y0
	and y=y1
    */
    double integ_y(double x, double y0, double y1);

    /** \brief Compute the mixed partial derivative 
	\f$ \frac{\partial^2 f}{\partial x \partial y} \f$
    */
    double deriv_xy(double x, double y);

    /** 
	\brief Specify the base interpolation objects to use

	This allows the user to provide new interpolation objects for
	use in the two-dimensional interpolation. For
	example,

	\code
	twod_intp ti;
	ovector x(20), y(40);
	omatrix d(40,20);

	// Fill x, y, and d with the data, choose linear interpolation
	// instead of the default cubic spline
	def_interp_mgr<ovector_const_view,linear_interp> dim1;
	def_interp_mgr<ovector_const_subvector,linear_interp> dim2;
	
	ti.set_interp(dim1,dim2);
	ti.set_data(20,40,x,y,d,true);
	\endcode

	This function automatically calls reset_interp() and
	reset_data() if the data has already been set to reset the
	internal interpolation objects.
    */
    int set_interp(base_interp_mgr<ovector_const_view> &b1,
		   base_interp_mgr<ovector_const_subvector> &b2) {
      bimp1=&b1;
      bimp2=&b2;
      return 0;
    }

    /**
       \brief Inform the class the data has been modified or
       changed in a way that set_data() will need to be called 
       again.
    */
    int unset_data();

#ifndef DOXYGENP
    
  protected:

    /// \name Default interpolation manager
    //@{
    def_interp_mgr<ovector_const_view,cspline_interp> dim1;
    def_interp_mgr<ovector_const_subvector,cspline_interp> dim2;
    //@}

    /// \name Interpolation manager pointers
    //@{
    base_interp_mgr<ovector_const_view> *bimp1;
    base_interp_mgr<ovector_const_subvector> *bimp2;
    //@}

    /// The array of sm_interp pointers 
    sm_interp_vec **si;
    /// An array of omatrix_rows
    omatrix_row **ari;
    /// An array of omatrix_cols
    omatrix_col **aci;

    /// The number of x grid points
    size_t nx;

    /// The number of y grid points
    size_t ny;

    /// If true, the user has specified the base interpolation objects
    bool it_set;

    /// True if the x interpolation should be done first
    bool xfirst;

    /// True if the data has been specified by the user
    bool data_set;

    /// The x grid
    ovector *xfun;

    /// The y grid
    ovector *yfun;

    /// The data
    omatrix *datap;

    /// Allocate the base interpolation objects
    int alloc_interp(size_t n);

    /// Free the base interpolation objects
    int free_interp();

    /// Free the data and interpolation objects
    int free_data();

#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif



