/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_VEC_STATS_H
#define O2SCL_VEC_STATS_H

/** \file vec_stats.h
    \brief File containing statistics template functions

    \future Add chi squared test?
    \future Consider generalizing to other data types
*/

#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/err_hnd.h>
#include <gsl/gsl_ieee_utils.h>
#include <gsl/gsl_sort.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Compute the mean of the first \c n elements of a vector

      If \c n is zero, this will return zero.
  */
  template<class vec_t>
    double vector_mean(size_t n, vec_t &data) {
    
    long double mean=0.0;
    for(size_t i=0;i<n;i++) {
      mean+=(data[i]-mean)/(i+1);
    }
    return mean;
  }

  /** \brief Compute variance with specified mean known in advance
      
      This function computes
      \f[
      \frac{1}{N} \sum_{i} \left( x_i - \mu \right)^2
      \f]
      where the value of \f$ \mu \f$ is given in \c mean. 
  */
  template<class vec_t>
    double vector_variance_fmean(size_t n, vec_t &data, 
				 double mean) {
    long double var=0.0;
    for(size_t i=0;i<n;i++) {
      long double delta=(data[i]-mean);
      var+=(delta*delta-var)/(i+1);
    }
    return var;
  }

  /** \brief Compute the variance with specified mean

      This function computes
      \f[
      \frac{1}{N-1} \sum_{i} \left( x_i - \mu \right)^2
      \f]
      where the value of \f$ \mu \f$ is given in \c mean.
      
      If \c n is 0 or 1, this function will call the error
      handler.
  */
  template<class vec_t>
    double vector_variance(size_t n, vec_t &data, double mean) {

    if (n<2) {
      O2SCL_ERR2_RET("Cannot compute variance with less than 2 elements",
		     " in vector_variance().",gsl_einval);
    }

    double var=vector_variance_fmean<vec_t>(n,data,mean);
    return var*n/(n-1);
  }

  /** \brief Compute the variance 

      This function computes
      \f[
      \frac{1}{N-1} \sum_{i} \left( x_i - \mu \right)n^2
      \f]
      where \f$ \mu \f$ is the mean computed with \ref vector_mean().
      
      If \c n is 0 or 1, this function will call the error
      handler.
  */
  template<class vec_t>
    double vector_variance(size_t n, vec_t &data) {

    if (n<2) {
      O2SCL_ERR2_RET("Cannot compute variance with less than 2 elements",
		     " in vector_variance().",gsl_einval);
    }
    
    double mean=vector_mean<vec_t>(n,data);
    double var=vector_variance_fmean<vec_t>(n,data,mean);
    return var*n/(n-1);
  }

  /** \brief Standard deviation with specified mean known in advance

      This function computes
      \f[
      \sqrt{\frac{1}{N} \sum_{i} \left( x_i - \mu \right)^2}
      \f]
      where the value of \f$ \mu \f$ is given in \c mean. 

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_stddev_fmean(size_t n, vec_t &data, 
			       double mean) {
    double sd=vector_variance_fmean<vec_t>(n,data,mean);
    return std::sqrt(sd);
  }

  /** \brief Standard deviation with specified mean

      This function computes
      \f[
      \sqrt{\frac{1}{N-1} \sum_{i} \left( x_i - \mu \right)^2}
      \f]
      where \f$ \mu \f$ is the mean computed with \ref vector_mean().

      If \c n is 0 or 1, this function will call the error
      handler.
  */
  template<class vec_t>
    double vector_stddev(size_t n, vec_t &data) {
    
    if (n<2) {
      O2SCL_ERR2_RET("Cannot compute std. dev. with less than 2 elements",
		     " in vector_stddev().",gsl_einval);
    }
    
    double mean=vector_mean<vec_t>(n,data);
    double var=vector_variance_fmean<vec_t>(n,data,mean);
    return std::sqrt(var*n/(n-1));
  }

  /** \brief Standard deviation with specified mean

      This function computes
      \f[
      \sqrt{\frac{1}{N-1} \sum_{i} \left( x_i - \mu \right)^2}
      \f]
      where the value of \f$ \mu \f$ is given in \c mean. 

      If \c n is 0 or 1, this function will call the error
      handler.
  */
  template<class vec_t>
    double vector_stddev(size_t n, vec_t &data, 
			 double mean) {
    
    if (n<2) {
      O2SCL_ERR2_RET("Cannot compute std. dev. with less than 2 elements",
		     " in vector_stddev().",gsl_einval);
    }
    
    double sd=vector_variance_fmean<vec_t>(n,data,mean);
    return std::sqrt(sd*n/(n-1));
  }
  
  /** \brief Absolute deviation from the specified mean

      This function computes
      \f[
      \sum_i | x_i - \mu |
      \f]
      where the value of \f$ \mu \f$ is given in \c mean. 

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_absdev(size_t n, vec_t &data, 
			 double mean) {

    if (n==0) return 0.0;

    long double sum=0.0;
    for(size_t i=0;i<n;i++) {
      sum+=fabs(data[i]-mean);
    }
    return sum/n;
  }
  
  /** \brief Absolute deviation from the computed mean

      This function computes
      \f[
      \sum_i \left| x_i - \frac{1}{N} \left(\sum_j x_j\right) \right|
      \f]
      where the value of \f$ \mu \f$ is given in \c mean. 

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_absdev(size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    return vector_absdev(n,data,mean);
  }

  /** \brief Skewness with specified mean and standard deviation

      This function computes 
      \f[
      \frac{1}{N} \sum_i \left[ 
      \frac{ \left(x_i - \mu \right)}{ \sigma }\right]^3
      \f]
      where the values of \f$ \mu \f$ and \f$ \sigma \f$ 
      are given in \c mean and \c stddev.

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_skew(size_t n, vec_t &data, double mean,
		       double stddev) {
    long double skew=0.0;
    for(size_t i=0;i<n;i++) {
      long double x=(data[i]-mean)/stddev;
      skew+=(x*x*x-skew)/(i+1);
    }
    return skew;
  }

  /** \brief Skewness with computed mean and standard deviation

      This function computes 
      \f[
      \frac{1}{N} \sum_i \left[ 
      \frac{ \left(x_i - \mu \right)}{ \sigma }\right]^3
      \f]
      where the values of \f$ \mu \f$ and \f$ \sigma \f$ 
      are computed using \ref vector_mean() and \ref vector_stddev().

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_skew(size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    double sd=vector_stddev<vec_t>(n,data,mean);
    return vector_skew(n,data,mean,sd);
  }

  /** \brief Kurtosis with specified mean and standard deviation

      This function computes 
      \f[
      -3 + \frac{1}{N} \sum_i \left[ 
      \frac{ \left(x_i - \mu \right)}{ \sigma }\right]^4
      \f]
      where the values of \f$ \mu \f$ and \f$ \sigma \f$ 
      are given in \c mean and \c stddev.

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_kurtosis(size_t n, vec_t &data, double mean,
			   double stddev) {
    long double avg=0.0;
    for(size_t i=0;i<n;i++) {
      long double x=(data[i]-mean)/stddev;
      avg+=(x*x*x*x-avg)/(i+1);
    }
    return avg-3.0;
  }

  /** \brief Kurtosis with computed mean and standard deviation

      This function computes 
      \f[
      -3 + \frac{1}{N} \sum_i \left[ 
      \frac{ \left(x_i - \mu \right)}{ \sigma }\right]^4
      \f]
      where the values of \f$ \mu \f$ and \f$ \sigma \f$ 
      are computed using \ref vector_mean() and \ref vector_stddev().

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_kurtosis(size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    double sd=vector_stddev<vec_t>(n,data,mean);
    return vector_kurtosis(n,data,mean,sd);
  }

  /** \brief Lag-1 autocorrelation

      This function computes
      \f[
      \left[
      \sum_i \left(x_i - \mu\right) \left(x_{i-1} - \mu \right)
      \right] \left[ 
      \sum_i \left(x_i - \mu\right)^2 
      \right]^{-1}
      \f]

      If \c n is zero, this function will call the error handler.
  */
  template<class vec_t>
    double vector_lag1_autocorr(size_t n, vec_t &data, double mean) {
    
    if (n<1) {
      O2SCL_ERR2_RET("Cannot compute lag1 with no elements",
		     " in vector_lag1_autocorr().",gsl_einval);
    }

    long double q=0.0;
    long double v=(data[0]-mean)*(data[0]-mean);
    for(size_t i=1;i<n;i++) {
      long double delta0=data[i-1]-mean;
      long double delta1=data[i]-mean;
      q+=(delta0*delta1-q)/(i+1);
      v+=(delta1*delta1-v)/(i+1);
    }
    return q/v;
  }

  /** \brief Lag-1 autocorrelation

      This function computes
      \f[
      \left[
      \sum_i \left(x_i - \mu\right) \left(x_{i-1} - \mu \right)
      \right] \left[ 
      \sum_i \left(x_i - \mu\right)^2 
      \right]^{-1}
      \f]
      
      If \c n is zero, this function will call the error handler.
  */
  template<class vec_t>
    double vector_lag1_autocorr(size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    return vector_lag1_autocorr(n,data,mean);
  }

  /** \brief Covariance

      This function computes
      \f[
      \frac{1}{n-1} \sum_i \left(x_i - {\bar{x}}\right)
      \left(y_i - {\bar{y}}\right)
      \f]
      where \f$ {\bar{x}} \f$ and \f$ {\bar{y}} \f$ are specified
      in \c mean1 and \c mean2, respectively.

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_covariance(size_t n, vec_t &data1, vec_t &data2,
			     double mean1, double mean2) {
    double covar=0.0;
    for(size_t i=0;i<n;i++) {
      double delta1=(data1[i]-mean1);
      double delta2=(data2[i]-mean2);
      covar+=(delta1*delta2-covar)/(i+1);
    }
    return covar;
  }

  /** \brief Covariance

      This function computes
      \f[
      \frac{1}{n-1} \sum_i \left(x_i - {\bar{x}}\right)
      \left(y_i - {\bar{y}}\right)
      \f]
      where \f$ {\bar{x}} \f$ and \f$ {\bar{y}} \f$ are 
      the averages of \c data1 and \c data2 and are computed
      automatically.

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_covariance(size_t n, vec_t &data1, vec_t &data2) {
    double covar=0.0;
    double mean1=vector_mean<vec_t>(n,data1);
    double mean2=vector_mean<vec_t>(n,data2);
    for(size_t i=0;i<n;i++) {
      long double delta1=(data1[i]-mean1);
      long double delta2=(data2[i]-mean2);
      covar+=(delta1*delta2-covar)/(i+1);
    }
    return covar;
  }
  
  /** \brief Pearson's correlation
      
      \comment
      r = cov(x, y) / (\Hat\sigma_x \Hat\sigma_y)
      = {1/(n-1) \sum (x_i - \Hat x) (y_i - \Hat y)
      \over
      \sqrt{1/(n-1) \sum (x_i - \Hat x)^2} \sqrt{1/(n-1) 
      \sum (y_i - \Hat y)^2}
      }
      \endcomment

      If \c n is zero, this function will call the error handler.
  */
  template<class vec_t>
    double vector_correlation(size_t n, vec_t &data1, vec_t &data2) {
    size_t i;

    if (n<1) {
      O2SCL_ERR2_RET("Cannot compute correlation with no elements",
		     " in vector_correlation().",gsl_einval);
    }

    double sum_xsq = 0.0;
    double sum_ysq = 0.0;
    double sum_cross = 0.0;
    double ratio;
    double delta_x, delta_y;
    double mean_x, mean_y;
    double r;

    /*
     * Compute:
     * sum_xsq = Sum [ (x_i - mu_x)^2 ],
     * sum_ysq = Sum [ (y_i - mu_y)^2 ] and
     * sum_cross = Sum [ (x_i - mu_x) * (y_i - mu_y) ]
     * using the above relation from Welford's paper
     */

    mean_x = data1[0];
    mean_y = data2[0];

    for (i = 1; i < n; ++i) {
      ratio = i / (i + 1.0);
      delta_x = data1[i] - mean_x;
      delta_y = data2[i] - mean_y;
      sum_xsq += delta_x * delta_x * ratio;
      sum_ysq += delta_y * delta_y * ratio;
      sum_cross += delta_x * delta_y * ratio;
      mean_x += delta_x / (i + 1.0);
      mean_y += delta_y / (i + 1.0);
    }
    
    r = sum_cross / (std::sqrt(sum_xsq) * std::sqrt(sum_ysq));
    
    return r;
  }

  /** \brief Pooled variance

      \todo Document this
      
      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_pvariance(size_t n1, vec_t &data1, 
			    size_t n2, vec_t &data2) {
    double var1=vector_variance<vec_t>(n1,data1);
    double var2=vector_variance<vec_t>(n2,data2);
    return (((n1-1)*var1)+((n2-1)*var2))/(n1+n2-2);
  }

  /** \brief Quantile

      \todo Document this

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_quantile_sorted(size_t n, vec_t &data, 
				  const double f) {

    double index=f*(n-1);
    size_t lhs=((size_t)index);
    double delta=index-lhs;
    if (n==0) return 0.0;
    if (lhs==n-1) return data[lhs];
    return (1-delta)*data[lhs]+delta*data[lhs+1];
  }
  
  /** \brief Median of sorted data

      \todo Document this

      If \c n is zero, this function will return zero
      without calling the error handler.
  */
  template<class vec_t>
    double vector_median_sorted(size_t n, vec_t &data) {
    
    if (n==0) return 0.0;
    
    size_t lhs=(n-1)/2;
    size_t rhs=n/2;
    
    if (lhs==rhs) return data[lhs];

    return (data[lhs]+data[rhs])/2.0;
  }

  /** \brief Compute the mean of weighted data

      \comment
      M(n) = M(n-1) + (data[n] - M(n-1)) (w(n)/(W(n-1) + w(n)))
      W(n) = W(n-1) + w(n)
      \endcomment
  */
  template<class vec_t>
    double wvector_mean(size_t n, vec_t &data, vec_t &weights) {

    long double wmean=0.0;
    long double W=0.0;
    for(size_t i=0;i<n;i++) {
      double wi=weights[i];
      if (wi>0.0) {
	W+=wi;
	wmean+=(data[i]-wmean)*(wi/W);
      }
    }
    
    return wmean;
  }

  /** \brief Compute a factor for weighted data

      This function is used in \ref wvector_variance(size_t n,
      vec_t &data, vec2_t &weights, double wmean) and \ref
      wvector_stddev(size_t n, vec_t &data, vec2_t &weights, double
      wmean) .
  */
  template<class vec_t>
    double wvector_factor(size_t n, vec_t &weights) {
    
    long double a=0.0;
    long double b=0.0;
    long double factor;
    for(size_t i=0;i<n;i++) {
      double wi=weights[i];
      if (wi>0.0) {
	a+=wi;
	b+=wi*wi;
      }
    }
    factor=a*a/(a*a-b);
    return factor;
  }
  
  /** \brief Compute the variance of a weighted vector 
      with a mean known in advance
  */
  template<class vec_t, class vec2_t>
    double wvector_variance_fmean(size_t n, vec_t &data,
				  vec_t &weights, double wmean) {
    long double wvariance=0.0;
    long double W=0.0;
    for(size_t i=0;i<n;i++) {
      double wi=weights[i];
      if (wi>0.0) {
	const long double delta=data[i]-wmean;
	W+=wi;
	wvariance+=(delta*delta-wvariance)*(wi/W);
      }
    }

    return wvariance;
  }

  /** \brief Compute the variance of a weighted vector where mean
      is computed automatically
   */
  template<class vec_t, class vec2_t>
    double wvector_variance(size_t n, vec_t &data,
			    vec2_t &weights) {

    double wmean=wvector_mean(n,data,weights);
    return wvector_variance_fmean(n,data,weights,wmean);
  }

  /** \brief Compute the variance of a weighted vector with
      specified mean
  */
  template<class vec_t, class vec2_t>
    double wvector_variance(size_t n, vec_t &data,
			    vec2_t &weights, double wmean) {

    const double variance=wvector_variance_fmean
      (n,data,weights,wmean);
    const double scale=wvector_factor(n,weights);
    const double wvar=scale*variance;
    return wvar;
  }

  /** \brief Compute the standard deviation of a weighted vector 
      with a mean known in advance
  */
  template<class vec_t, class vec2_t>
    double wvector_stddev_fmean(size_t n, vec_t &data,
				vec2_t &weights, double wmean) {
    return sqrt(wvector_variance_fmean(n,data,weights,wmean));
  }

  /** \brief Compute the standard deviation of a weighted vector where mean
      is computed automatically
   */
  template<class vec_t, class vec2_t>
    double wvector_stddev(size_t n, vec_t &data,
			  vec2_t &weights) {
    double wmean=wvector_mean(n,data,weights);
    return sqrt(wvector_variance_fmean(n,data,weights,wmean));
  }

  /** \brief Compute the standard deviation of a weighted vector with
      specified mean
  */
  template<class vec_t, class vec2_t>
    double wvector_stddev(size_t n, vec_t &data,
			  vec2_t &weights, double wmean) {
    const double variance=wvector_variance_fmean
      (n,data,weights,wmean);
    const double scale=wvector_factor(n,weights);
    const double wvar=scale*variance;
    return sqrt(wvar);
  }

  /// Desc
  template<class vec_t, class vec2_t>
    double wvector_sumsq_fmean(size_t n, vec_t &data,
			       vec2_t &weights, double wmean) {
    long double wtss=0.0;
    for(size_t i=0;i<n;i++) {
      double wi=weights[i];
      if (wi>0.0) {
	const long double delta=data[i]-wmean;
	wtss+=wi*delta*delta;
      }
    }
    
    return wtss;
  }

  /// Desc
  template<class vec_t, class vec2_t>
    double wvector_sumsq(size_t n, vec_t &data,
			 vec2_t &weights) {
    
    double wmean=wvector_mean(n,data,weights);
    return wvector_sumsq(n,data,weights,wmean);
  }

  /// Desc
  template<class vec_t, class vec2_t> 
    double wvector_absdev(size_t n, vec_t &data, vec2_t &weights, 
			  double wmean) {
    long double wabsdev=0.0;
    long double W=0.0;
    for(size_t i=0;i<n;i++) {
      double wi=weights[i];
      if (wi>0.0) {
	const long double delta=fabs(data[i]-wmean);
	W+=wi;
	wabsdev+=(delta-wabsdev)*(wi/W);
      }
    }
    return wabsdev;
  }

  /// Desc
  template<class vec_t, class vec2_t>
    double wvector_skew(size_t n, vec_t &data, vec2_t &weights,
			double wmean, double wsd) {
    long double wskew=0.0;
    long double W=0.0;
    for(size_t i=0;i<n;i++) {
      double wi=weights[i];
      if (wi>0.0) {
	const long double x=(data[i]-wmean)/wsd;
	W+=wi;
	wskew+=(x*x*x-wskew)*(wi/W);
      }
    }
    return wskew;
  }
  
  /// Desc
  template<class vec_t, class vec2_t>
    double wvector_skew(size_t n, vec_t &data, vec2_t &weights) {
    double wmean=wvector_mean(n,data,weights);
    double wsd=wvector_stddev(n,data,weights,wmean);
    return wvector_skew(n,data,weights,wmean,wsd);
  }

  /// Desc
  template<class vec_t, class vec2_t>
    double wvector_kurtosis(size_t n, vec_t &data, vec2_t &weights,
			    double wmean, double wsd) {
    long double wavg=0.0;
    long double W=0.0;
    for(size_t i=0;i<n;i++) {
      double wi=weights[i];
      if (wi>0.0) {
	const long double x=(data[i]-wmean)/wsd;
	W+=wi;
	wavg+=(x*x*x*x-wavg)*(wi/W);
      }
    }
    return wavg-3.0;
  }

  /// Desc
  template<class vec_t, class vec2_t>
    double wvector_kurtosis(size_t n, vec_t &data, vec2_t &weights) {
    double wmean=wvector_mean(n,data,weights);
    double wsd=wvector_stddev(n,data,weights,wmean);
    return wvector_kurtosis(n,data,weights,wmean,wsd);
  }

#ifdef O2SCL_NEVER_DEFINED

  // These need to be redone similar to the GSL functions

  /// Weighted mean
  template<class vec_t, class vec2_t> 
    double wvector_mean(size_t n, vec_t &data, vec_t &stddev) {
    double sum1=0.0, err=0.0;
    for(size_t i=0;i<n;i++) {
      sum1+=data[i]/stddev[i]/stddev[i];
      err+=1.0/stddev[i]/stddev[i];
    }
    return sum1/err;
  }

  /// Weighted mean with error
  template<class vec_t, class vec2_t> 
    int wvector_mean_err(size_t n, vec_t &data, vec_t &stddev,
			 double &mean, double &err) {
    mean=0.0;
    err=0.0;
    for(size_t i=0;i<n;i++) {
      mean+=data[i]/stddev[i]/stddev[i];
      err+=1.0/stddev[i]/stddev[i];
    }
    mean=mean/err;
    err=sqrt(1.0/err);
    return 0;
  }

#endif

#ifndef DOXYGENP
}
#endif

#endif
