/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_NONREL_FERMION_H
#define O2SCL_NONREL_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/fermion.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_inte_qagiu.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Nonrelativistic fermion class

      The rest mass energy density is given by n*m not n*ms. Note that
      the effective mass here is the Landau mass, not the Dirac mass.
      
      Pressure is computed with
      \f[
      P = 2 \varepsilon/3
      \f]
      and entropy density with
      \f[
      s = \frac{5 \varepsilon}{3 T} - \frac{n \mu}{T}
      \f]
      These relations can be verified with an integration by
      parts. See, e.g. \ref Callen pg. 403 or \ref Landau pg. 164.

      \comment
      This was in an earlier version and is probably unnecessary.

      Note that the energy density integral can be rescaled:
      \f[
      \varepsilon = 
      \frac{T^{5/2}(2 m^{*})^{3/2}}{\pi^2}
      \int_0^{\infty} d u \frac{u^{3/2}}{1 + \exp(u-y)}
      \f]
      where \f$ u = k^2/2/m^{*}/T \f$ and \f$ y=\mu/T \f$.
      \endcomment

      The functions fermion::pair_density() and pair_mu() have not 
      been implemented. 

      \todo Check behaviour of calc_density() at zero density, and
      compare with that from eff_fermion, rel_fermion, and \ref
      classical.

      \todo Make sure to test with non-interacting equal to 
      true or false, and document whether or not it works
      with both inc_rest_mass equal to true or false

      \future This could be improved by performing a Chebyshev
      approximation (for example) to invert the density integral so
      that we don't need to use a solver.
  */
  class nonrel_fermion : public fermion_T {

  public:

    /// Create a nonrelativistic fermion with mass 'm' and degeneracy 'g'
    nonrel_fermion(double m=0.0, double g=0.0);

    virtual ~nonrel_fermion();
    
    /** 
	\brief Zero temperature fermions
    */
    virtual int calc_mu_zerot();

    /** 
	\brief Zero temperature fermions
    */
    virtual int calc_density_zerot();
    
    /** 
	\brief Calculate properties as function of chemical potential
     */
    virtual int calc_mu(const double temper);

    /** \brief Calculate properties as function of density

	If the density is zero, this function just sets part::mu,
	part::nu, part::ed, part::pr, and part::en to zero and returns
	without calling the error handler (even though at 
	zero density and finite	temperature, the chemical potentials
	formally are equal to \f$ -\infty \f$). 
     */
    virtual int calc_density(const double temper);

    virtual int pair_mu(const double temper) {
      return gsl_eunimpl;
    }

    virtual int pair_density(const double temper) {
      return gsl_eunimpl;
    }
    
    /// Calculate effective chemical potential from density
    virtual int nu_from_n(const double temper);

    /** \brief Set the solver for use in calculating the chemical
	potential from the density 
    */
    int set_density_root(root<double,funct<double> > &rp) {
      density_root=&rp;
      return 0;
    }

    /// The default solver for calc_density().
    cern_mroot_root<double,funct<double> > def_density_root;

    /// Return string denoting type ("nonrel_fermion")
    virtual const char *type() { return "nonrel_fermion"; }

  protected:

#ifndef DOXYGENP

    /// Solver to compute chemical potential from density
    root<double,funct<double> > *density_root;
    
    /// Function to compute chemical potential from density
    int solve_fun(double x, double &yy, double &T);

  private:

    nonrel_fermion(const nonrel_fermion &);
    nonrel_fermion& operator=(const nonrel_fermion&);

#endif

  };

  /// A zero temperature non-relativistic fermion
  class nonrel_fermion_zerot : public fermion {

  public:

    /// Create a nonrelativistic fermion with mass 'm' and degeneracy 'g'
    nonrel_fermion_zerot(double m=0.0, double g=0.0);

    virtual ~nonrel_fermion_zerot();
    
    /** 
	\brief Zero temperature fermions
    */
    virtual int calc_mu_zerot();

    /** 
	\brief Zero temperature fermions
    */
    virtual int calc_density_zerot();
    
    /// Return string denoting type ("nonrel_fermion_zerot")
    virtual const char *type() { return "nonrel_fermion_zerot"; }

#ifndef DOXYGENP

  private:

    nonrel_fermion_zerot(const nonrel_fermion_zerot &);
    nonrel_fermion_zerot& operator=(const nonrel_fermion_zerot&);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
