/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/rel_fermion.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

rel_fermion::rel_fermion(double mass, double dof) : 
  part(mass,dof), fermion_T(mass,dof) {
  deg_limit=2.0;
  nit=&def_nit;
  dit=&def_dit;
  density_root=&def_density_root;

  guess_from_nu=true;
  exp_limit=200.0;
  upper_limit_fac=20.0;
  deg_entropy_fac=30.0;
}

rel_fermion::~rel_fermion() {
}

int rel_fermion::set_inte(inte<double,funct<double> > &l_nit, 
			  inte<double,funct<double> > &l_dit) {
  nit=&l_nit;
  dit=&l_dit;
  return gsl_success;
}

int rel_fermion::calc_mu(double temper) {
  if (temper<=0.0) return fermion::calc_mu_zerot();
  if (non_interacting==true) { nu=mu; ms=m; }
  
  double T=temper;

  // If the temperature is large enough, perform the full integral
  
  if ((nu-ms)/temper<deg_limit) {
    
    funct_mfptr_noerr<rel_fermion,double> mfd(this,&rel_fermion::density_fun);
    funct_mfptr_noerr<rel_fermion,double> mfe(this,&rel_fermion::energy_fun);
    funct_mfptr_noerr<rel_fermion,double> mfs(this,&rel_fermion::entropy_fun);
    
    n=nit->integ(mfd,0.0,0.0,T);
    n*=g*pow(temper,3.0)/2.0/pi2;
    unc.n=nit->get_error()*g*pow(temper,3.0)/2.0/pi2;

    ed=nit->integ(mfe,0.0,0.0,T);
    ed*=g*pow(temper,4.0)/2.0/pi2;
    unc.ed=nit->get_error()*g*pow(temper,4.0)/2.0/pi2;
    
    en=nit->integ(mfs,0.0,0.0,T);
    en*=g*pow(temper,3.0)/2.0/pi2;
    unc.en=nit->get_error()*g*pow(temper,3.0)/2.0/pi2;

  } else {
    
    // Otherwise, apply a degenerate approximation, by making the
    // upper integration limit finite
    
    funct_mfptr_noerr<rel_fermion,double> 
      mfd(this,&rel_fermion::deg_density_fun);
    funct_mfptr_noerr<rel_fermion,double> 
      mfe(this,&rel_fermion::deg_energy_fun);
    funct_mfptr_noerr<rel_fermion,double> 
      mfs(this,&rel_fermion::deg_entropy_fun);

    double ul=sqrt(pow(upper_limit_fac*temper+nu,2.0)-ms*ms);

    n=dit->integ(mfd,0.0,ul,T);
    n*=g/2.0/pi2;
    unc.n=dit->get_error()*g/2.0/pi2;
    
    ed=dit->integ(mfe,0.0,ul,T);
    ed*=g/2.0/pi2;
    unc.ed=dit->get_error()*g/2.0/pi2;
    
    en=dit->integ(mfs,0.0,ul,T);
    en*=g/2.0/pi2;
    unc.en=dit->get_error()*g/2.0/pi2;
  }

  pr=-ed+temper*en+nu*n;
  unc.pr=sqrt(unc.ed*unc.ed+temper*unc.en*temper*unc.en+nu*unc.n*nu*unc.n);

  return gsl_success;
}

int rel_fermion::nu_from_n(double temper) {
  double nex;
  double T=temper;

  if (guess_from_nu) {
    if (nu>ms) {
      nex=nu/temper;
    } else {
      nex=ms*1.01/temper;
    }
  } else {
    nex=ms*1.01/temper;
  }
  
  funct_mfptr<rel_fermion,double> mf(this,&rel_fermion::solve_fun);
  int ret=density_root->solve(nex,T,mf);
  nu=nex*temper;
  
  return ret;
}

int rel_fermion::calc_density(double temper) {
  int mret;
  double T=temper;

  if (non_interacting==true) { nu=mu; ms=m; }
  
  mret=nu_from_n(temper);

  if (non_interacting) { mu=nu; }

  if ((nu-ms)/temper<deg_limit) {

    funct_mfptr_noerr<rel_fermion,double> mfe(this,&rel_fermion::energy_fun);
    funct_mfptr_noerr<rel_fermion,double> mfs(this,&rel_fermion::entropy_fun);

    ed=nit->integ(mfe,0.0,0.0,T);
    ed*=g*pow(temper,4.0)/2.0/pi2;
    unc.ed=nit->get_error()*g*pow(temper,4.0)/2.0/pi2;
    
    en=nit->integ(mfs,0.0,0.0,T);
    en*=g*pow(temper,3.0)/2.0/pi2;
    unc.en=nit->get_error()*g*pow(temper,3.0)/2.0/pi2;

  } else {

    funct_mfptr_noerr<rel_fermion,double> 
      mfe(this,&rel_fermion::deg_energy_fun);
    funct_mfptr_noerr<rel_fermion,double> 
      mfs(this,&rel_fermion::deg_entropy_fun);

    double ul=sqrt(pow(upper_limit_fac*temper+nu,2.0)-ms*ms);

    ed=dit->integ(mfe,0.0,ul,T);
    ed*=g/2.0/pi2;
    unc.ed=dit->get_error()*g/2.0/pi2;

    en=dit->integ(mfs,0.0,ul,T);
    en*=g/2.0/pi2;
    unc.en=dit->get_error()*g/2.0/pi2;

  }

  pr=-ed+temper*en+mu*n;
  unc.pr=sqrt(unc.ed*unc.ed+temper*unc.en*temper*unc.en+mu*unc.n*mu*unc.n);
  
  return mret;
}

double rel_fermion::deg_density_fun(double k, double &temper) {

  double E=gsl_hypot(k,ms), ret;
  
  ret=k*k/(1.0+exp((E-nu)/(temper)));
  
  return ret;
}

double rel_fermion::deg_energy_fun(double k, double &temper) {

  double E=gsl_hypot(k,ms), ret;

  ret=k*k*E/(1.0+exp((E-nu)/(temper)));
  
  return ret;
}
  
double rel_fermion::deg_entropy_fun(double k, double &temper) {
  
  double E=gsl_hypot(k,ms), ret;
  
  // If the argument to the exponential is really small, then the
  // value of the integrand is just zero
  if (((E-nu)/(temper))<-exp_limit) {
    ret=0.0;
    // Otherwise, if the argument to the exponential is still small,
    // then addition of 1 makes us lose precision, so we use an
    // alternative:
  } else if (((E-nu)/(temper))<-deg_entropy_fac) {
    double arg=E/temper-nu/temper;
    ret=-k*k*arg*exp(arg);
  } else {
    double nx=1.0/(1.0+exp(E/(temper)-nu/(temper)));
    ret=-k*k*(nx*log(nx)+(1.0-nx)*log(1.0-nx));
  }

  return ret;
}
  
double rel_fermion::density_fun(double u, double &temper) {
  double ret, y, mx;

  y=nu/(temper);
  mx=ms/(temper);
  
  if (y-mx-u>exp_limit) {
    ret=(mx+u)*sqrt(u*u+2.0*mx*u);
  } else if (y>u+exp_limit && mx>u+exp_limit) {
    ret=(mx+u)*sqrt(u*u+2.0*mx*u)/(exp(mx+u-y)+1.0);
  } else {
    ret=(mx+u)*sqrt(u*u+2.0*mx*u)*exp(y)/(exp(mx+u)+exp(y));
  }

  if (!finite(ret)) {
    ret=0.0;
  }

  return ret;
}

double rel_fermion::energy_fun(double u, double &temper) {
  double ret, y, mx;

  y=nu/(temper);
  mx=ms/(temper);
  
  if (y>u+exp_limit && mx>u+exp_limit) {
    ret=(mx+u)*(mx+u)*sqrt(u*u+2.0*mx*u)/(exp(mx+u-y)+1.0);
  } else {
    ret=(mx+u)*(mx+u)*sqrt(u*u+2.0*mx*u)*exp(y)/(exp(mx+u)+exp(y));
  }
  
  if (!finite(ret)) {
    return 0.0;
  }

  return ret;
}

double rel_fermion::entropy_fun(double u, double &temper) {
  double ret, y, mx, term1, term2;

  y=mu/(temper);
  mx=ms/(temper);

  term1=log(1.0+exp(y-mx-u))/(1.0+exp(y-mx-u));
  term2=log(1.0+exp(mx+u-y))/(1.0+exp(mx+u-y));
  ret=(mx+u)*sqrt(u*u+2.0*mx*u)*(term1+term2);
  
  if (!finite(ret)) {
    return 0.0;
  }

  return ret;
}

int rel_fermion::solve_fun(double x, double &yy, double &temper) {
  double nden;
  
  nu=(temper)*x;

  if (non_interacting) mu=nu;

  if ((nu-ms)/(temper)< deg_limit) {
    funct_mfptr_noerr<rel_fermion,double> mfe(this,&rel_fermion::density_fun);

    nden=nit->integ(mfe,0.0,0.0,temper);
    nden*=g*pow((temper),3.0)/2.0/pi2;
    unc.n=nit->get_error()* g*pow((temper),3.0)/2.0/pi2;
    
    yy=(n-nden)/n;

  } else {
    
    funct_mfptr_noerr<rel_fermion,double> 
      mfe(this,&rel_fermion::deg_density_fun);
    
    double ul=sqrt(pow(upper_limit_fac*temper+nu,2.0)-ms*ms);

    nden=dit->integ(mfe,0.0,ul,temper);
    nden*=g/2.0/pi2;
    unc.n=err_hnd->get_errno()*g/2.0/pi2;

    yy=(n-nden)/n;
  }
  
  return gsl_success;
}

int rel_fermion::pair_mu(double temper) {
  int ret=gsl_success;

  rel_fermion antip(ms,g);

  if (non_interacting) { nu=mu; ms=m; }
  anti(antip);

  ret=calc_mu(temper);
  if (ret==0) {
    ret=antip.calc_mu(temper);
    n-=antip.n;
    pr+=antip.pr;
    ed+=antip.ed;
    en+=antip.en;
    unc.n=sqrt(unc.n*unc.n+antip.unc.n*antip.unc.n);
    unc.ed=sqrt(unc.ed*unc.ed+antip.unc.ed*antip.unc.ed);
    unc.pr=sqrt(unc.pr*unc.pr+antip.unc.pr*antip.unc.pr);
    unc.en=sqrt(unc.en*unc.en+antip.unc.en*antip.unc.en);
  }

  return ret;

}

int rel_fermion::pair_density(double temper) {
  double nex;
  int ret;

  if (temper<=0.0) return fermion::calc_density_zerot();
  if (non_interacting==true) { nu=mu; ms=m; }

  double T=temper;

  nex=nu/temper;
  funct_mfptr<rel_fermion,double> mf(this,&rel_fermion::pair_fun);
  ret=density_root->solve(nex,T,mf);
  if (ret!=0) {
    O2SCL_ERR_RET("Solver failed in rel_fermion::pair_density().",ret);
  }
  nu=nex*temper;
  
  if (non_interacting==true) { mu=nu; }
  
  ret=pair_mu(temper);

  return ret;

}

int rel_fermion::pair_fun(double x, double &yy, double &temper) {
  double nden;

  nu=(temper)*x;

  if (non_interacting) mu=nu;

  if ((nu-ms)/(temper)< deg_limit) {
    funct_mfptr_noerr<rel_fermion,double> mfe(this,&rel_fermion::density_fun);

    nden=nit->integ(mfe,0.0,0.0,temper);
    nden*=g*pow((temper),3.0)/2.0/pi2;
    yy=nden;
  } else {
    funct_mfptr_noerr<rel_fermion,double> 
      mfe(this,&rel_fermion::deg_density_fun);

    double ul=sqrt(pow(upper_limit_fac*temper+nu,2.0)-ms*ms);

    nden=dit->integ(mfe,0.0,ul,temper);
    nden*=g/2.0/pi2;
    yy=nden;
  }

  nu=-(temper)*x ;
  if ((nu-ms)/(temper)< deg_limit) {
    funct_mfptr_noerr<rel_fermion,double> mf(this,&rel_fermion::density_fun);

    nden=nit->integ(mf,0.0,0.0,temper);
    nden*=g*pow((temper),3.0)/2.0/pi2;
    yy-=nden;
  } else {
    funct_mfptr_noerr<rel_fermion,double> 
      mf(this,&rel_fermion::deg_density_fun);
    
    double ul=sqrt(pow(upper_limit_fac*temper+nu,2.0)-ms*ms);

    nden=dit->integ(mf,0.0,ul,temper);
    nden*=g/2.0/pi2;
    yy-=nden;
  }

  yy=yy/n-1.0;

  return gsl_success;
}

