/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_REL_FERMION_H
#define O2SCL_REL_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/fermion.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_inte_qagiu.h>
#include <o2scl/gsl_inte_qag.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Equation of state for a relativistic fermion

      \hline 
      <b>Method:</b>

      This implements an equation of state for a relativistic fermion
      using direct integration. Define the degeneracy parameter 
      \f[
      \psi=(\nu-m^{*})/T 
      \f] 
      where \f$ \nu \f$ is the effective chemical potential and \f$
      m^{*} \f$ is the effective mass. For \f$ \psi \f$ greater than
      \ref deg_limit (degenerate regime), a finite interval integrator
      is used and for \f$ \psi \f$ less than \ref deg_limit
      (non-degenerate regime), an integrator over the interval from
      \f$ [0,\infty) \f$ is used. The upper limit on the degenerate
      integration is given by
      \f[
      \sqrt{(u T+{\nu})^2-m^{*,2}}
      \f]
      where \f$ u \f$ is \ref rel_fermion::upper_limit_fac .

      In the degenerate regime, the entropy integrand
      \f[
      - k^2 \left[ n \log n + \left(1-n\right) \log 
      \left(1-n \right) \right]
      \f]
      can lose precision when \f$ (E - \nu)/T \f$ is negative
      and sufficiently large in absolute magnitude. Thus when 
      \f$ (E - \nu)/T < S \f$ where \f$ S \f$ is stored in 
      \ref deg_entropy_fac (default is -30), the integrand
      is written as
      \f[
      -k^2 \left( E/T-\nu/T \right) e^{E/T-\nu/T} \, .
      \f]
      If \f$ (E - \nu)/T < S \f$ is less than -1 times \ref exp_limit
      (e.g. less than -200), then the entropy integrand is assumed 
      to be zero.
      
      The integrands in the non-degenerate regime are written
      in a dimensionless form, by defining 
      \f$ u \equiv p/T, y \equiv \nu/T, \f$ and 
      \f$ \mathrm{mx} \equiv m^{*}/T \f$. 
      The density integrand is 
      \f[
      \left(\mathrm{mx}+u\right) \sqrt{u^2+2 \mathrm{mx} u}
      \left(\frac{e^{y}}{e^{\mathrm{mx}+u}+e^{y}}\right) \, , 
      \f]
      the energy integrand is 
      \f[
      \left(\mathrm{mx}+u\right)^2 \sqrt{u^2+2 \mathrm{mx} u}
      \left(\frac{e^{y}}{e^{\mathrm{mx}+u}+e^{y}}\right) \, ,
      \f]
      and the entropy integrand is 
      \f[
      \left(\mathrm{mx}+u\right) \sqrt{u^2+2 \mathrm{mx} u} 
      \left(t_1+t_2\right) \, ,
      \f]
      where 
      \f{eqnarray*}
      t_1 &=& \log \left(1+e^{y-\mathrm{mx}-u}\right)/
      \left(1+e^{y-\mathrm{mx}-u}\right) \nonumber \\
      t_2 &=& \log \left(1+e^{\mathrm{mx}+u-y}\right)/
      \left(1+e^{\mathrm{mx}+u-y}\right) \, .
      \f}

      The default integrators are gsl_inte_qag (for degenerate particles)
      and gsl_inte_qagiu (for non-degenerate particles).

      \hline 
      <b>Accuracy:</b>

      The default settings for for this class give an accuracy of at
      least 1 part in \f$ 10^6 \f$ (and frequently better than this).

      When the integrators provide numerical uncertainties, these
      uncertainties are stored in \ref unc. In the case of
      calc_density() and pair_density(), the uncertainty from the
      numerical accuracy of the solver is not included. (There is also
      a relatively small inaccuracy due to the mathematical evaluation
      of the integrands which is not included in \ref unc.)
     
      One way to improve the accuracy of the computation is just to 
      decrease the tolerances on the default integration objects. This 
      can be done, using, for example
      \code
      rel_fermion rf(1.0,2.0);
      rf.def_dit.tolx/=1.0e2;
      rf.def_dit.tolf/=1.0e2;
      rf.def_nit.tolx/=1.0e2;
      rf.def_nit.tolf/=1.0e2;
      \endcode
      which decreases the both the relative and absolute tolerances
      for both the degenerate and non-degenerate integrators. If one
      is using either the calc_density() or pair_density() functions,
      one may also have to improve the accuracy of the solver which
      determines the chemical potential from the density. For
      the default solver, this could be done with
      \code
      rf.def_density_root.tolx/=1.0e2;
      rf.def_density_root.tolf/=1.0e2;
      \endcode
      Of course if these tolerances are too small, the calculation
      may fail.

      \hline 
      <b>Todos:</b>

      \note This does not work with inc_rest_mass=false (3/30/09:
      There is a significant amount of testing code to ensure that
      it does work with inc_rest_mass=false, so this may have
      been fixed already.)

      \future Improve the asymptotics of the non-degenerate 
      integrands.

      \future It appears this doesn't compute the uncertainty in the
      chemical potential or density with calc_density(). This could
      be fixed.
      
  */
  class rel_fermion : public fermion_T {

  public:

    /// \name Numerical parameters
    //@{
    /** \brief The critical degeneracy at which to switch integration 
	techniques (default 2)
    */
    double deg_limit;
    
    /** \brief The limit for exponentials to ensure integrals are finite 
	(default 200)
    */
    double exp_limit;

    /// The factor for the degenerate upper limits (default 20)
    double upper_limit_fac;

    /// A factor for the degenerate entropy integration (default 30)
    double deg_entropy_fac;
    //@}

    /// Storage for the uncertainty
    fermion unc;

    /** \brief If true, use the present value of the chemical potential as
	a guess for the new chemical potential
    */
    bool guess_from_nu;
    
    /// Create a fermion with mass \c m and degeneracy \c g
    rel_fermion(double m=0.0, double g=0.0);
    virtual ~rel_fermion();
    
    /** 
	\brief Calculate properties as function of chemical potential
    */
    virtual int calc_mu(double temper);

    /** \brief Calculate properties as function of density
     */
    virtual int calc_density(double temper);

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual int pair_mu(double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
     */
    virtual int pair_density(double temper);

    /// Calculate effective chemical potential from density
    virtual int nu_from_n(double temper);
    
    /// Set integrators
    int set_inte(inte<double,funct<double> > &non_it, 
		 inte<double,funct<double> > &deg_it);
    
    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_root(root<double,funct<double> > &rp) {
      density_root=&rp;
      return 0;
    }
    
    /// The default solver for calc_density().
    cern_mroot_root<double,funct<double> > def_density_root;

    /// The default integrator for degenerate fermions
    gsl_inte_qag<double,funct<double> > def_dit;

    /// The default integrator for non-degenerate fermions
    gsl_inte_qagiu<double,funct<double> > def_nit;

    /// Return string denoting type ("rel_fermion")
    virtual const char *type() { return "rel_fermion"; }

  protected:
    
#ifndef DOXYGEN_INTERNAL
    
    /// The non-degenerate integrator
    inte<double,funct<double> > *nit;
    /// The degenerate integrator
    inte<double,funct<double> > *dit;
    /// The solver for calc_density()
    root<double,funct<double> > *density_root;

    /// The integrand for the density for non-degenerate fermions
    double density_fun(double u, double &pa);

    /// The integrand for the energy density for non-degenerate fermions
    double energy_fun(double u, double &pa);

    /// The integrand for the entropy density for non-degenerate fermions
    double entropy_fun(double u, double &pa);

    /// The integrand for the density for degenerate fermions
    double deg_density_fun(double u, double &pa);

    /// The integrand for the energy density for degenerate fermions
    double deg_energy_fun(double u, double &pa);

    /// The integrand for the entropy density for degenerate fermions
    double deg_entropy_fun(double u, double &pa);

    /// Solve for the chemical potential given the density
    int solve_fun(double x, double &yy, double &pa);

    /// Solve for the chemical potential given the density with antiparticles
    int pair_fun(double x, double &yy, double &pa);
    
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
