/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SN_FERMION_H
#define O2SCL_SN_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/inte.h>
#include <o2scl/deriv_part.h>
#include <o2scl/gsl_inte_qag.h>
#include <o2scl/gsl_inte_qagiu.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Equation of state for a relativistic fermion

      \note This class does not work with inc_rest_mass=true.

      This implements an equation of state for a relativistic fermion
      using direct integration. After subtracting the rest mass from
      the chemical potentials, the distribution function is
      \f[
      \left\{1+\exp[(\sqrt{k^2+m^{* 2}}-m-\nu)/T]\right\}^{-1}
      \f]
      where \f$ k \f$ is the momentum, \f$ \nu \f$ is the effective
      chemical potential, \f$ m \f$ is the rest mass, and \f$ m^{*}
      \f$ is the effective mass.  For later use, we define \f$ E^{*} =
      \sqrt{k^2 + m^{*2}} \f$ . The degeneracy parameter is
      \f[
      \psi=(\nu+(m-m^{*}))/T 
      \f] 
      For \f$ \psi \f$ greater than \ref deg_limit (degenerate
      regime), a finite interval integrator is used and for \f$ \psi
      \f$ less than \ref deg_limit (non-degenerate regime), an
      integrator over the interval from \f$ [0,\infty) \f$ is
      used. The upper limit on the degenerate integration is given by
      the solution of
      \f[
      (\sqrt{k^2+m^{*,2}}-m-\nu)/T=\mathrm{f{l}imit}
      \f]
      which is
      \f[
      \sqrt{(m+{\cal L})^2-m^{*2}}
      \f]
      where \f$ {\cal L}\equiv\mathrm{f{l}imit}\times T+\nu \f$ .

      In the non-degenerate regime, we make the substitution
      \f$ u=k/T \f$ to ensure that the variable of integration
      scales properly.

      Uncertainties are given in \ref unc.

      \todo This needs to be corrected to calculate \f$ \sqrt{k^2+m^{*
      2}}-m \f$ gracefully when \f$ m^{*}\approx m << k \f$ .
      \todo Call error handler if inc_rest_mass is true or update
      to properly treat the case when inc_rest_mass is true.
      
      \b Evaluation \b of \b the \b derivatives

      The relevant
      derivatives of the distribution function are
      \f[
      \frac{\partial f}{\partial T}=
      f(1-f)\frac{E^{*}-m-\nu}{T^2}
      \f]
      \f[
      \frac{\partial f}{\partial \nu}=
      f(1-f)\frac{1}{T}
      \f]
      \f[
      \frac{\partial f}{\partial k}=
      -f(1-f)\frac{k}{E^{*} T}
      \f]
      \f[
      \frac{\partial f}{\partial m^{*}}=
      -f(1-f)\frac{m^{*}}{E^{*} T}
      \f]

      We also need the derivative of the entropy integrand w.r.t. the 
      distribution function, which is
      \f[
      {\cal S}\equiv f \ln f +(1-f) \ln (1-f) \qquad
      \frac{\partial {\cal S}}{\partial f} = \ln 
      \left(\frac{f}{1-f}\right) = 
      \left(\frac{\nu-E^{*}+m}{T}\right)
      \f]
      where the entropy density is
      \f[
      s = - \frac{g}{2 \pi^2} \int_0^{\infty} {\cal S} k^2 d k
      \f]

      The derivatives can be integrated directly (\ref method = \ref
      direct) or they may be converted to integrals over the
      distribution function through an integration by parts (\ref
      method = \ref byparts)
      \f[
      \int_a^b f(k) \frac{d g(k)}{dk} dk = \left.f(k) g(k)\right|_{k=a}^{k=b}
      - \int_a^b g(k) \frac{d f(k)}{dk} dk 
      \f]
      using the distribution function for \f$ f(k) \f$ and 0 and 
      \f$ \infty \f$ as the limits, we have
      \f[
      \frac{g}{2 \pi^2} \int_0^{\infty} \frac{d g(k)}{dk} f dk =
      \frac{g}{2 \pi^2} \int_0^{\infty} g(k) f (1-f) \frac{k}{E^{*} T} dk 
      \f]
      as long as \f$ g(k) \f$ vanishes at \f$ k=0 \f$ .
      Rewriting,
      \f[
      \frac{g}{2 \pi^2} \int_0^{\infty} h(k) f (1-f) dk =
      \frac{g}{2 \pi^2} \int_0^{\infty} f \frac{T}{k} 
      \left[ h^{\prime} E^{*}-\frac{h E^{*}}{k}+\frac{h k}{E^{*}} \right] dk
      \f]
      as long as \f$ h(k)/k \f$ vanishes at \f$ k=0 \f$ .

      \b Explicit \b forms

      1) The derivative of the density wrt the chemical potential
      \f[
      \left(\frac{d n}{d \mu}\right)_T = 
      \frac{g}{2 \pi^2} \int_0^{\infty} \frac{k^2}{T} f (1-f) dk
      \f]
      Using \f$ h(k)=k^2/T \f$ we get
      \f[
      \left(\frac{d n}{d \mu}\right)_T = 
      \frac{g}{2 \pi^2} \int_0^{\infty} 
      \left(\frac{k^2+E^{*2}}{E^{*}}\right) f dk
      \f]

      2) The derivative of the density wrt the temperature
      \f[
      \left(\frac{d n}{d T}\right)_{\mu} = 
      \frac{g}{2 \pi^2} \int_0^{\infty} \frac{k^2(E^{*}-m-\nu)}{T^2} 
      f (1-f) dk
      \f]
      Using \f$ h(k)=k^2(E^{*}-\nu)/T^2 \f$ we get
      \f[
      \left(\frac{d n}{d T}\right)_{\mu} = 
      \frac{g}{2 \pi^2} \int_0^{\infty} \frac{f}{T} 
      \left[2 k^2+E^{*2}-E^{*}\left(\nu+m\right)-
      k^2 \left(\frac{\nu+m}{E^{*}}\right)\right] dk
      \f]

      3) The derivative of the entropy wrt the chemical potential
      \f[
      \left(\frac{d s}{d \mu}\right)_T = 
      \frac{g}{2 \pi^2} \int_0^{\infty} k^2 f (1-f) 
      \frac{(E^{*}-m-\nu)}{T^2} dk
      \f]
      This verifies the Maxwell relation
      \f[
      \left(\frac{d s}{d \mu}\right)_T =
      \left(\frac{d n}{d T}\right)_{\mu}
      \f]

      4) The derivative of the entropy wrt the temperature
      \f[
      \left(\frac{d s}{d T}\right)_{\mu} = 
      \frac{g}{2 \pi^2} \int_0^{\infty} k^2 f (1-f) 
      \frac{(E^{*}-m-\nu)^2}{T^3} dk
      \f]
      Using \f$ h(k)=k^2 (E^{*}-\nu)^2/T^3 \f$ 
      \f[
      \left(\frac{d s}{d T}\right)_{\mu} = 
      \frac{g}{2 \pi^2} \int_0^{\infty} \frac{f(E^{*}-m-\nu)}{E^{*}T^2} 
      \left[E^{* 3}+3 E^{*} k^2- (E^{* 2}+k^2)(\nu+m)\right] d k
      \f]

      5) The derivative of the density wrt the effective mass
      \f[
      \left(\frac{d n}{d m^{*}}\right)_{T,\mu} = 
      -\frac{g}{2 \pi^2} \int_0^{\infty} 
      \frac{k^2 m^{*}}{E^{*} T} f (1-f) dk
      \f]
      Using \f$ h(k)=-(k^2 m^{*})/(E^{*} T) \f$ we get
      \f[
      \left(\frac{d n}{d m^{*}}\right)_{T,\mu} = 
      -\frac{g}{2 \pi^2} \int_0^{\infty} 
      m^{*} f dk
      \f]

      \note The dsdT integration may fail if the system is
      very degenerate. When method is byparts, the integral involves a
      large cancellation between the regions from \f$ k \in (0,
      \mathrm{ulimit/2}) \f$ and \f$ k \in (\mathrm{ulimit/2},
      \mathrm{ulimit}) \f$. Switching to method=direct and setting the
      lower limit to \f$ \mathrm{llimit} \f$, may help, but recent
      testing on this gave negative values for dsdT. For very
      degenerate systems, an expansion may be better than trying
      to perform the integration. 

      \future It might be worth coding up direct differentiation, or
      differentiating the eff results, as these may succeed more
      generally.

      \future This class will have difficulty with extremely degenerate
      or extremely non-degnerate systems. Fix this.
      \future Create a more intelligent method for dealing with bad 
      initial guesses for the chemical potential in calc_density().
  */
  class sn_fermion : public fermion_T, public deriv_part {
    
  public:

    /// Create a fermion with mass \c m and degeneracy \c g
    sn_fermion(double m=0.0, double g=0.0);
    virtual ~sn_fermion();
    
    /** \brief The critical degeneracy at which to switch integration 
	techniques (default 2.0)
    */
    double deg_limit;
    
    /** 
	\brief The limit for the Fermi functions (default 20.0)
	
	sn_fermion will ignore corrections smaller than about
	\f$ \exp(-\mathrm{f{l}imit}) \f$ . 
    */
    double flimit;
    
    /// Storage for the most recently calculated uncertainties 
    fermion unc;
    /// Storage for the most recently calculated uncertainties 
    deriv_part dunc;
    
    /** \name Method of computing derivatives
     */
    //@{
    /// Method (default is \ref byparts)
    int method;
    /// In the form containing \f$ f(1-f) \f$ .
    static const int direct=1;
    /// Integrate by parts
    static const int byparts=2;
    //@}

    /** 
	\brief Calculate properties as function of chemical potential
    */
    virtual int calc_mu(double temper);

    /** \brief Calculate properties as function of density
     */
    virtual int calc_density(double temper);

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual int pair_mu(double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
     */
    virtual int pair_density(double temper);

    /// Calculate effective chemical potential from density
    virtual int nu_from_n(double temper);

    /** 
	\brief Set inte objects
	
	The first integrator is used for non-degenerate integration
	and should integrate from 0 to \f$ \infty \f$ (like \ref
	gsl_inte_qagiu). The second integrator is for the degenerate
	case, and should integrate between two finite values.
    */
    int set_inte(inte<double,funct<double> > &unit, 
		 inte<double,funct<double> > &udit);

    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_root(root<double,funct<double> > &rp) {
      density_root=&rp;
      return 0;
    }

    /// The default integrator for the non-degenerate regime
    gsl_inte_qagiu<double,funct<double> > def_nit;

    /// The default integrator for the degenerate regime
    gsl_inte_qag<double,funct<double> > def_dit;

    /// The default solver for npen_density() and pair_density()
    cern_mroot_root<double,funct<double> > def_density_root;

    /// Return string denoting type ("sn_fermion")
    virtual const char *type() { return "sn_fermion"; };

  protected:

#ifndef DOXYGENP

    /// The integrator for non-degenerate fermions
    inte<double,funct<double> > *nit;

    /// The integrator for degenerate fermions
    inte<double,funct<double> > *dit;

    /// The solver for calc_density() and pair_density()
    root<double,funct<double> > *density_root;

    /** \name The integrands, as a function of \f$ u=k/T \f$, for 
	non-degenerate integrals
    */
    //@{
    double density_fun(double u, double &temper);
    double energy_fun(double u, double &temper);
    double entropy_fun(double u, double &temper);
    double density_T_fun(double k, double &temper);
    double density_mu_fun(double k, double &temper);
    double entropy_T_fun(double k, double &temper);
    double density_ms_fun(double k, double &temper);
    //@}

    /** \name The integrands, as a function of momentum, for the 
	degenerate integrals
    */
    //@{
    double deg_density_fun(double u, double &temper);
    double deg_energy_fun(double u, double &temper);
    double deg_entropy_fun(double u, double &temper);
    double deg_density_T_fun(double k, double &temper);
    double deg_density_mu_fun(double k, double &temper);
    double deg_entropy_T_fun(double k, double &temper);
    double deg_density_ms_fun(double k, double &temper);
    //@}

    /** \name The function pointers for the non-degenerate integrals
     */
    //@{
    funct_mfptr_noerr<sn_fermion,double> *density_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *energy_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *entropy_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *density_mu_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *density_T_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *entropy_T_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *density_ms_fun_f;
    //@}

    /** \name The function pointers for the degenerate integrals
     */
    //@{
    funct_mfptr_noerr<sn_fermion,double> *deg_density_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *deg_energy_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *deg_entropy_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *deg_density_mu_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *deg_density_T_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *deg_entropy_T_fun_f;
    funct_mfptr_noerr<sn_fermion,double> *deg_density_ms_fun_f;
    //@}

    /** \brief Solve for the chemical potential from the density 
	for calc_density()
    */
    int solve_fun(double x, double &yy, double &temper);

    /** \brief Solve for the chemical potential from the density 
	for pair_density()
    */
    int pair_fun(double x, double &yy, double &temper);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
