/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/* multiroots/enorm.c
 * multiroots/dogleg.c
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2007 Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/gsl_mroot_hybrids_b.h>

#include <iostream>
#include <o2scl/ovector_tlate.h>
#include <o2scl/omatrix_tlate.h>

using namespace std;
using namespace o2scl;

//-------------------------------------------------
// Functions from multiroots/enorm.c and 
// multiroots/dogleg.c
//-------------------------------------------------

double mroot_hybrids_base::enorm(const gsl_vector *f) {
  double e2 = 0 ;
  size_t i, n = f->size ;
  for (i = 0; i < n ; i++) {
    double fi= gsl_vector_get(f, i);
    e2 += fi * fi ;
  }
  return sqrt(e2);
}

double mroot_hybrids_base::scaled_enorm (const gsl_vector *d, 
				   const gsl_vector *f) {
  double e2 = 0 ;
  size_t i, n = f->size ;
  for (i = 0; i < n ; i++) {
    double fi= gsl_vector_get(f, i);
    double di= gsl_vector_get(d, i);
    double u = di * fi;
    e2 += u * u ;
  }
  return sqrt(e2);
}

double mroot_hybrids_base::enorm_sum(const gsl_vector *a, const gsl_vector *b) {
  double e2 = 0 ;
  size_t i, n = a->size ;
  for (i = 0; i < n ; i++) {
    double ai= gsl_vector_get(a, i);
    double bi= gsl_vector_get(b, i);
    double u = ai + bi;
    e2 += u * u ;
  }
  return sqrt(e2);
}

double mroot_hybrids_base::compute_actual_reduction(double fnorm, double fnorm1) {
  double actred;
  
  if (fnorm1 < fnorm) {
    double u = fnorm1 / fnorm;
    actred = 1 - u * u;
  } else {
    actred = -1;
  }
  
  return actred;
}

double mroot_hybrids_base::compute_predicted_reduction(double fnorm, 
						 double fnorm1) {
  double prered;

  if (fnorm1 < fnorm) {
    double u = fnorm1 / fnorm;
    prered = 1 - u * u;
  } else {
    prered = 0;
  }

  return prered;
}

void mroot_hybrids_base::compute_qtf(const gsl_matrix * q, 
			       const gsl_vector *f, gsl_vector *qtf) {
  size_t i, j, N = f->size ;

  for (j = 0; j < N; j++) {
    double sum = 0;
    for (i = 0; i < N; i++) {
      sum += gsl_matrix_get (q, i, j) * gsl_vector_get (f, i);
    }
    gsl_vector_set (qtf, j, sum);
  }
}

int mroot_hybrids_base::newton_direction(const gsl_matrix * r, 
				   const gsl_vector *qtf, gsl_vector *p) {
  const size_t N = r->size2;
  size_t i;
  int status;

  status = gsl_linalg_R_solve (r, qtf, p);

  for (i = 0; i < N; i++) {
    double pi = gsl_vector_get (p, i);
    gsl_vector_set (p, i, -pi);
  }
  
  return status;
}

void mroot_hybrids_base::gradient_direction(const gsl_matrix * r, 
				      const gsl_vector *qtf,
				      const gsl_vector *diag, 
				      gsl_vector *g) {
  const size_t M = r->size1;
  const size_t N = r->size2;

  size_t i, j;

  for (j = 0; j < M; j++) {

    double sum = 0;
    double dj;

    for (i = 0; i < N; i++) {
      sum += gsl_matrix_get (r, i, j) * gsl_vector_get (qtf, i);
    }

    dj = gsl_vector_get (diag, j);
    gsl_vector_set (g, j, -sum / dj);
  }
}

void mroot_hybrids_base::minimum_step(double gnorm, const gsl_vector *diag, 
				gsl_vector *g) {
  const size_t N = g->size;
  size_t i;

  for (i = 0; i < N; i++) {

    double gi = gsl_vector_get (g, i);
    double di = gsl_vector_get (diag, i);
    gsl_vector_set (g, i, (gi / gnorm) / di);
  }
}

